/* a basic ray-tracer */
/* http://www.scratchapixel.com/lessons/3d-basic-rendering/introduction-to-ray-tracing/implementing-the-raytracing-algorithm */
/* https://en.wikipedia.org/wiki/Ray_tracing_(graphics) */

#define _USE_MATH_DEFINES
#include <math.h>
#include <time.h>
#include "elabmm.h"

extern const float NaNf = ((float)((1e308 * 10)*0.));
extern const float Inff = ((float)(1e308 * 10));

#define DEG2RAD(A) ((float) (A) * (float) M_PI / 180.0F)
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))

struct elmCOORD {
	float x, y, z;
};

struct elmCLR {
	float r, g, b;
};

struct elmMTR {
	struct elmCLR sc; /* basic color */
	float rr; /* reflective factor */
	float rf; /* refractive factor */
	float em; /* emissive facor */
};

struct elmROTM {
	float xx, xy, xz;
	float yx, yy, yz;
	float zx, zy, zz;
};

struct elmRAY {
	struct elmCOORD s;
	struct elmCOORD d;
};

struct elmCAM {
	struct elmCOORD o;
	float theta, phi;
	float fov;
	float ar;
	struct {
		float hs, vs, dx, dy;
	} wk;
};

struct elmPOLY {
	struct elmCOORD v1, v2, v3;
	struct elmMTR sm;
};

struct elmSPHE {
	struct elmCOORD c;
	float r;
	struct elmMTR sm;
};

struct elmOBJ {
	enum {ELMP, ELMS} t;
	union {
		struct elmPOLY p;
		struct elmSPHE s;
	} o;
	struct elmMTR *sm;
};

struct elmIMGF {
	int width, height;
	float *buf;
};

struct elmCLR skycolor = {0.25F,0.25F,0.3F};

struct elmCOORD versor(struct elmCOORD *v1, struct elmCOORD *v2)
{
	/* versor from v1 towards v2 */
	struct elmCOORD d = {v2->x - v1->x, v2->y - v1->y, v2->z - v1->z};
	float len = sqrtf(d.x * d.x + d.y * d.y + d.z * d.z);
	d.x /= len;
	d.y /= len;
	d.z /= len;
	return d;
}

int prepare_cam(struct elmCAM *cam, struct elmIMGF *imgf)
{
	/* camera size on the unit circle */
	cam->wk.hs = 2.0F * tanf(cam->fov / 2.0F);
	cam->wk.dx = cam->wk.hs / (float) imgf->width;
	cam->wk.dy = cam->wk.dx;
	cam->wk.vs = cam->wk.hs / cam->ar;

	return 0;
}

int primary_ray(struct elmRAY *pray, struct elmCAM *cam, int row, int col, int dorandom)
{
	struct elmCOORD orig = {0, 0, 0};
	struct elmCOORD pixelcoord = {
		cosf(cam->fov / 2.0F),
		cam->wk.hs / 2.0F - cam->wk.dx * (col + dorandom * ((float) rand() / (float) RAND_MAX)),
		cam->wk.vs / 2.0F - cam->wk.dy * (row + dorandom * ((float) rand() / (float) RAND_MAX))
	};

	/* origin of ray is the cam */
	memcpy(&pray->s, &cam->o, sizeof(struct elmCOORD));

	/* direction versor */
	pray->d = versor(&orig, &pixelcoord);

	/* rotate around z-axis according to theta */
	pray->d.x = pray->d.x * cosf(cam->theta) - pray->d.y * sinf(cam->theta);
	pray->d.y = pray->d.x * sinf(cam->theta) + pray->d.y * cosf(cam->theta);

	/* rotate around new x-axis according to phi */
	pray->d.x = -pray->d.z * sinf(cam->phi) + pray->d.x * cosf(cam->phi);
	pray->d.z =  pray->d.z * cosf(cam->phi) + pray->d.x * sinf(cam->phi);

	return 0;
}

float intersection(struct elmCOORD *hitp, struct elmCOORD *hitn, struct elmOBJ *obj, struct elmRAY *pray) 
{
	if (obj->t == ELMS) {

		/* sphere */
		struct elmSPHE *sphe = &(obj->o.s);
		struct elmCOORD v = {pray->s.x - sphe->c.x, pray->s.y - sphe->c.y, pray->s.z - sphe->c.z};
		float vd = (v.x * pray->d.x + v.y * pray->d.y + v.z * pray->d.z);
		float disc = vd * vd - (v.x * v.x + v.y * v.y + v.z * v.z) + sphe->r * sphe->r;
		if (disc < 0)
			return -1;
		else {
			float discs = sqrtf(disc);
			float t1 = - vd + discs, t2 = - vd - discs;
			if (t2 >= 0) {
				if (hitp) {
					hitp->x = pray->s.x + t2 * pray->d.x; 
					hitp->y = pray->s.y + t2 * pray->d.y; 
					hitp->z = pray->s.z + t2 * pray->d.z; 
					if (hitn)
						*hitn = versor(&sphe->c, hitp);
				}
				return t2;
			} else if (t1 >= 0) {
				if (hitp) {
					hitp->x = pray->s.x + t1 * pray->d.x; 
					hitp->y = pray->s.y + t1 * pray->d.y; 
					hitp->z = pray->s.z + t1 * pray->d.z; 
					if (hitn)
						*hitn = versor(&sphe->c, hitp);
				}
				return t1;
			} else
				return -1;
		}

	} else if (obj->t == ELMP) {

		/* polygon */
		return -1;
	}

	return -1;
}

struct elmCLR trace(struct elmRAY *ray, struct elmOBJ *obj, int nobj)
{
	struct elmCLR ilmout = {1,1,1};
	int oo;

	/* trace ray */
	float mindist = +Inff;
	struct elmOBJ *cobj = NULL;

	/* search all objects */
	for (oo = 0; oo < nobj; oo++) {

		/* find intersection if any (not with lights) */
		float dist;
		if (obj[oo].sm->em == 0 && (dist = intersection(NULL, NULL, &obj[oo], ray)) >= 0) {

			/* save nearest object */
			if (dist < mindist) {
				mindist = dist;
				cobj = &obj[oo];
			}
		}
	}

	/* from far object */
	if (cobj) {

		/* final color */
		ilmout.r = cobj->sm->sc.r;
		ilmout.g = cobj->sm->sc.g;
		ilmout.b = cobj->sm->sc.b;

	} else {

		/* hit the sky */
		ilmout.r = skycolor.r;
		ilmout.g = skycolor.g;
		ilmout.b = skycolor.b;
	}

	return ilmout;
}

int main(void)
{
	int row = 0, col = 0, samp = 0;
	struct elmIMGF imgf = {640, 480};
	struct elmCAM cam = {
		{-2, 0, 1}, /* o */
		DEG2RAD(0), /* theta */
		DEG2RAD(0), /* phi */
		DEG2RAD(60), /* fov */
		(float) imgf.width / (float) imgf.height /* ar */
	};
	struct elmPOLY polys[] = {
		{{4,0,0}, {5,0,0}, {4,1,0}}
	};
	int npoly = sizeof(polys) / sizeof(struct elmPOLY);
	struct elmSPHE sphes[] = {
		{{4,0,0.8F}, 0.8F, {{0.2F,1,0.5F}, 1, 0, 0}}, // large frontal
		{{3,1.1F,0.8F}, 0.5F, {{1,0,0}, 1, 0, 0}}, // small side
		{{0,0,-100000}, 100000, {{0.95F,0.95F,0.8F}, 0, 0, 0}}, // floor
		{{3.8F,0,2.5F}, 0.3F, {{1,1,1}, 0, 0, 1}} // white light
	};
	int nsphe = sizeof(sphes) / sizeof(struct elmSPHE);
	struct elmOBJ obj[(sizeof(polys)/sizeof(struct elmPOLY)) + (sizeof(sphes)/sizeof(struct elmSPHE))] = {0};
	int nobj = 0, oo;
	int nsamples = 0;

	/* prepare list of objects */
	for (oo = 0; oo < npoly; oo++, nobj++) {
		obj[nobj].t = ELMP;
		obj[nobj].o.p = polys[oo];
		obj[nobj].sm = &obj[nobj].o.p.sm;
	}
	for (oo = 0; oo < nsphe; oo++, nobj++) {
		obj[nobj].t = ELMS;
		obj[nobj].o.s = sphes[oo];
		obj[nobj].sm = &obj[nobj].o.s.sm;
	}

	/* initialize random seed: */
	srand((unsigned int) time(NULL));

	/* prepare image */
	imgf.buf = (float *) calloc(imgf.height * imgf.width * 3, sizeof(float));

	/* prepare camera */
	prepare_cam(&cam, &imgf);

	/* scan image pixels */
	for (row = 0; row < imgf.height; row++) {
		for (col = 0; col < imgf.width; col++) {

			/* start from black */
			imgf.buf[(row * imgf.width + col) * 3 + 0] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 1] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 2] = 0;

			/* multiple samples */
			for (samp = 0; samp < max(1, nsamples); samp++) {

				/* primary ray for that pixel */
				struct elmCLR ilm = {0,0,0};
				struct elmRAY pray = {0};
				primary_ray(&pray, &cam, row, col, nsamples > 0);

				/* trace ray */
				ilm = trace(&pray, obj, nobj);

				/* assign illumination */
				imgf.buf[(row * imgf.width + col) * 3 + 0] += ilm.r;
				imgf.buf[(row * imgf.width + col) * 3 + 1] += ilm.g;
				imgf.buf[(row * imgf.width + col) * 3 + 2] += ilm.b;
			}

			/* renormalize */
			if (nsamples > 1) {
				imgf.buf[(row * imgf.width + col) * 3 + 0] /= (float) nsamples;
				imgf.buf[(row * imgf.width + col) * 3 + 1] /= (float) nsamples;
				imgf.buf[(row * imgf.width + col) * 3 + 2] /= (float) nsamples;
			}
		}
	}

	/* write to file */
	{
		struct elmIMG img = {imgf.width, imgf.height, 3, 255};
		img.bytes = img.maxval > 255 ? 2 : 1;
		img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
		img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

		for (row = 0; row < imgf.height; row++) {
			for (col = 0; col < imgf.width; col++) {
				img.buffer[row * img.stride + col * img.depth * img.bytes + 0] = FLOAT2UCHAR(imgf.buf[(row * imgf.width + col) * 3 + 0]);
				img.buffer[row * img.stride + col * img.depth * img.bytes + 1] = FLOAT2UCHAR(imgf.buf[(row * imgf.width + col) * 3 + 1]);
				img.buffer[row * img.stride + col * img.depth * img.bytes + 2] = FLOAT2UCHAR(imgf.buf[(row * imgf.width + col) * 3 + 2]);
			}
		}

		/* dump image */
		bmp_write("out.bmp", &img);

		/* free image */
		free(img.buffer);
		memset(&img, 0, sizeof(struct elmIMG));
	}

	/* free float image */
	free(imgf.buf);
	memset(&imgf, 0, sizeof(struct elmIMGF));

	return 0;
}
