/* restart again from scratch */
/* conversion of smallpt using prof. Cline slides */
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "elabmm.h"

/* coordinate/vector/color triplets */
struct elmVEC {
	double x, y, z; /* the three components */
};

/* constructor */
struct elmVEC elmvec(double x, double y, double z)
{
	struct elmVEC c = {x, y, z};
	return c;
}

/* component-wise addition */
struct elmVEC vec_plus(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x + b.x, a.y + b.y, a.z + b.z};
	return c;
}

/* component-wise subtraction */
struct elmVEC vec_minus(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x - b.x, a.y - b.y, a.z - b.z};
	return c;
}

/* component-wise scalar-vector multiplication */
struct elmVEC vec_times(struct elmVEC a, double b)
{
	struct elmVEC c = {b * a.x, b * a.y, b * a.z};
	return c;
}

/* component-wise vector-vector multiplication */
struct elmVEC vec_mult(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x * b.x, a.y * b.y, a.z * b.z};
	return c;
}

/* normalization to 1 */
struct elmVEC vec_norm(struct elmVEC a)
{
	return vec_times(a, 1 / sqrt(a.x * a.x + a.y * a.y + a.z * a.z));
}

/* scalar (dot) product of vectors */
double vec_dot(struct elmVEC a, struct elmVEC b)
{
	return (a.x * b.x + a.y * b.y + a.z * b.z);
}

/* vector (cross) product of vectors */
struct elmVEC vec_cross(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x};
	return c;
}

/* a ray */ 
struct elmRAY {
	struct elmVEC orig; /* origin */
	struct elmVEC dir; /* direction */
};

/* constructor */
struct elmRAY elmray(struct elmVEC orig, struct elmVEC dir)
{
	struct elmRAY r;
	r.orig = orig;
	r.dir = dir;
	return r;
}

/* a sphere */
struct elmSPHERE {
	double rad; /* radius */
	struct elmVEC pos; /* position */
	struct elmVEC em; /* emission */
	struct elmVEC clr; /* color */
	enum {elmDIFF, elmSPEC, elmREFR} refl; /* reflection type: diffuse, specular, refractive */
};

/* intersection between sphere and ray: returns distance or 0 if miss */
double sphere_intersect(struct elmSPHERE *sph, struct elmRAY ray)
{
	/* solve the 2nd degree equation */
	struct elmVEC op = vec_minus(sph->pos, ray.orig);
	double t, eps = 1E-4; /* fudge factor */
	double b = vec_dot(op, ray.dir); /* in quadratic solver */
	double det = b * b - vec_dot(op, op) + sph->rad * sph->rad;

	if (det < 0)
		return 0; /* ray misses sphere */
	else
		det = sqrt(det);

	/* return smaller positive t */
	if ((t = b - det) > eps)
		return t;
	else if ((t = b + det) > eps)
		return t;
	else
		return 0;
}

struct elmIMGF {
	int width, height;
	int bord;
	float *buf;
};


/* Scene: radius, position, emission, color, material */
struct elmSPHERE spheres[] = {
	{1e5, {1e5+1,40.8,81.6}, {0}, {.75,.25,.25}, elmDIFF}, //Left
	{1e5, {-1e5+99,40.8,81.6}, {0}, {.25,.25,.75}, elmDIFF}, //Rght
	{1e5, {50,40.8, 1e5}, {0}, {.75,.75,.75}, elmDIFF},//Back
	{1e5, {50,40.8,-1e5+170}, {0}, {0}, elmDIFF},//Frnt
	{1e5, {50, 1e5, 81.6}, {0}, {.75,.75,.75}, elmDIFF},//Botm
	{1e5, {50,-1e5+81.6,81.6}, {0}, {.75,.75,.75}, elmDIFF},//Top
	{16.5, {27,16.5,47}, {0}, {.999,.999,.999}, elmSPEC},//Mirr
	{16.5, {73,16.5,78}, {0}, {.999,.999,.999}, elmREFR},//Glas
	{600, {50,681.6-.27,81.6}, {12,12,12}, {0}, elmDIFF} //Lite
};

int numSpheres = sizeof(spheres) / sizeof(struct elmSPHERE);

/* clamping function */
_inline double clamp(double x)
{
	return x < 0 ? 0 : (x > 1 ? 1 : x);
}

/* double to integer conversion for saving in file */
_inline int toInt(double x)
{
	return (int) (pow(clamp(x), 1.0 / 2.2) * 255.0 + 0.5);
}

/* intersects ray with scene */
_inline int intersect(struct elmRAY ray, double *t, int *id)
{
  double n = sizeof(spheres) / sizeof(struct elmSPHERE), d, inf = *t = 1e20;
  int i;
  for (i = (int) n; i--;) {
	  if ((d = sphere_intersect(&spheres[i], ray)) && d < *t) {
		  *t = d;
		  *id = i;
	  }
  }

  return *t < inf;
}

/* random number between 0 and 1 */
double erand48(unsigned short xseed[3])
{
	return (double) rand() / (double) RAND_MAX;
}

// radiance along ray
struct elmVEC radiance(struct elmRAY ray, int depth, unsigned short *Xi, int E)
{
	double t; // distance to intersection
	int id = 0; // id of intersected object

	if (!intersect(ray, &t, &id))

		return elmvec(0, 0, 0); // if miss, return black

	else {

		struct elmVEC x = vec_plus(ray.orig, vec_times(ray.dir, t)); /* ray intersection point */
		struct elmSPHERE *obj = &spheres[id]; // the hit object
		struct elmVEC n = vec_norm(vec_minus(x, obj->pos)); // sphere normal
		struct elmVEC nl = vec_dot(n, ray.dir) < 0 ? n : vec_times(n, -1); // properly oriented surface normal
		struct elmVEC f = obj->clr; /* object color */
		double p = f.x > f.y && f.x > f.z ? f.x : f.y > f.z ? f.y : f.z; /* use maximum reflectivity amount for Russian roulette */
				
		// too deep recursion
		if (depth > 10)
			return elmvec(0,0,0);

		// max refl
		if (++depth > 5 || !p)
			if (erand48(Xi) < p)
				f = vec_times(f, 1 / p);
			else
				return vec_times(obj->em, E); //R.R.

		/* depends on object material */
		if (obj->refl == elmDIFF) {

			// Ideal DIFFUSE reflection
			double r1 = 2 * M_PI * erand48(Xi); // angle around
			double r2 = erand48(Xi), r2s = sqrt(r2); // distance from center
			struct elmVEC w = nl; // w =  normal
			struct elmVEC u = vec_norm(vec_cross(fabs(w.x) > .1 ? elmvec(0, 1, 0) : elmvec(1, 0, 0), w)); // u is perpendicular to w
			struct elmVEC v = vec_cross(w, u); // v is perpendicular to v and w
			struct elmVEC d = vec_norm(vec_plus(vec_times(u, cos(r1) * r2s), vec_plus(vec_times(v, sin(r1) * r2s), vec_times(w, sqrt(1 - r2))))); // d is a random reflection ray
			
			// Loop over any lights (explicit lighting)
			struct elmVEC e = {0,0,0};
			int i;
			for (i = 0; i < numSpheres; i++) {

				struct elmSPHERE *s = &spheres[i];
				if (s->em.x <= 0 && s->em.y <= 0 && s->em.z <= 0)

					continue; // skip non-lights 

				else {

					// create random direction towards sphere using method from Realistic Ray Tracing
					struct elmVEC sw = vec_minus(s->pos, x),
						su = vec_norm(vec_cross(fabs(sw.x) > .1 ? elmvec(0,1,0) : elmvec(1,0,0), sw)),
						sv = vec_cross(sw, su);
					double cos_a_max = sqrt(1 - s->rad * s->rad / vec_dot(vec_minus(x, s->pos), vec_minus(x, s->pos)));
					double eps1 = erand48(Xi), eps2 = erand48(Xi);
					double cos_a = 1 - eps1 + eps1 * cos_a_max;
					double sin_a = sqrt(1 - cos_a * cos_a);
					double phi = 2 * M_PI * eps2;
					struct elmVEC l = vec_plus(vec_plus(vec_times(su, cos(phi) * sin_a), vec_times(sv, sin(phi) * sin_a)), vec_times(sw, cos_a));
					l = vec_norm(l);

					// shoot shadow ray
					if (intersect(elmray(x, l), &t, &id) && id == i) {  // shadow ray
						double omega = 2 * M_PI * (1 - cos_a_max);
						//e = vec_plus(e, vec_times(vec_mult(f, vec_times(s->em, vec_dot(l, nl) * omega)), M_1_PI));  // 1/pi for brdf
					}
				}
			}

			// recursive call with random ray and no emissive term
			return vec_plus(e, vec_plus(vec_times(obj->em, E), vec_mult(f, radiance(elmray(x, d), depth, Xi, 0))));

		} else if (obj->refl == elmSPEC) {

			// Ideal SPECULAR reflection
			return vec_plus(obj->em, vec_mult(f, radiance(elmray(x, vec_minus(ray.dir, vec_times(n, 2 * vec_dot(n, ray.dir)))), depth, Xi, 1)));
		
		} else {

			/* REFRACTION on dielectric surface - glass */
			struct elmRAY reflRay = elmray(x, vec_minus(ray.dir, vec_times(n, 2 * vec_dot(n, ray.dir))));     
			int into = vec_dot(n, nl) > 0; /* Ray from outside going in? */
			double nc = 1, nt = 1.5, nnt = into ? nc / nt : nt / nc, ddn = vec_dot(ray.dir, nl), cos2t;
			
			/* if total internal reflection, reflect */
			if ((cos2t = 1 - nnt * nnt * (1 - ddn * ddn)) < 0) {

				/* Total internal reflection */
				return vec_plus(obj->em, vec_mult(f, radiance(reflRay, depth, Xi, 1)));

			} else {

				/* choose reflection or refraction */
				struct elmVEC tdir = vec_norm(vec_minus(vec_times(ray.dir, nnt), vec_times(n, (into ? 1 : -1) * (ddn * nnt + sqrt(cos2t)))));
				double a = nt - nc, b = nt + nc, R0 = a * a / (b * b), c = 1 - (into ? -ddn : vec_dot(tdir, n));
				double Re = R0 + (1 - R0) * c * c * c * c * c, Tr = 1 - Re, P = .25 + .5 * Re,
					RP = Re / P, TP = Tr / (1 - P);
				
				// Russian roulette
				return vec_plus(obj->em, vec_mult(f, depth > 2 ? (erand48(Xi) < P ? 
					vec_times(radiance(reflRay, depth, Xi, 1), RP) : vec_times(radiance(elmray(x, tdir), depth, Xi, 1), TP)) :
					vec_plus(vec_times(radiance(reflRay, depth, Xi, 1), Re), vec_times(radiance(elmray(x, tdir), depth, Xi, 1), Tr))));
			}
		}
	}
}

/* exposure correction function */
float apply_exposure(float l)
{
	float expo = -1.00f;
	return (1.0f - expf(l * expo));
}

/* encoding to srgb */
float srgb_encode(float c)
{
	if (c <= 0.0031308f) 
		return 12.92f * c;
	else
		return 1.055f * powf(c, 0.4166667f) - 0.055f; // Inverse gamma 2.4
}

/* clip float to unsigned char */
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))

/* encode a pixel */
unsigned char encode_pixel(float pixval)
{
	// exposure correction
	float pixvalec = apply_exposure(pixval);

	// gamma correction
	float pixvalge = srgb_encode(pixvalec);

	// ready for file
	unsigned char pixvaluc = FLOAT2UCHAR(pixvalge);

	return pixvaluc;
}

/* main function */
int main(int argc, char **argv)
{
	int w = 1024, h = 768; // image size
	struct elmIMGF imgf = {w, h, 8};
	int samps = argc == 2 ? atoi(argv[1]) / 4 : 32; // # samples, default 1
	struct elmRAY cam = elmray(elmvec(50,52,295.6), vec_norm(elmvec(0,-0.042612,-1))); // camera pos, dir
	struct elmVEC cx = elmvec(w * .5135 / h, 0, 0); // x dir increment
	struct elmVEC cy = vec_times(vec_norm(vec_cross(cx, cam.dir)), .5135); // y direction increment with cross product
	struct elmVEC r; // colors of smples
	int y, x, s;

	/* prepare image */
	imgf.buf = (float *) calloc(imgf.height * imgf.width * 3, sizeof(float));

	for (s = 0; s < samps; s++) { // loop samples

		for (y = 0; y < h; y++) { // Loop over image rows

			unsigned short Xi[3] = {0, 0, y * y * y};
			int sx, sy;

			fprintf(stderr, "\rRendering (%d spp) %5.2f%%", samps * 4, 100.0 * (s * h + y + 1) / (samps * h));

			for (x = 0; x < w; x++) {  // Loop cols

				for (sy = 0; sy < 2; sy++) { // 2x2 subpixel rows

					for (sx = 0; sx < 2; sx++, r = elmvec(0, 0, 0)) { // 2x2 subpixel cols

						/* tent filter? */
						double r1 = 2 * erand48(Xi), dx = r1 < 1 ? sqrt(r1) - 1 : 1 - sqrt(2 - r1);
						double r2 = 2 * erand48(Xi), dy = r2 < 1 ? sqrt(r2) - 1 : 1 - sqrt(2 - r2);
						struct elmVEC d = vec_plus(vec_plus(vec_times(cx, (((sx + .5 + dx) / 2 + x) / w - .5)),
							vec_times(cy, (((sy + .5 + dy) / 2 + y) / h - .5))), cam.dir);
						
						// Camera rays are pushed ^^^^^ forward to start in interior
						r = radiance(elmray(vec_plus(cam.orig, vec_times(d, 140)), vec_norm(d)), 0, Xi, 1); 
						
						// accumulate radiance on this point
						imgf.buf[(y * imgf.width + x) * 3 + 0] += (float) r.x;
						imgf.buf[(y * imgf.width + x) * 3 + 1] += (float) r.y;
						imgf.buf[(y * imgf.width + x) * 3 + 2] += (float) r.z;
					}
				}
			}
		}

		/* write to file */
		if (1) {

			struct elmIMGF imgf2 = imgf;
			struct elmIMG img = {imgf.width, imgf.height, 3, 255};
			img.bytes = img.maxval > 255 ? 2 : 1;
			img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
			img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

			imgf2.buf = (float *) calloc(imgf2.height * imgf2.width * 3, sizeof(float));
			memcpy(imgf2.buf, imgf.buf, imgf2.height * imgf2.width * 3 * sizeof(float));
			//post_process(&imgf2, &imgf);

			for (y = 0; y < imgf.height; y++) {
				for (x = 0; x < imgf.width; x++) {
					img.buffer[y * img.stride + x * img.depth * img.bytes + 0] =
						encode_pixel(imgf2.buf[((h - 1 - y) * imgf2.width + x) * 3 + 0] / (samps * 4));
					img.buffer[y * img.stride + x * img.depth * img.bytes + 1] =
						encode_pixel(imgf2.buf[((h - 1 - y) * imgf2.width + x) * 3 + 1] / (samps * 4));
					img.buffer[y * img.stride + x * img.depth * img.bytes + 2] =
						encode_pixel(imgf2.buf[((h - 1 - y) * imgf2.width + x) * 3 + 2] / (samps * 4));
				}
			}

			free(imgf2.buf);
		
			/* dump image */
			bmp_write("image.bmp", &img);

			/* free image */
			free(img.buffer);
			memset(&img, 0, sizeof(struct elmIMG));
		}
	} 


	/* free image */
	free(imgf.buf);

	return EXIT_SUCCESS;
}
