/* restart again from scratch */
/* conversion of smallpt using prof. Cline slides */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <conio.h>
#include <time.h>
#define PNG_DEBUG 3
#include <png.h>
#include "elabmm.h"

extern const double Inf = ((double)(1e308 * 10));

#define DOUBLE_SWAP(A, B) {double temp = (A); (A) = (B); (B) = temp;}

/* coordinate/vector/color triplets */
struct elmVEC {
	double x, y, z; /* the three components */
};

/* constructor */
struct elmVEC elmvec(double x, double y, double z)
{
	struct elmVEC c = {x, y, z};
	return c;
}

/* component-wise addition */
struct elmVEC vec_plus(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x + b.x, a.y + b.y, a.z + b.z};
	return c;
}

/* component-wise subtraction */
struct elmVEC vec_minus(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x - b.x, a.y - b.y, a.z - b.z};
	return c;
}

/* component-wise scalar-vector multiplication */
struct elmVEC vec_times(struct elmVEC a, double b)
{
	struct elmVEC c = {b * a.x, b * a.y, b * a.z};
	return c;
}

/* component-wise vector-vector multiplication */
struct elmVEC vec_mult(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x * b.x, a.y * b.y, a.z * b.z};
	return c;
}

/* normalization to 1 */
struct elmVEC vec_norm(struct elmVEC a)
{
	return vec_times(a, 1 / sqrt(a.x * a.x + a.y * a.y + a.z * a.z));
}

/* scalar (dot) product of vectors */
double vec_dot(struct elmVEC a, struct elmVEC b)
{
	return (a.x * b.x + a.y * b.y + a.z * b.z);
}

/* vector (cross) product of vectors */
struct elmVEC vec_cross(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x};
	return c;
}

/* maximum of vector */
double vec_max(struct elmVEC f)
{
	return f.x > f.y && f.x > f.z ? f.x : f.y > f.z ? f.y : f.z;
}

/* a ray */ 
struct elmRAY {
	struct elmVEC orig; /* origin */
	struct elmVEC dir; /* direction */
	int objid; /* ray origin object id: -1 means camera */
};

/* constructor */
struct elmRAY elmray(struct elmVEC orig, struct elmVEC dir, int objid)
{
	struct elmRAY r;
	r.orig = orig;
	r.dir = dir;
	r.objid = objid;
	return r;
}

/* a generic object */
struct elmOBJ {
	enum {ELMPLANE, ELMSPHERE, ELMDISC, ELMCYLINDER, ELMCONE, ELMTRIANGLE} kind;
	/* plane, disc, cylinder, cone */
	struct {
		struct elmVEC nor; /* normal or base normal */
	};
	/* plane, sphere, disc, cylinder, cone, triangle */
	struct {
		struct elmVEC pos; /* center, base center, or first vertex*/
	};
	/* sphere, disc, cylinder, cone */
	struct {
		double rad; /* radius or base radius */
	};
	/* cylinder, cone */
	struct {
		double hght; /* height */
	};
	/* cylinder, cone, triangle */
	struct {
		struct elmVEC pos2; /* top center or second vertex */
	};
	/* cone */
	struct {
		double rad2; /* top radius */
		double tga; /* tangent of half angle */
		struct elmVEC apx; /* apex */
	};
	/* triangle */
	struct {
		struct elmVEC pos3; /* third vertex */
	};
	struct elmVEC em; /* emission */
	struct elmVEC clr; /* color */
	enum {elmDIFF, elmSPEC, elmREFR, elmGLOSS} refl; /* reflection type: diffuse, specular, refractive */
	char name[32]; /* name */
	struct {
		struct elmVEC vmin, vmax;
	} aabb; /* axis aligned bounding box */
	int id; /* object identifier */
};

/* the camera */
struct elmCAM {
	struct elmRAY view; // camera pos, dir
	double hfov; // horizontal field of view, in degrees
	struct elmVEC cx; // x dir increment
	struct elmVEC cy; // y direction increment with cross product
	struct elmVEC ipo; // image plane origin
};

/* a floating point image */
struct elmIMGF {
	int width, height;
	float *buf;
	int *samps;
};

/* the scene */
struct elmSCENE {
	int nobjs; /* number of objects */
	struct elmOBJ *obj; /* pointer to objects */
	struct elmCAM cam; /* camera */
	struct {
		struct elmVEC clr; /* ambient color */
		struct elmIMGF map; /* cubemap image */
		double str; /* ambient strength */
	} amb; /* ambient */
};

/* intersection between sphere and ray: returns distance or 0 if miss */
double intersect_sphere(struct elmOBJ *sph, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* solve the 2nd degree equation */
	struct elmVEC op = vec_minus(sph->pos, ray->orig);
	double t, eps = 1E-4; /* fudge factor */
	double b = vec_dot(op, ray->dir); /* in quadratic solver */
	double det = b * b - vec_dot(op, op) + sph->rad * sph->rad;

	if (det < 0)
		return 0; /* ray misses sphere */
	else
		det = sqrt(det);

	/* return smaller positive t */
	if ((t = b - det) > eps) {
		*hitp = vec_plus(ray->orig, vec_times(ray->dir, t)); /* ray intersection point */
		*hitn = vec_norm(vec_minus(*hitp, sph->pos)); // sphere normal
		return t;
	} else if ((t = b + det) > eps) {
		*hitp = vec_plus(ray->orig, vec_times(ray->dir, t)); /* ray intersection point */
		*hitn = vec_norm(vec_minus(*hitp, sph->pos)); // sphere normal
		return t;
	} else
		return 0;
}

/* intersection between plane and ray: returns distance or 0 if miss */
double intersect_plane(struct elmOBJ *pla, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* self intersection */
	if (ray->objid == pla->id)
		return 0;
	else {
		/* search */
		double denom = vec_dot(pla->nor, ray->dir);
		if (denom < -1e-12 || denom > 1e-12) {
			/* not parallel */
			double t = vec_dot(vec_minus(pla->pos, ray->orig), pla->nor) / denom; 
			if (t > 0) {
				/* in front */
				*hitn = pla->nor;
				*hitp = vec_plus(ray->orig, vec_times(ray->dir, t));
				return t;
			} else
				/* behind */
				return 0;
		} else
			/* parallel */
			return 0;
	}
}

/* intersection between disc and ray: returns distance or 0 if miss */
double intersect_disc(struct elmOBJ *dsc, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* self intersection */
	if (ray->objid == dsc->id)
		return 0;
	else {
		/* search */
		double denom = vec_dot(dsc->nor, ray->dir);
		if (denom < -1e-12 || denom > 1e-12) {
			/* not parallel */
			double t = vec_dot(vec_minus(dsc->pos, ray->orig), dsc->nor) / denom; 
			if (t > 0) {
				/* in front */
				struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t));
				struct elmVEC v = vec_minus(p, dsc->pos);
				if (vec_dot(v, v) <= dsc->rad * dsc->rad) {
					/* in disc */
					*hitn = dsc->nor;
					*hitp = p;
					return t;
				} else
					return 0;
			} else
				/* behind */
				return 0;
		} else
			/* parallel */
			return 0;
	}
}

/* intersection between disc and cylinder: returns distance or 0 if miss */
double intersect_cylinder(struct elmOBJ *cyl, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	struct elmVEC v1 = vec_times(cyl->nor, vec_dot(ray->dir, cyl->nor));
	struct elmVEC v2 = vec_minus(ray->dir, v1);
	double A = vec_dot(v2, v2);
	struct elmVEC v3 = vec_minus(ray->orig, cyl->pos);
	struct elmVEC v4 = vec_times(cyl->nor, vec_dot(v3, cyl->nor));
	struct elmVEC v5 = vec_minus(v3, v4);
	double B = 2 * vec_dot(v2, v5);
	double C = vec_dot(v5, v5) - cyl->rad * cyl->rad;
	double det = B * B - 4 * A * C;
	double t1, t2, t3, t4, eps = 1e-4;
	double denom;

	/* ray misses cylinder */
	if (det < 0)
		return 0; 
	else
		det = sqrt(det);

	/* first possible intersection point - front side of tube */
	t1 = (-B - det) / (2 * A);
	if (t1 > eps) {
		/* in front of us */
		struct elmVEC q1 = vec_plus(ray->orig, vec_times(ray->dir, t1));
		if (vec_dot(cyl->nor, vec_minus(q1, cyl->pos)) > 0 && vec_dot(cyl->nor, vec_minus(q1, cyl->pos2)) < 0) {
			/* is in the external tube */
			*hitp = q1; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q1, vec_times(cyl->nor, vec_dot(cyl->nor, vec_minus(q1, cyl->pos))))) ; // normal
			return t1;
		}
	}	

	/* try caps */
	denom = vec_dot(cyl->nor, ray->dir);
	if (denom < -1e-12 || denom > 1e-12) {
		/* not parallel */
		/* bottom cap */
		t3 = vec_dot(vec_minus(cyl->pos, ray->orig), cyl->nor) / denom; 
		if (t3 > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t3));
			struct elmVEC v = vec_minus(p, cyl->pos);
			if (vec_dot(v, v) <= cyl->rad * cyl->rad) {
				/* in disc */
				*hitn = vec_times(cyl->nor, -1);
				*hitp = p;
				return t3;
			} 
		} 
		/* top cap */
		t4 = vec_dot(vec_minus(cyl->pos2, ray->orig), cyl->nor) / denom; 
		if (t4 > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t4));
			struct elmVEC v = vec_minus(p, cyl->pos2);
			if (vec_dot(v, v) <= cyl->rad * cyl->rad) {
				/* in disc */
				*hitn = cyl->nor;
				*hitp = p;
				return t4;
			} 
		} 
	}

	/* last possible intersection point - back side of tube */
	t2 = (-B + det) / (2 * A);
	if (t2 > eps) {
		/* in front of us */
		struct elmVEC q2 = vec_plus(ray->orig, vec_times(ray->dir, t2));
		if (vec_dot(cyl->nor, vec_minus(q2, cyl->pos)) > 0 && vec_dot(cyl->nor, vec_minus(q2, cyl->pos2)) < 0) {
			/* is in the external tube */
			*hitp = q2; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q2, vec_times(cyl->nor, vec_dot(cyl->nor, vec_minus(q2, cyl->pos))))) ; // normal
			return t2;
		}
	}	

	return 0;
}

/* intersection between ray and cone: returns distance or 0 if miss */
double intersect_cone(struct elmOBJ *con, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	double dirdotnor = vec_dot(ray->dir, con->nor);
	struct elmVEC v1 = vec_minus(ray->dir, vec_times(con->nor, dirdotnor));
	double cos2a = 1 / (1 + con->tga * con->tga);
	double sin2a = cos2a * con->tga * con->tga;
	double A = cos2a * vec_dot(v1, v1) - sin2a * dirdotnor * dirdotnor;
	struct elmVEC v2 = vec_minus(ray->orig, con->apx);
	double v2dotnor = vec_dot(v2, con->nor);
	struct elmVEC v3 = vec_minus(v2, vec_times(con->nor, v2dotnor));
	double B = 2 * cos2a * vec_dot(v1, v3) - 2 * sin2a * dirdotnor * v2dotnor;
	double C = cos2a * vec_dot(v3, v3) - sin2a * v2dotnor * v2dotnor;
	double det = B * B - 4 * A * C;
	double t1, t2, t3, t4, eps = 1e-4;
	double denom;

	/* ray misses cone */
	if (det < 0)
		return 0; 
	else
		det = sqrt(det);

	/* first possible intersection point - front side of cone */
	t1 = (-B - det) / (2 * A);
	if (t1 > eps) {
		/* in front of us */
		struct elmVEC q1 = vec_plus(ray->orig, vec_times(ray->dir, t1));
		if (vec_dot(con->nor, vec_minus(q1, con->pos)) > 0 && vec_dot(con->nor, vec_minus(q1, con->pos2)) < 0) {
			/* is in the external tube */
			struct elmVEC x = vec_times(con->nor, vec_dot(vec_minus(q1, con->pos), con->nor));
			struct elmVEC O = vec_times(con->nor, sqrt(vec_dot(x, x)) * (1 + con->tga * con->tga) - con->tga * con->rad);
			*hitp = q1; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q1, O));
			return t1;
		}
	}	

	/* try caps */
	denom = vec_dot(con->nor, ray->dir);
	if (denom < -1e-12 || denom > 1e-12) {
		/* not parallel */
		/* bottom cap */
		t3 = vec_dot(vec_minus(con->pos, ray->orig), con->nor) / denom; 
		if (t3 > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t3));
			struct elmVEC v = vec_minus(p, con->pos);
			if (vec_dot(v, v) <= con->rad * con->rad) {
				/* in disc */
				*hitn = vec_times(con->nor, -1);
				*hitp = p;
				return t3;
			} 
		} 
		/* top cap */
		if (con->rad2) {
			t4 = vec_dot(vec_minus(con->pos2, ray->orig), con->nor) / denom; 
			if (t4 > 0) {
				/* in front */
				struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t4));
				struct elmVEC v = vec_minus(p, con->pos2);
				if (vec_dot(v, v) <= con->rad2 * con->rad2) {
					/* in disc */
					*hitn = con->nor;
					*hitp = p;
					return t4;
				} 
			}
		}
	}

	/* last possible intersection point - back side of tube */
	t2 = (-B + det) / (2 * A);
	if (t2 > eps) {
		/* in front of us */
		struct elmVEC q2 = vec_plus(ray->orig, vec_times(ray->dir, t2));
		if (vec_dot(con->nor, vec_minus(q2, con->pos)) > 0 && vec_dot(con->nor, vec_minus(q2, con->pos2)) < 0) {
			/* is in the external tube */
			struct elmVEC x = vec_times(con->nor, vec_dot(vec_minus(q2, con->pos), con->nor));
			struct elmVEC O = vec_times(con->nor, sqrt(vec_dot(x, x)) * (1 + con->tga * con->tga) - con->tga * con->rad);
			*hitp = q2; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q2, O));
			return t2;
		}
	}

	return 0;
}

/* intersection between ray and triangle: returns distance or 0 if miss */
double intersect_triangle(struct elmOBJ *tri, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* self intersection */
	if (ray->objid == tri->id)
		return 0;
	else {
		double nordotdir = vec_dot(tri->nor, ray->dir);
		if (nordotdir < -1e-12 || nordotdir > 1e-12) {
			// not parallel
			double t = vec_dot(vec_minus(tri->pos, ray->orig), tri->nor) / nordotdir; 
			if (t < 0)
				// behind
				return 0;
			else {
				struct elmVEC P = vec_plus(ray->orig, vec_times(ray->dir, t));
				struct elmVEC edge0 = vec_minus(tri->pos2, tri->pos);
				struct elmVEC VP0 = vec_minus(P, tri->pos);
				struct elmVEC C = vec_cross(edge0, VP0);
				if (vec_dot(tri->nor, C) < 0)
					return 0;
				else {
					struct elmVEC edge1 = vec_minus(tri->pos3, tri->pos2);
					struct elmVEC VP1 = vec_minus(P, tri->pos2);
					C = vec_cross(edge1, VP1);
					if (vec_dot(tri->nor, C) < 0)
						return 0;
					else {
						struct elmVEC edge2 = vec_minus(tri->pos, tri->pos3);
						struct elmVEC VP2 = vec_minus(P, tri->pos3);
						C = vec_cross(edge2, VP2);
						if (vec_dot(tri->nor, C) < 0)
							return 0;
						else {
							*hitp = P;
							*hitn = tri->nor;
							return t;
						}
					}
				}
			}
		} else
			return 0;
	}
}

/* intersect aabb */
int intersect_aabb(struct elmRAY *ray, struct elmVEC vmin, struct elmVEC vmax)
{
	double tmin, tymin, tzmin;
	double tmax, tymax, tzmax;
	tmin = (vmin.x - ray->orig.x) / ray->dir.x;
	tmax = (vmax.x - ray->orig.x) / ray->dir.x;
	if (tmin > tmax)
		DOUBLE_SWAP(tmin, tmax);
	tymin = (vmin.y - ray->orig.y) / ray->dir.y;
	tymax = (vmax.y - ray->orig.y) / ray->dir.y;
	if (tymin > tymax)
		DOUBLE_SWAP(tymin, tymax);
	if ((tmin > tymax) || (tymin > tmax))
		return 0;
	if (tymin > tmin)
		tmin = tymin;
	if (tymax < tmax)
		tmax = tymax;
	tzmin = (vmin.z - ray->orig.z) / ray->dir.z;
	tzmax = (vmax.z - ray->orig.z) / ray->dir.z;
	if (tzmin > tzmax)
		DOUBLE_SWAP(tzmin, tzmax);
	if ((tmin > tzmax) || (tzmin > tmax))
		return 0;
	if (tzmin > tmin)
		tmin = tzmin;
	if (tzmax < tmax)
		tmax = tzmax;
	/*if ((tmin > r.tmax) || (tmax < r.tmin)) return false;
	if (r.tmin < tmin) r.tmin = tmin;
	if (r.tmax > tmax) r.tmax = tmax;*/
	return 1;
}

/* intersect object: return distance or 0 if miss */
double intersect_obj(struct elmOBJ *obj, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* fast intersect aabb */
	if (1 /*intersect_aabb(ray, obj->aabb.vmin, obj->aabb.vmax)*/) {

		/* depending on type */
		if (obj->kind == ELMSPHERE)
			return intersect_sphere(obj, ray, hitp, hitn);
		else if (obj->kind == ELMPLANE)
			return intersect_plane(obj, ray, hitp, hitn);
		else if (obj->kind == ELMDISC)
			return intersect_disc(obj, ray, hitp, hitn);
		else if (obj->kind == ELMCYLINDER)
			return intersect_cylinder(obj, ray, hitp, hitn);
		else if (obj->kind == ELMCONE)
			return intersect_cone(obj, ray, hitp, hitn);
		else if (obj->kind == ELMTRIANGLE)
			return intersect_triangle(obj, ray, hitp, hitn);
		else
			return 0;

	} else

		return 0;
}

/* intersects ray with scene */
int intersect_scene(struct elmRAY *ray, struct elmSCENE *sce, double *t, int *id, struct elmVEC *hitp, struct elmVEC *hitn)
{
	double d = 0;
	int i = 0;
	struct elmVEC bhitp = {0}, bhitn = {0};

	/* explore all objects */
	*t = Inf;
	for (i = 0; i < sce->nobjs; i++) {
		/* try to intersect the object */
		if ((d = intersect_obj(&sce->obj[i], ray, &bhitp, &bhitn)) && d < *t) {
			*t = d;
			*id = i;
			*hitp = bhitp;
			*hitn = bhitn;
		}
	}

	/* return success or not */
	return *t < Inf;
}

int png_read(char *fname, struct elmIMG *img)
{
    unsigned char header[8];    // 8 is the maximum size that can be checked
	png_structp png_ptr;
	png_infop info_ptr;
	png_byte color_type;
	png_byte bit_depth;
	int number_of_passes;
	png_bytep *row_pointers;
	int y;

	/* open file and test for it being a png */
	FILE *fp = fopen(fname, "rb");
	if (!fp)
		return -1;
	fread(header, 1, 8, fp);
	if (png_sig_cmp(header, 0, 8))
		return -1;

    /* initialize stuff */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
		return -1;

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
		return -1;

    if (setjmp(png_jmpbuf(png_ptr)))
		return -1;

    /* Expand any grayscale, RGB, or palette images to RGBA */
    png_set_expand(png_ptr);

    /* Reduce any 16-bits-per-sample images to 8-bits-per-sample */
    png_set_strip_16(png_ptr);

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

	memset(img, 0, sizeof(struct elmIMG));

    img->width = png_get_image_width(png_ptr, info_ptr);
    img->height = png_get_image_height(png_ptr, info_ptr);
    color_type = png_get_color_type(png_ptr, info_ptr);
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);

    number_of_passes = png_set_interlace_handling(png_ptr);
    png_read_update_info(png_ptr, info_ptr);

	img->depth = 4;
	img->maxval = 255;
	img->bytes = img->maxval > 255 ? 2 : 1;
	img->stride = 16 * ((img->width * img->depth * img->bytes + 15) / 16);
	img->buffer = (unsigned char *) calloc(img->height * img->stride, 1);
	if (!img->buffer)
		return -1;

    /* read file */
    row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * img->height);
    for (y = 0; y < img->height; y++)
		row_pointers[y] = img->buffer + y * img->stride;

    png_read_image(png_ptr, row_pointers);

	free(row_pointers);
    fclose(fp);
}

// *Really* minimal PCG32 code / (c) 2014 M.E. O'Neill / pcg-random.org
// Licensed under Apache License 2.0 (NO WARRANTY, etc. see website)
typedef struct { unsigned __int64 state;  unsigned __int64 inc; } pcg32_random_t;
unsigned long int pcg32_random_r(pcg32_random_t* rng)
{
    unsigned __int64 oldstate = rng->state;
	unsigned long int xorshifted;
	long int rot;

    // Advance internal state
    rng->state = oldstate * 6364136223846793005ULL + (rng->inc|1);

    // Calculate output function (XSH RR), uses old state for max ILP
    xorshifted = (unsigned long int) (((oldstate >> 18u) ^ oldstate) >> 27u);
    rot = oldstate >> 59u;

    return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
}

/* random number between 0 and 1 */
double myrand(pcg32_random_t *rng)
{
	return (double) pcg32_random_r(rng) / (double) UINT_MAX;
}

/* get a texture pixel with bilinear interpolation */
struct elmVEC map_texture(struct elmIMGF *imgf, double x, double y)
{
	struct elmVEC ilmout = {0,0,0};
	double u, v;
	int col, row;
	struct elmVEC P0, P1, P2, P3;
	if (x < 0)
		x = 0;
	else if (x > imgf->width - 2)
		x = imgf->width - 2;
	if (y < 1)
		y = 1;
	else if (y > imgf->height - 1)
		y = imgf->height - 1;
	u = x - floorf(x);
	v = 1 - (y - floorf(y));
	col = (int) x;
	row = imgf->height - 1 - (int) y;
	P0.x = imgf->buf[(row * imgf->width + col) * 3 + 0];
	P0.y = imgf->buf[(row * imgf->width + col) * 3 + 1];
	P0.z = imgf->buf[(row * imgf->width + col) * 3 + 2];
	P1.x = imgf->buf[(row * imgf->width + col + 1) * 3 + 0];
	P1.y = imgf->buf[(row * imgf->width + col + 1) * 3 + 1];
	P1.z = imgf->buf[(row * imgf->width + col + 1) * 3 + 2];
	P2.x = imgf->buf[((row + 1) * imgf->width + col) * 3 + 0];
	P2.y = imgf->buf[((row + 1) * imgf->width + col) * 3 + 1];
	P2.z = imgf->buf[((row + 1) * imgf->width + col) * 3 + 2];
	P3.x = imgf->buf[((row + 1) * imgf->width + col + 1) * 3 + 0];
	P3.y = imgf->buf[((row + 1) * imgf->width + col + 1) * 3 + 1];
	P3.z = imgf->buf[((row + 1) * imgf->width + col + 1) * 3 + 2];
	ilmout.x = ((1 - u) * P0.x + u * P1.x) * (1 - v) + ((1 - u) * P2.x + u * P3.x) * v;
	ilmout.y = ((1 - u) * P0.y + u * P1.y) * (1 - v) + ((1 - u) * P2.y + u * P3.y) * v;
	ilmout.z = ((1 - u) * P0.z + u * P1.z) * (1 - v) + ((1 - u) * P2.z + u * P3.z) * v;
	return ilmout;
}

/* get the ambient occlusion light */
struct elmVEC ambient(struct elmSCENE *sce, struct elmRAY *ray)
{
	/* map on the cube */
	if (sce->amb.map.width) {

		double cx2 = sce->amb.map.width / 8.0, cy2 = sce->amb.map.height / 6.0F;
		double xm, ym;

		if (fabs(ray->dir.x) > fabs(ray->dir.y) && fabs(ray->dir.x) > fabs(ray->dir.z)) {
			if (ray->dir.x > 0) {
				/* facing front */
				double xo = 0.375 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/*ilmout.r = 0.5F; ilmout.g = 0; ilmout.b = 0;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.x;
				ym = yo + cy2 * ray->dir.z / ray->dir.x;
			} else {
				double xo = 0.875 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/* facing back */
				/*ilmout.r = 0.5F; ilmout.g = 0.5F; ilmout.b = 0;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.x;
				ym = yo - cy2 * ray->dir.z / ray->dir.x;
			}
		} else if (fabs(ray->dir.y) > fabs(ray->dir.x) && fabs(ray->dir.y) > fabs(ray->dir.z)) {
			if (ray->dir.y > 0) {
				double xo = 0.125 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/* facing left */
				/*ilmout.r = 0; ilmout.g = 0.5F; ilmout.b = 0;*/
				xm = xo + cx2 * ray->dir.x / ray->dir.y;
				ym = yo + cy2 * ray->dir.z / ray->dir.y;
			} else {
				/* facing right */
				double xo = 0.625 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/*ilmout.r = 0; ilmout.g = 0.5F; ilmout.b = 0.5F;*/
				xm = xo + cx2 * ray->dir.x / ray->dir.y;
				ym = yo - cy2 * ray->dir.z / ray->dir.y;
			}
		} else if (fabs(ray->dir.z) > fabs(ray->dir.x) && fabs(ray->dir.z) > fabs(ray->dir.y)) {
			if (ray->dir.z > 0) {
				double xo = 0.375 * sce->amb.map.width, yo = 0.83333333 * sce->amb.map.height;
				/* facing top */
				/*ilmout.r = 0; ilmout.g = 0; ilmout.b = 0.5F;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.z;
				ym = yo - cy2 * ray->dir.x / ray->dir.z;
			} else {
				double xo = 0.375 * sce->amb.map.width, yo = 0.16666666 * sce->amb.map.height;
				/* facing bottom */
				/*ilmout.r = 0.5F; ilmout.g = 0; ilmout.b = 0.5F;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.z;
				ym = yo + cy2 * ray->dir.x / ray->dir.z;
			}
		} 

		return map_texture(&sce->amb.map, xm, ym);

	} else {

		/* only ambient color */
		return sce->amb.clr;
	}
}

// radiance along ray
struct elmVEC radiance(struct elmRAY ray, struct elmSCENE *sce, int depth, pcg32_random_t* rng)
{
	double t; // distance to intersection
	int id = 0; // id of intersected object
	struct elmVEC n = {0}, x = {0}; // object normal and hit point

	if (!intersect_scene(&ray, sce, &t, &id, &x, &n))

		// if miss, return ambient
		return depth ? vec_times(ambient(sce, &ray), sce->amb.str) : ambient(sce, &ray); 

	else {

		struct elmOBJ *obj = &sce->obj[id]; // the hit object
		double n_dot_raydir = vec_dot(n, ray.dir);
		struct elmVEC nl = n_dot_raydir < 0 ? n : vec_times(n, -1); // properly oriented surface normal
		struct elmVEC f = obj->clr; /* object color */
		double p = vec_max(f); /* use maximum reflectivity amount for Russian roulette */
		struct elmVEC objem = n_dot_raydir < 0 ? obj->em : elmvec(0, 0, 0); // emits only in the normal direction
				
		// too deep recursion
		if (depth > 15)
			// return current emission
			return objem; 

		// max refl with Russian roulette
		if (++depth > 5)
			if (myrand(rng) < p)
				f = vec_times(f, 1 / p);
			else
				return objem;

		/* depends on object material */
		if (obj->refl == elmDIFF) {

			// Ideal DIFFUSE reflection: generate random point on the unit hemisphere

			// random azimuth angle
			double r1 = 2 * M_PI * myrand(rng);

			// random distance from center
			double r2 = myrand(rng), r2s = sqrt(r2);

			// the z axis is the normal
			struct elmVEC w = nl;

			// the u axis is perpendicular to w and is normalized
			struct elmVEC u = vec_norm(vec_cross(fabs(w.x) > .1 ? elmvec(0, 1, 0) : elmvec(1, 0, 0), w)); 

			// v is perpendicular to v and w and is normalized
			struct elmVEC v = vec_cross(w, u); 

			// move along u then along v then along W (hit the unit hemisphere)
			// and obtain a random point rdir on it
			struct elmVEC rdir = vec_norm(vec_plus(vec_times(u, cos(r1) * r2s), vec_plus(vec_times(v, sin(r1) * r2s), vec_times(w, sqrt(1 - r2))))); // d is a random reflection ray
			
			// find radiance on this random direction, weight by surface color, and add emissivity
			return vec_plus(objem, vec_mult(f, radiance(elmray(x, rdir, id), sce, depth, rng)));

		} else if (obj->refl == elmSPEC) {

			// Ideal SPECULAR reflection

			// reflected ray
			struct elmRAY refray = elmray(x, vec_minus(ray.dir, vec_times(n, 2 * n_dot_raydir)), id);

			// return emissivity plus reflected radiance weighted by specular color
			return vec_plus(objem, vec_mult(f, radiance(refray, sce, depth, rng)));
				
		} else  if (obj->refl == elmREFR) {

			/* REFRACTION on dielectric surface - glass */
			struct elmRAY reflRay = elmray(x, vec_minus(ray.dir, vec_times(n, 2 * n_dot_raydir)), id);     
			int into = vec_dot(n, nl) > 0; /* Ray from outside going in? */
			double nc = 1, nt = 1.5, nnt = into ? nc / nt : nt / nc, ddn = vec_dot(ray.dir, nl), cos2t;
			
			/* if total internal reflection, reflect */
			if ((cos2t = 1 - nnt * nnt * (1 - ddn * ddn)) < 0) {

				/* Total internal reflection */
				return vec_plus(objem, vec_mult(f, radiance(reflRay, sce, depth, rng)));

			} else {

				/* choose reflection or refraction */
				struct elmVEC tdir = vec_norm(vec_minus(vec_times(ray.dir, nnt), vec_times(n, (into ? 1 : -1) * (ddn * nnt + sqrt(cos2t)))));
				double a = nt - nc, b = nt + nc, R0 = a * a / (b * b), c = 1 - (into ? -ddn : vec_dot(tdir, n));
				double Re = R0 + (1 - R0) * c * c * c * c * c, Tr = 1 - Re, P = .25 + .5 * Re,
					RP = Re / P, TP = Tr / (1 - P);
				
				// Russian roulette
				return vec_plus(objem, vec_mult(f, depth > 2 ? (myrand(rng) < P ? 
					vec_times(radiance(reflRay, sce, depth, rng), RP) : vec_times(radiance(elmray(x, tdir, id), sce, depth, rng), TP)) :
					vec_plus(vec_times(radiance(reflRay, sce, depth, rng), Re), vec_times(radiance(elmray(x, tdir, id), sce, depth, rng), Tr))));
			}
		
		} else

			return elmvec(0,0,0);
	}
}

/* exposure correction function */
float apply_exposure(float l)
{
	float expo = -1.00f;
	return (1.0f - expf(l * expo));
}

/* encoding to srgb */
float srgb_encode(float c)
{
	if (c <= 0.0031308f) 
		return 12.92f * c;
	else
		return 1.055f * powf(c, 0.4166667f) - 0.055f; // Inverse gamma 2.4
}

/* decoding from srgb */
float srgb_decode(float c)
{
	if (c <= 0.04045f) 
		return c * 0.07739938F;
	else
		return powf((c + 0.055F) * 0.947867F, 2.4F); // Direct gamma 2.4
}

/* clip float to unsigned char */
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))

/* encode a pixel */
unsigned char encode_pixel(float pixval)
{
	// exposure correction
	float pixvalec = apply_exposure(pixval);

	// gamma correction
	float pixvalge = srgb_encode(pixvalec);

	// ready for file
	unsigned char pixvaluc = FLOAT2UCHAR(pixvalge);

	return pixvaluc;
}

struct threadenv {
	int nt, startcol, stepcol;
	int *die;
	struct elmSCENE *sce;
	struct elmIMGF *imgf;
	int samps;
};

/* explore outer space with rays */
void threadproc(void *env)
{
	int y, x, s;
	struct threadenv *te = (struct threadenv *) env;
	pcg32_random_t rng = {time(NULL) ^ (intptr_t) &printf, (intptr_t) env};

	fprintf(stderr, "Thread %d cols %d,%d,%d,...\n", te->nt,
		te->startcol, te->startcol + te->stepcol, te->startcol + 2 * te->stepcol);

	// loop samples
	for (s = 0; s < te->samps && !*(te->die); s++) { 

		// Loop over image rows (top-bottom)
		for (y = 0; y < te->imgf->height; y++) { 

			// Loop cols (left-right)
			for (x = te->startcol; x < te->imgf->width; x += te->stepcol) {  

				int sx, sy, ns = 0;
				struct elmVEC r = {0,0,0}; // colors of smples

				// 2x2 subpixel rows
				for (sy = 0; sy < 2; sy++) { 

					// 2x2 subpixel cols
					for (sx = 0; sx < 2; sx++) { 

						/* random displacement around subpixel center */
						double r1 = 2 * myrand(&rng), dx = r1 < 1 ? sqrt(r1) - 1 : 1 - sqrt(2 - r1);
						double r2 = 2 * myrand(&rng), dy = r2 < 1 ? sqrt(r2) - 1 : 1 - sqrt(2 - r2);

						/* image plane pixel coord */
						struct elmVEC pixcoord = vec_plus(vec_plus(te->sce->cam.ipo, vec_times(te->sce->cam.cx, (sx + .5 + dx) / 2 + x)),
							vec_times(te->sce->cam.cy, (sy + .5 + dy) / 2 + y));

						/* primary ray direction */
						struct elmVEC pdir = vec_norm(vec_minus(pixcoord, te->sce->cam.view.orig));

						/* primary ray origin (out of camera) */
						struct elmVEC porig = vec_plus(te->sce->cam.view.orig, pdir);

						/* accumulate radiance in that pixel */
						r = vec_plus(r, radiance(elmray(porig, pdir, -1), te->sce, 0, &rng)); 	

						/* count samples done */
						ns++;
					}
				}

				// accumulate radiance on this point
				te->imgf->buf[(y * te->imgf->width + x) * 3 + 0] += (float) r.x;
				te->imgf->buf[(y * te->imgf->width + x) * 3 + 1] += (float) r.y;
				te->imgf->buf[(y * te->imgf->width + x) * 3 + 2] += (float) r.z;
				te->imgf->samps[y * te->imgf->width + x] += ns;
			}
		}
	}

	fprintf(stderr, "Thread %d exiting\n", te->nt);
}

/* add plane */
int add_plane(struct elmSCENE *sce, struct elmVEC pos, struct elmVEC nor, struct elmVEC em, struct elmVEC clr, int refl, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMPLANE;
	cobj->pos = pos;
	cobj->nor = vec_norm(nor);
	cobj->em = em;
	cobj->clr = clr;
	cobj->refl = refl;
	strncpy(cobj->name, name, 32);
	cobj->aabb.vmin = elmvec(vec_dot(cobj->nor, elmvec(1,0,0)) == 1 ? 0 : -Inf,
		vec_dot(cobj->nor, elmvec(0,1,0)) == 1 ? 0 : -Inf, vec_dot(cobj->nor, elmvec(0,0,1)) == 1 ? 0 : -Inf);
	cobj->aabb.vmax = elmvec(vec_dot(cobj->nor, elmvec(1,0,0)) == 1 ? 0 : +Inf,
		vec_dot(cobj->nor, elmvec(0,1,0)) == 1 ? 0 : +Inf, vec_dot(cobj->nor, elmvec(0,0,1)) == 1 ? 0 : +Inf);
	cobj->id = sce->nobjs - 1;
	return 1;
}

/* add sphere */
int add_sphere(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC em, struct elmVEC clr, int refl, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMSPHERE;
	cobj->rad = rad;
	cobj->pos = pos;
	cobj->em = em;
	cobj->clr = clr;
	cobj->refl = refl;
	strncpy(cobj->name, name, 32);
	cobj->aabb.vmin = vec_minus(cobj->pos, elmvec(cobj->rad, cobj->rad, cobj->rad));
	cobj->aabb.vmax = vec_plus(cobj->pos, elmvec(cobj->rad, cobj->rad, cobj->rad));
	cobj->id = sce->nobjs - 1;
	return 1;
}

/* add disc */
int add_disc(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, struct elmVEC em, struct elmVEC clr, int refl, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMDISC;
	cobj->rad = rad;
	cobj->pos = pos;
	cobj->nor = vec_norm(nor);
	cobj->em = em;
	cobj->clr = clr;
	cobj->refl = refl;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return 1;
}

/* add cylinder */
int add_cylinder(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, double height, struct elmVEC em, struct elmVEC clr, int refl, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMCYLINDER;
	cobj->rad = rad;
	cobj->pos = pos;
	cobj->nor = vec_norm(nor);
	cobj->hght = height;
	cobj->pos2 = vec_plus(cobj->pos, vec_times(cobj->nor, cobj->hght));
	cobj->em = em;
	cobj->clr = clr;
	cobj->refl = refl;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return 1;
}

/* add cone */
int add_cone(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, double height, double rad2, struct elmVEC em, struct elmVEC clr, int refl, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMCONE;
	cobj->rad = rad;
	cobj->pos = pos;
	cobj->nor = vec_norm(nor);
	cobj->hght = height;
	cobj->rad2 = rad2;
	cobj->pos2 = vec_plus(cobj->pos, vec_times(cobj->nor, cobj->hght));
	cobj->apx = vec_plus(cobj->pos, vec_times(vec_minus(cobj->pos2, cobj->pos), cobj->rad / (cobj->rad - cobj->rad2)));
	cobj->tga = (cobj->rad - cobj->rad2) / cobj->hght;
	cobj->em = em;
	cobj->clr = clr;
	cobj->refl = refl;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return 1;
}

/* add triangle */
int add_triangle(struct elmSCENE *sce, struct elmVEC pos, struct elmVEC pos2, struct elmVEC pos3, struct elmVEC em, struct elmVEC clr, int refl, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMTRIANGLE;
	cobj->pos = pos;
	cobj->pos2 = pos2;
	cobj->pos3 = pos3;
	cobj->nor = vec_norm(vec_cross(vec_minus(cobj->pos2, cobj->pos), vec_minus(cobj->pos3, cobj->pos)));
	cobj->em = em;
	cobj->clr = clr;
	cobj->refl = refl;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return 1;
}

/* prepare camera */
int prepare_camera(struct elmCAM *cam, struct elmIMGF *imgf, struct elmVEC pos, struct elmVEC aim, double hfov)
{
	struct elmVEC ipc;
	double fl, ipsw;

	// assign camera data
	cam->view = elmray(pos, vec_norm(vec_minus(aim, pos)), -1); // camera pos, d
	cam->hfov = hfov;

	// focal length
	fl = cos(0.5 * hfov * M_PI / 180);

	// semi-width of image plane
	ipsw = tan(0.5 * hfov * M_PI / 180);

	// image plane center
	ipc = vec_plus(cam->view.orig, vec_times(cam->view.dir, fl));

	// image plane horizontal director
	cam->cx = vec_times(vec_norm(vec_cross(elmvec(cam->view.dir.x, cam->view.dir.y, 0), elmvec(0, 0, 1))), 2 * ipsw / imgf->width);

	// image plane vertical director
	cam->cy = vec_times(vec_norm(vec_cross(cam->cx, cam->view.dir)), -(2 * ipsw / imgf->width));

	// image plane origin
	cam->ipo = vec_plus(ipc, vec_plus(vec_times(cam->cx, - imgf->width / 2.0), vec_times(cam->cy, -imgf->height / 2.0)));

	return 0;
}

/* populate scene */
/*int prepare_scene(struct elmSCENE *sce, struct elmIMGF *imgf)
{
	// camera
	prepare_camera(&sce->cam, imgf, elmvec(10,52,10), elmvec(27,16.5,47), 54);

	// smallpt scene
	add_sphere(sce, 1e5, elmvec(1e5+1,40.8,81.6), elmvec(0,0,0), elmvec(.75,.25,.25), elmDIFF, "Left");
	add_sphere(sce, 1e5, elmvec(-1e5+99,40.8,81.6), elmvec(0,0,0), elmvec(.25,.25,.75), elmDIFF, "Right");
	add_sphere(sce, 1e5, elmvec(50,40.8, 1e5), elmvec(0,0,0), elmvec(.75,.75,.75), elmDIFF, "Back");
	add_sphere(sce, 1e5, elmvec(50,40.8,-1e5+170), elmvec(0,0,0), elmvec(0,0,0), elmDIFF, "Front");
	add_sphere(sce, 1e5, elmvec(50, 1e5, 81.6), elmvec(0,0,0), elmvec(.75,.75,.75), elmDIFF, "Bottom");
	add_sphere(sce, 1e5, elmvec(50,-1e5+81.6,81.6), elmvec(0,0,0), elmvec(.75,.75,.75), elmDIFF, "Top");
	add_sphere(sce, 16.5, elmvec(27,16.5,47), elmvec(0,0,0), elmvec(.999,.999,.999), elmSPEC, "Mirror");
	add_sphere(sce, 16.5, elmvec(73,16.5,78), elmvec(0,0,0), elmvec(.999,.999,.999), elmREFR, "Glass");
	add_sphere(sce, 600, elmvec(50,681.6-.27,81.6), elmvec(12,12,12), elmvec(0,0,0), elmDIFF, "Light");
	return 1;
}*/

/*int prepare_scene(struct elmSCENE *sce, struct elmIMGF *imgf)
{
	// camera
	sce->cam.view = elmray(elmvec(50,52,295.6), vec_norm(elmvec(0,-0.062612,-1))); // camera pos, dir
	sce->cam.cx = elmvec(imgf->width * .5135 / imgf->height, 0, 0); // x dir increment
	sce->cam.cy = vec_times(vec_norm(vec_cross(sce->cam.cx, sce->cam.view.dir)), .5135); // y direction increment with cross product

	// plane hovered by light
	add_plane(sce, elmvec(0,0,0), elmvec(0,-1,0), elmvec(0,0,0), elmvec(.999,0,0), elmDIFF, "Floor plane");
	add_sphere(sce, 16, elmvec(50,16,81.6), elmvec(0,0,0), elmvec(.999,.999,.999), elmREFR, "Glass sphere");
	add_sphere(sce, 12, elmvec(35,12,50), elmvec(0,0,0), elmvec(.999,.999,.999), elmDIFF, "White sphere");
	add_sphere(sce, 6, elmvec(65,100,81.6), elmvec(50,50,50), elmvec(0,0,0), elmDIFF, "Light");
	return 1;
}*/

int prepare_scene(struct elmSCENE *sce, struct elmIMGF *imgf)
{
	// camera
	prepare_camera(&sce->cam, imgf, elmvec(-90,0,45), elmvec(5,0,0), 54);

	// plane hovered by light
	add_plane(sce, elmvec(0,0,0), elmvec(0,0,1), elmvec(0,0,0), elmvec(.67,.56,.56), elmDIFF, "Floor plane");
	add_sphere(sce, 12, elmvec(0,0,12), elmvec(0,0,0), elmvec(.32,.32,.32), elmDIFF, "Ball");
	add_sphere(sce, 8, elmvec(-60,-30,62), elmvec(50,50,50), elmvec(.32,.32,.32), elmDIFF, "Light 1");
	add_sphere(sce, 7, elmvec(-200,5,85), elmvec(60,60,50), elmvec(.32,.32,.32), elmDIFF, "Light 2");
	add_sphere(sce, 10, elmvec(-20,25,10), elmvec(0,0,0), elmvec(.90,.90,.90), elmSPEC, "Mirror");
	add_sphere(sce, 10, elmvec(-20,-25,10), elmvec(0,0,0), elmvec(.93,.93,.98), elmREFR, "Glass");
	add_disc(sce, 3, elmvec(-40,15,3), elmvec(-1,0,0), elmvec(0,0,0), elmvec(.4,.4,.5), elmDIFF, "Disc");
	add_cylinder(sce, 5, elmvec(-40,4,0), elmvec(0,0,1), 5, elmvec(0,0,0), elmvec(.98,.98,.93), elmREFR, "Cylinder");
	add_cone(sce, 5, elmvec(-40,-10,0), elmvec(0,0,1), 16, 0, elmvec(0,0,0), elmvec(.99,.99,.99), elmDIFF, "Cone");
	add_triangle(sce, elmvec(-40,22,0), elmvec(-40,26,5), elmvec(-40,26,0), elmvec(5,1,1), elmvec(.76,.15,.13), elmDIFF, "Triangle");

	// ambient occlusion
	sce->amb.clr = elmvec(0.3,0.3,0.5);
	sce->amb.str = 0.3;

	/* load the sky */
	if (0) {

		struct elmIMG sky = {0};
		int row, col;
		char ambientfile[] = "skyboxsun25degtest.bmp";
		//char ambientfile[] = "skybox_texture.bmp";
		//char ambientfile[] = "grimmnight_large.bmp";

		fprintf(stderr, "Loading ambient \"%s\"...", ambientfile);
		bmp_read(ambientfile, &sky);
		sce->amb.map.width = sky.width;
		sce->amb.map.height = sky.height;
		sce->amb.map.buf = calloc(sky.width * sky.height * 3, sizeof(float));
		for (row = 0; row < sky.height; row++) {
			for (col = 0; col < sky.width; col++) {
				float P0 = srgb_decode((float) sky.buffer[row * sky.stride + col * sky.depth * sky.bytes + 0] / 255.0F);
				float P1 = srgb_decode((float) sky.buffer[row * sky.stride + col * sky.depth * sky.bytes + 1] / 255.0F);
				float P2 = srgb_decode((float) sky.buffer[row * sky.stride + col * sky.depth * sky.bytes + 2] / 255.0F);
				float thr = 0.995F, gain = 15000.0F;
				if (P0 > thr && P1 > thr && P2 > thr) {
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 0] = (P0 - thr) * gain + thr;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 1] = (P1 - thr) * gain + thr;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 2] = (P2 - thr) * gain + thr;
				} else {
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 0] = P0;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 1] = P1;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 2] = P2;
				}
			}
		}
		free(sky.buffer);
		fprintf(stderr, " done\n");
	}

	return 1;
}

/* dump to output file */
int image_dump(char *filename, struct elmIMGF *imgf)
{
	int x, y;
	struct elmIMGF imgf2 = *imgf;
	struct elmIMG img = {imgf->width, imgf->height, 3, 255};
	img.bytes = img.maxval > 255 ? 2 : 1;
	img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
	img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

	imgf2.buf = (float *) calloc(imgf2.height * imgf2.width * 3, sizeof(float));
	memcpy(imgf2.buf, imgf->buf, imgf2.height * imgf2.width * 3 * sizeof(float));
	//post_process(&imgf2, &imgf);

	for (y = 0; y < imgf->height; y++) {
		for (x = 0; x < imgf->width; x++) {
			img.buffer[y * img.stride + x * img.depth * img.bytes + 0] =
				encode_pixel(imgf2.buf[(y * imgf2.width + x) * 3 + 0] / imgf2.samps[y * imgf2.width + x]);
			img.buffer[y * img.stride + x * img.depth * img.bytes + 1] =
				encode_pixel(imgf2.buf[(y * imgf2.width + x) * 3 + 1] / imgf2.samps[y * imgf2.width + x]);
			img.buffer[y * img.stride + x * img.depth * img.bytes + 2] =
				encode_pixel(imgf2.buf[(y * imgf2.width + x) * 3 + 2] / imgf2.samps[y * imgf2.width + x]);
		}
	}

	free(imgf2.buf);

	/* dump image */
	bmp_write(filename, &img);

	/* free image */
	free(img.buffer);
	memset(&img, 0, sizeof(struct elmIMG));

	return 0;
}

/* main function */
int main(int argc, char **argv)
{
	int y, nt, die = 0;
	int numthreads = 6;
	double totsamples = 0, lastsamples = 0;
	struct elmSCENE scene = {0};
	struct threadenv *te = NULL;
	struct elmIMGF imgf = {640, 480};
	int samps = 16384; // # samples
	HANDLE *hands = NULL;

	/* prepare image */
	imgf.buf = (float *) calloc(imgf.height * imgf.width * 3, sizeof(float));
	imgf.samps = (int *) calloc(imgf.height * imgf.width, sizeof(int));
	totsamples = imgf.width * imgf.height * 4 * samps;

	/* prepare scene */
	prepare_scene(&scene, &imgf);

	/* start threads */
	te = calloc(numthreads, sizeof(struct threadenv));
	hands = calloc(numthreads, sizeof(HANDLE));
	for (nt = 0; nt < numthreads; nt++) {
		te[nt].nt = nt;
		te[nt].startcol = nt;
		te[nt].stepcol = numthreads;
		te[nt].die = &die;
		te[nt].sce = &scene;
		te[nt].imgf = &imgf;
		te[nt].samps = samps;
		hands[nt] = (HANDLE) _beginthread(threadproc, 0, &te[nt]);
	}

	do {
		int sleepms = 2000;
		int KeyInfo = _kbhit() ? _getch() : 0;
		double currsamples = 0;
		static int dumpimg = 1;
		static double currtime, nextdumptime = 0;

		currtime = timeGetTime();
		Sleep(sleepms);

		if (tolower(KeyInfo) == 'd')
			dumpimg = !dumpimg;
		else if (tolower(KeyInfo) == 'k')
			die = 1;

		/* write to file */
		if (dumpimg && timeGetTime() >= nextdumptime) {
			image_dump("image.bmp", &imgf);
			fprintf(stderr, "image dumped\n");
			nextdumptime = timeGetTime() + 20000;
		}

		/* count samples done */
		for (y = 0; y < imgf.height * imgf.width; y++)
			currsamples += imgf.samps[y];

		fprintf(stderr, "%.1f krays/s (%.1f spp)\n", (currsamples - lastsamples) / (timeGetTime() - currtime), currsamples / (imgf.height * imgf.width));
		lastsamples = currsamples;

	} while (!die);

	/* wait all threads */
	WaitForMultipleObjects(numthreads, hands, TRUE, INFINITE);

	/* save safely to disk */
	image_dump("image.bmp", &imgf);
	fprintf(stderr, "image dumped\n");

	/* free image, scene, and all */
	free(imgf.buf);
	free(imgf.samps);
	free(te);
	free(scene.obj);
	free(scene.amb.map.buf);
	free(hands);

	return EXIT_SUCCESS;
}
