/* restart again from scratch */
/* conversion of smallpt using prof. Cline slides */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <conio.h>
#include <time.h>
#include "elabmm.h"

/* the infinite value */
extern const double Inf = ((double) (1e308 * 10));

/* swap two doubles */
#define DOUBLE_SWAP(A, B) {double temp = (A); (A) = (B); (B) = temp;}

/******************************/
/********* VECTORS ************/
/******************************/

/* coordinate/vector/color triplets */
struct elmVEC {
	double x, y, z; /* the three components */
};

/* constructor */
struct elmVEC elmvec(double x, double y, double z)
{
	struct elmVEC c = {x, y, z};
	return c;
}

/* component-wise addition */
struct elmVEC vec_plus(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x + b.x, a.y + b.y, a.z + b.z};
	return c;
}

/* component-wise subtraction */
struct elmVEC vec_minus(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x - b.x, a.y - b.y, a.z - b.z};
	return c;
}

/* component-wise scalar-vector multiplication */
struct elmVEC vec_times(struct elmVEC a, double b)
{
	struct elmVEC c = {b * a.x, b * a.y, b * a.z};
	return c;
}

/* component-wise vector-vector multiplication */
struct elmVEC vec_mult(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.x * b.x, a.y * b.y, a.z * b.z};
	return c;
}

/* length of vector */
double vec_length(struct elmVEC a)
{
	return sqrt(a.x * a.x + a.y * a.y + a.z * a.z);
}

/* normalization to length 1 */
struct elmVEC vec_norm(struct elmVEC a)
{
	double ilen = 1 / sqrt(a.x * a.x + a.y * a.y + a.z * a.z);
	struct elmVEC c = {a.x * ilen, a.y * ilen, a.z * ilen};
	return c;
}

/* scalar (dot) product of vectors */
double vec_dot(struct elmVEC a, struct elmVEC b)
{
	return (a.x * b.x + a.y * b.y + a.z * b.z);
}
double vecp_dot(struct elmVEC *a, struct elmVEC *b)
{
	return (a->x * b->x + a->y * b->y + a->z * b->z);
}

/* vector (cross) product of vectors */
struct elmVEC vec_cross(struct elmVEC a, struct elmVEC b)
{
	struct elmVEC c = {a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x};
	return c;
}
struct elmVEC vecp_cross(struct elmVEC *a, struct elmVEC *b)
{
	struct elmVEC c = {a->y * b->z - a->z * b->y, a->z * b->x - a->x * b->z, a->x * b->y - a->y * b->x};
	return c;
}

/* maximum of vector */
double vec_max(struct elmVEC f)
{
	return f.x > f.y && f.x > f.z ? f.x : f.y > f.z ? f.y : f.z;
}

/******************************/
/*********** RAY **************/
/******************************/

/* a ray */ 
struct elmRAY {
	struct elmVEC orig; /* origin */
	struct elmVEC dir; /* direction */
	int objid; /* ray origin object id: -1 means camera */
};

/* constructor */
struct elmRAY elmray(struct elmVEC orig, struct elmVEC dir, int objid)
{
	struct elmRAY r;
	r.orig = orig;
	r.dir = dir;
	r.objid = objid;
	return r;
}

/******************************/
/********** MATERIAL **********/
/******************************/

/* reflection type: diffuse, specular, refractive */
enum {elmDIFF, elmSPEC, elmREFR}; 

/* a material */
struct elmMAT {
    struct elmVEC emiclr; /* emission color */
    struct elmVEC diffclr; /* diffusion color */
    struct elmVEC specclr; /* specular color */
    struct elmVEC refrclr; /* refractive color */
	double comp[3]; /* relative amounts of diffusive, specular, refractive */
    struct elmVEC (*emiFunc)(struct elmVEC*, struct elmOBJ*); /* pointer to function for procedural emission color: takes the hit point and object, returns the color */
    struct elmVEC (*diffFunc)(struct elmVEC*, struct elmOBJ*); /* pointer to function for procedural diffusion color: takes the hit point and object, returns the color */
};

/* constructor */
struct elmMAT elmmat(struct elmVEC emiclr, struct elmVEC diffclr, struct elmVEC specclr, double comp0, double comp1, double comp2)
{
	struct elmMAT m;
    m.emiclr = emiclr;
    m.diffclr = diffclr;
    m.specclr = specclr;
	m.comp[0] = comp0;
	m.comp[1] = comp1;
	m.comp[2] = comp2;
    m.diffFunc = NULL;
    m.emiFunc = NULL;
	return m;
}
	
/******************************/
/********** OBJECT ************/
/******************************/

/* a generic object */
struct elmOBJ {
	enum {ELMPLANE, ELMSPHERE, ELMDISC, ELMCYLINDER, ELMCONE, ELMTRIANGLE, ELMQUADRANGLE} kind;

	union {

		/* plane */
		struct elmPLA {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* center */
			struct elmVEC ori; /* orientation */
		} pla;

		/* sphere */
		struct elmSPH {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* center */
			struct elmVEC ori; /* orientation */
			double rad; /* radius */
		} sph;

		/* disc */
		struct elmDSC {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* center */
			struct elmVEC ori; /* orientation */
			double rad; /* radius */
		} dsc;

		/* cylinder */
		struct elmCYL {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* center */
			struct elmVEC ori; /* orientation */
			double rad; /* radius */
			double hght; /* height */
			struct elmVEC pos2; /* top center */
		} cyl;

		/* cone */
		struct elmCON {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* center */
			struct elmVEC ori; /* orientation */
			double rad; /* base radius */
			double hght; /* height */
			struct elmVEC pos2; /* top center */
			double rad2; /* top radius */
			double tga; /* tangent of half angle */
			struct elmVEC apx; /* apex */
		} con;

		/* triangle */
		struct elmTRI {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* first vertex */
			struct elmVEC ori; /* orientation */
			struct elmVEC pos2; /* second vertex */
			struct elmVEC pos3; /* third vertex */
		} tri;

		/* quadrate */
		struct elmQUAD {
			struct elmVEC nor; /* normal */
			struct elmVEC pos; /* first vertex */
			struct elmVEC ori; /* orientation */
			struct elmVEC pos2; /* second vertex */
			struct elmVEC pos3; /* third vertex */
			struct elmVEC pos4; /* fourth vertex */
		} quad;

	};

	struct elmMAT mat; /* object material */
	char name[32]; /* name */
	struct {
		struct elmVEC vmin, vmax;
	} aabb; /* axis aligned bounding box */
	int id; /* object identifier */
};

/* a floating point image */
struct elmIMGF {
	int width, height; /* width and height */
	float *buf; /* buffer for pixels */
	int *samps; /* buffer for samples/pixel */
};

/******************************/
/********** CAMERA ************/
/******************************/

/* the camera */
struct elmCAM {
	/* main parameters */
	struct elmRAY view; // camera pos & dir
	double hfov; // horizontal field of view, in degrees
	double fdist; // focal distance
	struct elmIMGF imgf; // image formed in the camera
	/* work variables */
	struct elmVEC cx; // x dir increment
	struct elmVEC cy; // y direction increment with cross product
	struct elmVEC ipo; // image plane origin
	double fl; /* focal length */
};

/******************************/
/********** SCENE *************/
/******************************/

/* the scene */
struct elmSCENE {
	int nobjs; /* number of objects */
	struct elmOBJ *obj; /* pointer to objects */
	struct elmCAM cam; /* camera */
	struct {
		struct elmVEC clr; /* ambient color */
		struct elmIMGF map; /* cubemap image */
		double str; /* ambient strength */
	} amb; /* ambient */
};

/******************************/
/****** INTERSECTIONS *********/
/******************************/

/* intersection between sphere and ray: returns distance or 0 if miss */
double intersect_sphere(struct elmSPH *sph, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* solve the 2nd degree equation */
	struct elmVEC op = vec_minus(sph->pos, ray->orig);
	double t, eps = 1E-4; /* fudge factor */
	double b = vec_dot(op, ray->dir); /* in quadratic solver */
	double det = b * b - vec_dot(op, op) + sph->rad * sph->rad;

	if (det < 0)
		return 0; /* ray misses sphere */
	else
		det = sqrt(det);

	/* return smaller positive t */
	if ((t = b - det) > eps) {
		*hitp = vec_plus(ray->orig, vec_times(ray->dir, t)); /* ray intersection point */
		*hitn = vec_norm(vec_minus(*hitp, sph->pos)); // sphere normal
		return t;
	} else if ((t = b + det) > eps) {
		*hitp = vec_plus(ray->orig, vec_times(ray->dir, t)); /* ray intersection point */
		*hitn = vec_norm(vec_minus(*hitp, sph->pos)); // sphere normal
		return t;
	} else
		return 0;
}

/* intersection between plane and ray: returns distance or 0 if miss */
double intersect_plane(struct elmPLA *pla, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* search */
	double denom = vec_dot(pla->nor, ray->dir);
	if (denom < -1e-12 || denom > 1e-12) {
		/* not parallel */
		double t = vec_dot(vec_minus(pla->pos, ray->orig), pla->nor) / denom; 
		if (t > 0) {
			/* in front */
			*hitn = pla->nor;
			*hitp = vec_plus(ray->orig, vec_times(ray->dir, t));
			return t;
		} else
			/* behind */
			return 0;
	} else
		/* parallel */
		return 0;
}

/* intersection between disc and ray: returns distance or 0 if miss */
double intersect_disc(struct elmDSC *dsc, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* search */
	double denom = vec_dot(dsc->nor, ray->dir);
	if (denom < -1e-12 || denom > 1e-12) {
		/* not parallel */
		double t = vec_dot(vec_minus(dsc->pos, ray->orig), dsc->nor) / denom; 
		if (t > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t));
			struct elmVEC v = vec_minus(p, dsc->pos);
			if (vec_dot(v, v) <= dsc->rad * dsc->rad) {
				/* in disc */
				*hitn = dsc->nor;
				*hitp = p;
				return t;
			} else
				return 0;
		} else
			/* behind */
			return 0;
	} else
		/* parallel */
		return 0;
}

/* intersection between ray and cylinder: returns distance or 0 if miss */
double intersect_cylinder(struct elmCYL *cyl, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	struct elmVEC v1 = vec_times(cyl->nor, vec_dot(ray->dir, cyl->nor));
	struct elmVEC v2 = vec_minus(ray->dir, v1);
	double A = vec_dot(v2, v2);
	struct elmVEC v3 = vec_minus(ray->orig, cyl->pos);
	struct elmVEC v4 = vec_times(cyl->nor, vec_dot(v3, cyl->nor));
	struct elmVEC v5 = vec_minus(v3, v4);
	double B = 2 * vec_dot(v2, v5);
	double C = vec_dot(v5, v5) - cyl->rad * cyl->rad;
	double det = B * B - 4 * A * C;
	double t1, t2, t3, t4, eps = 1e-4;
	double denom;

	/* ray misses cylinder */
	if (det < 0)
		return 0; 
	else
		det = sqrt(det);

	/* first possible intersection point - front side of tube */
	t1 = (-B - det) / (2 * A);
	if (t1 > eps) {
		/* in front of us */
		struct elmVEC q1 = vec_plus(ray->orig, vec_times(ray->dir, t1));
		if (vec_dot(cyl->nor, vec_minus(q1, cyl->pos)) > 0 && vec_dot(cyl->nor, vec_minus(q1, cyl->pos2)) < 0) {
			/* is in the external tube */
			*hitp = q1; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q1, vec_plus(cyl->pos, vec_times(cyl->nor, vec_dot(cyl->nor, vec_minus(q1, cyl->pos)))))) ; // normal
			return t1;
		}
	}	

	/* try caps */
	denom = vec_dot(cyl->nor, ray->dir);
	if (denom < -1e-12 || denom > 1e-12) {
		/* not parallel */
		/* bottom cap */
		t3 = vec_dot(vec_minus(cyl->pos, ray->orig), cyl->nor) / denom; 
		if (t3 > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t3));
			struct elmVEC v = vec_minus(p, cyl->pos);
			if (vec_dot(v, v) <= cyl->rad * cyl->rad) {
				/* in disc */
				*hitn = vec_times(cyl->nor, -1);
				*hitp = p;
				return t3;
			} 
		} 
		/* top cap */
		t4 = vec_dot(vec_minus(cyl->pos2, ray->orig), cyl->nor) / denom; 
		if (t4 > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t4));
			struct elmVEC v = vec_minus(p, cyl->pos2);
			if (vec_dot(v, v) <= cyl->rad * cyl->rad) {
				/* in disc */
				*hitn = cyl->nor;
				*hitp = p;
				return t4;
			} 
		} 
	}

	/* last possible intersection point - back side of tube */
	t2 = (-B + det) / (2 * A);
	if (t2 > eps) {
		/* in front of us */
		struct elmVEC q2 = vec_plus(ray->orig, vec_times(ray->dir, t2));
		if (vec_dot(cyl->nor, vec_minus(q2, cyl->pos)) > 0 && vec_dot(cyl->nor, vec_minus(q2, cyl->pos2)) < 0) {
			/* is in the external tube */
			*hitp = q2; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q2, vec_plus(cyl->pos, vec_times(cyl->nor, vec_dot(cyl->nor, vec_minus(q2, cyl->pos)))))) ; // normal
			return t2;
		}
	}	

	return 0;
}

/* intersection between ray and cone: returns distance or 0 if miss */
double intersect_cone(struct elmCON *con, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	double dirdotnor = vec_dot(ray->dir, con->nor);
	struct elmVEC v1 = vec_minus(ray->dir, vec_times(con->nor, dirdotnor));
	double cos2a = 1 / (1 + con->tga * con->tga);
	double sin2a = cos2a * con->tga * con->tga;
	double A = cos2a * vec_dot(v1, v1) - sin2a * dirdotnor * dirdotnor;
	struct elmVEC v2 = vec_minus(ray->orig, con->apx);
	double v2dotnor = vec_dot(v2, con->nor);
	struct elmVEC v3 = vec_minus(v2, vec_times(con->nor, v2dotnor));
	double B = 2 * cos2a * vec_dot(v1, v3) - 2 * sin2a * dirdotnor * v2dotnor;
	double C = cos2a * vec_dot(v3, v3) - sin2a * v2dotnor * v2dotnor;
	double det = B * B - 4 * A * C;
	double t1, t2, t3, t4, eps = 1e-4;
	double denom;

	/* ray misses cone */
	if (det < 0)
		return 0; 
	else
		det = sqrt(det);

	/* first possible intersection point - front side of cone */
	t1 = (-B - det) / (2 * A);
	if (t1 > eps) {
		/* in front of us */
		struct elmVEC q1 = vec_plus(ray->orig, vec_times(ray->dir, t1));
		if (vec_dot(con->nor, vec_minus(q1, con->pos)) > 0 && vec_dot(con->nor, vec_minus(q1, con->pos2)) < 0) {
			/* is in the external tube */
			struct elmVEC x = vec_times(con->nor, vec_dot(vec_minus(q1, con->pos), con->nor));
			struct elmVEC O = vec_plus(con->pos, vec_times(con->nor, vec_length(x) * (1 + con->tga * con->tga) - con->tga * con->rad));
			*hitp = q1; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q1, O));
			return t1;
		}
	}	

	/* try caps */
	denom = vec_dot(con->nor, ray->dir);
	if (denom < -1e-12 || denom > 1e-12) {
		/* not parallel */
		/* bottom cap */
		t3 = vec_dot(vec_minus(con->pos, ray->orig), con->nor) / denom; 
		if (t3 > 0) {
			/* in front */
			struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t3));
			struct elmVEC v = vec_minus(p, con->pos);
			if (vec_dot(v, v) <= con->rad * con->rad) {
				/* in disc */
				*hitn = vec_times(con->nor, -1);
				*hitp = p;
				return t3;
			} 
		} 
		/* top cap */
		if (con->rad2) {
			t4 = vec_dot(vec_minus(con->pos2, ray->orig), con->nor) / denom; 
			if (t4 > 0) {
				/* in front */
				struct elmVEC p = vec_plus(ray->orig, vec_times(ray->dir, t4));
				struct elmVEC v = vec_minus(p, con->pos2);
				if (vec_dot(v, v) <= con->rad2 * con->rad2) {
					/* in disc */
					*hitn = con->nor;
					*hitp = p;
					return t4;
				} 
			}
		}
	}

	/* last possible intersection point - back side of tube */
	t2 = (-B + det) / (2 * A);
	if (t2 > eps) {
		/* in front of us */
		struct elmVEC q2 = vec_plus(ray->orig, vec_times(ray->dir, t2));
		if (vec_dot(con->nor, vec_minus(q2, con->pos)) > 0 && vec_dot(con->nor, vec_minus(q2, con->pos2)) < 0) {
			/* is in the external tube */
			struct elmVEC x = vec_times(con->nor, vec_dot(vec_minus(q2, con->pos), con->nor));
			struct elmVEC O = vec_plus(con->pos, vec_times(con->nor, vec_length(x) * (1 + con->tga * con->tga) - con->tga * con->rad));
			*hitp = q2; /* ray intersection point */
			*hitn = vec_norm(vec_minus(q2, O));
			return t2;
		}
	}

	return 0;
}

/* intersection between ray and triangle: returns distance or 0 if miss */
double intersect_triangle(struct elmTRI *tri, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	double nordotdir = vec_dot(tri->nor, ray->dir);
	if (nordotdir < -1e-12 || nordotdir > 1e-12) {
		// not parallel
		double t = vec_dot(vec_minus(tri->pos, ray->orig), tri->nor) / nordotdir; 
		if (t < 0)
			// behind
			return 0;
		else {
			struct elmVEC P = vec_plus(ray->orig, vec_times(ray->dir, t));
			struct elmVEC edge0 = vec_minus(tri->pos2, tri->pos);
			struct elmVEC VP0 = vec_minus(P, tri->pos);
			struct elmVEC C = vec_cross(edge0, VP0);
			if (vec_dot(tri->nor, C) < 0)
				return 0;
			else {
				struct elmVEC edge1 = vec_minus(tri->pos3, tri->pos2);
				struct elmVEC VP1 = vec_minus(P, tri->pos2);
				C = vec_cross(edge1, VP1);
				if (vec_dot(tri->nor, C) < 0)
					return 0;
				else {
					struct elmVEC edge2 = vec_minus(tri->pos, tri->pos3);
					struct elmVEC VP2 = vec_minus(P, tri->pos3);
					C = vec_cross(edge2, VP2);
					if (vec_dot(tri->nor, C) < 0)
						return 0;
					else {
						*hitp = P;
						*hitn = tri->nor;
						return t;
					}
				}
			}
		}
	} else
		return 0;
}

/* intersection between ray and rectangle: returns distance or 0 if miss */
double intersect_rectangle(struct elmQUAD *quad, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	double nordotdir = vec_dot(quad->nor, ray->dir);
	if (nordotdir < -1e-12 || nordotdir > 1e-12) {
		// not parallel
		double t = vec_dot(vec_minus(quad->pos, ray->orig), quad->nor) / nordotdir; 
		if (t < 0)
			// behind
			return 0;
		else {
			struct elmVEC P = vec_plus(ray->orig, vec_times(ray->dir, t));
			struct elmVEC edge0 = vec_minus(quad->pos2, quad->pos);
			struct elmVEC VP0 = vec_minus(P, quad->pos);
			struct elmVEC C = vec_cross(edge0, VP0);
			if (vec_dot(quad->nor, C) < 0)
				return 0;
			else {
				struct elmVEC edge1 = vec_minus(quad->pos3, quad->pos2);
				struct elmVEC VP1 = vec_minus(P, quad->pos2);
				C = vec_cross(edge1, VP1);
				if (vec_dot(quad->nor, C) < 0)
					return 0;
				else {
					struct elmVEC edge2 = vec_minus(quad->pos4, quad->pos3);
					struct elmVEC VP2 = vec_minus(P, quad->pos3);
					C = vec_cross(edge2, VP2);
					if (vec_dot(quad->nor, C) < 0)
						return 0;
					else {
						struct elmVEC edge3 = vec_minus(quad->pos, quad->pos4);
						struct elmVEC VP3 = vec_minus(P, quad->pos4);
						C = vec_cross(edge3, VP3);
						if (vec_dot(quad->nor, C) < 0)
							return 0;
						else {
							*hitp = P;
							*hitn = quad->nor;
							return t;
						}
					}
				}
			}
		}
	} else
		return 0;
}

/* intersect aabb */
int intersect_aabb(struct elmRAY *ray, struct elmVEC vmin, struct elmVEC vmax)
{
	double tmin, tymin, tzmin;
	double tmax, tymax, tzmax;
	tmin = (vmin.x - ray->orig.x) / ray->dir.x;
	tmax = (vmax.x - ray->orig.x) / ray->dir.x;
	if (tmin > tmax)
		DOUBLE_SWAP(tmin, tmax);
	tymin = (vmin.y - ray->orig.y) / ray->dir.y;
	tymax = (vmax.y - ray->orig.y) / ray->dir.y;
	if (tymin > tymax)
		DOUBLE_SWAP(tymin, tymax);
	if ((tmin > tymax) || (tymin > tmax))
		return 0;
	if (tymin > tmin)
		tmin = tymin;
	if (tymax < tmax)
		tmax = tymax;
	tzmin = (vmin.z - ray->orig.z) / ray->dir.z;
	tzmax = (vmax.z - ray->orig.z) / ray->dir.z;
	if (tzmin > tzmax)
		DOUBLE_SWAP(tzmin, tzmax);
	if ((tmin > tzmax) || (tzmin > tmax))
		return 0;
	if (tzmin > tmin)
		tmin = tzmin;
	if (tzmax < tmax)
		tmax = tzmax;
	/*if ((tmin > r.tmax) || (tmax < r.tmin)) return false;
	if (r.tmin < tmin) r.tmin = tmin;
	if (r.tmax > tmax) r.tmax = tmax;*/
	return 1;
}

/* intersect object: return distance or 0 if miss */
double intersect_obj(struct elmOBJ *obj, struct elmRAY *ray, struct elmVEC *hitp, struct elmVEC *hitn)
{
	/* fast intersect aabb */
	if (1 /*intersect_aabb(ray, obj->aabb.vmin, obj->aabb.vmax)*/) {

		/* depending on type */
		switch (obj->kind) {
		case ELMSPHERE:
			return intersect_sphere(&obj->sph, ray, hitp, hitn);
			break;
		case ELMPLANE:
			return ray->objid == obj->id ? 0 : intersect_plane(&obj->pla, ray, hitp, hitn);
			break;
		case ELMDISC:
			return ray->objid == obj->id ? 0 : intersect_disc(&obj->dsc, ray, hitp, hitn);
			break;
		case ELMCYLINDER:
			return intersect_cylinder(&obj->cyl, ray, hitp, hitn);
			break;
		case ELMCONE:
			return intersect_cone(&obj->con, ray, hitp, hitn);
			break;
		case ELMTRIANGLE:
			return ray->objid == obj->id ? 0 : intersect_triangle(&obj->tri, ray, hitp, hitn);
			break;
		case ELMQUADRANGLE:
			return ray->objid == obj->id ? 0 : intersect_rectangle(&obj->quad, ray, hitp, hitn);
			break;
		default:
			return 0;
		}
	} else
		return 0;
}

/* intersects ray with scene */
int intersect_scene(struct elmRAY *ray, struct elmSCENE *sce, double *t, int *id, struct elmVEC *hitp, struct elmVEC *hitn)
{
	double d = 0;
	int i = 0;
	struct elmVEC bhitp = {0}, bhitn = {0};

	/* explore all objects */
	*t = Inf;
	for (i = 0; i < sce->nobjs; i++) {
		/* try to intersect the object */
		if ((d = intersect_obj(&sce->obj[i], ray, &bhitp, &bhitn)) && d < *t) {
			*t = d;
			*id = i;
			*hitp = bhitp;
			*hitn = bhitn;
		}
	}

	/* return success or not */
	return *t < Inf;
}

// *Really* minimal PCG32 code / (c) 2014 M.E. O'Neill / pcg-random.org
// Licensed under Apache License 2.0 (NO WARRANTY, etc. see website)
typedef struct { unsigned __int64 state;  unsigned __int64 inc; } pcg32_random_t;
unsigned long int pcg32_random_r(pcg32_random_t* rng)
{
    unsigned __int64 oldstate = rng->state;
	unsigned long int xorshifted;
	long int rot;

    // Advance internal state
    rng->state = oldstate * 6364136223846793005ULL + (rng->inc|1);

    // Calculate output function (XSH RR), uses old state for max ILP
    xorshifted = (unsigned long int) (((oldstate >> 18u) ^ oldstate) >> 27u);
    rot = oldstate >> 59u;

    return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
}

/* random number between 0 and 1 */
double double_rand(pcg32_random_t *rng)
{
	return (double) pcg32_random_r(rng) / (double) UINT_MAX;
}

/* random integer with given probability */
int int_rand(pcg32_random_t *rng, double *prob, int nmax)
{
	int s = 0;
	double d = double_rand(rng), accp = 0;
	while (accp < d && s < nmax) {
		accp += prob[s++];
	}
	return s - 1;
}

/* get a texture pixel with bilinear interpolation */
struct elmVEC map_texture(struct elmIMGF *imgf, double x, double y, double *alpha)
{
	struct elmVEC ilmout = {0,0,0};
	double u, v;
	double alpha0, alpha1, alpha2, alpha3;
	int col, row, depth = alpha ? 4 : 3;
	struct elmVEC P0, P1, P2, P3;
	if (x < 0)
		x = 0;
	else if (x > imgf->width - 2)
		x = imgf->width - 2;
	if (y < 1)
		y = 1;
	else if (y > imgf->height - 1)
		y = imgf->height - 1;
	u = x - floorf(x);
	v = 1 - (y - floorf(y));
	col = (int) x;
	row = imgf->height - 1 - (int) y;
	P0.x = imgf->buf[(row * imgf->width + col) * depth + 0];
	P0.y = imgf->buf[(row * imgf->width + col) * depth + 1];
	P0.z = imgf->buf[(row * imgf->width + col) * depth + 2];
	P1.x = imgf->buf[(row * imgf->width + col + 1) * depth + 0];
	P1.y = imgf->buf[(row * imgf->width + col + 1) * depth + 1];
	P1.z = imgf->buf[(row * imgf->width + col + 1) * depth + 2];
	P2.x = imgf->buf[((row + 1) * imgf->width + col) * depth + 0];
	P2.y = imgf->buf[((row + 1) * imgf->width + col) * depth + 1];
	P2.z = imgf->buf[((row + 1) * imgf->width + col) * depth + 2];
	P3.x = imgf->buf[((row + 1) * imgf->width + col + 1) * depth + 0];
	P3.y = imgf->buf[((row + 1) * imgf->width + col + 1) * depth + 1];
	P3.z = imgf->buf[((row + 1) * imgf->width + col + 1) * depth + 2];
	if (alpha) {
		alpha0 = imgf->buf[(row * imgf->width + col) * depth + 3];
		alpha1 = imgf->buf[(row * imgf->width + col + 1) * depth + 3];
		alpha2 = imgf->buf[((row + 1) * imgf->width + col) * depth + 3];
		alpha3 = imgf->buf[((row + 1) * imgf->width + col + 1) * depth + 3];
	}
	ilmout.x = ((1 - u) * P0.x + u * P1.x) * (1 - v) + ((1 - u) * P2.x + u * P3.x) * v;
	ilmout.y = ((1 - u) * P0.y + u * P1.y) * (1 - v) + ((1 - u) * P2.y + u * P3.y) * v;
	ilmout.z = ((1 - u) * P0.z + u * P1.z) * (1 - v) + ((1 - u) * P2.z + u * P3.z) * v;
	if (alpha)
		*alpha = ((1 - u) * alpha0 + u * alpha1) * (1 - v) + ((1 - u) * alpha2 + u * alpha3) * v;
	return ilmout;
}

/* get the ambient occlusion light */
struct elmVEC ambient(struct elmSCENE *sce, struct elmRAY *ray)
{
	/* map on the cube */
	if (sce->amb.map.width) {

		double cx2 = sce->amb.map.width / 8.0, cy2 = sce->amb.map.height / 6.0F;
		double xm, ym;

		if (fabs(ray->dir.x) > fabs(ray->dir.y) && fabs(ray->dir.x) > fabs(ray->dir.z)) {
			if (ray->dir.x > 0) {
				/* facing front */
				double xo = 0.375 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/*ilmout.r = 0.5F; ilmout.g = 0; ilmout.b = 0;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.x;
				ym = yo + cy2 * ray->dir.z / ray->dir.x;
			} else {
				double xo = 0.875 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/* facing back */
				/*ilmout.r = 0.5F; ilmout.g = 0.5F; ilmout.b = 0;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.x;
				ym = yo - cy2 * ray->dir.z / ray->dir.x;
			}
		} else if (fabs(ray->dir.y) > fabs(ray->dir.x) && fabs(ray->dir.y) > fabs(ray->dir.z)) {
			if (ray->dir.y > 0) {
				double xo = 0.125 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/* facing left */
				/*ilmout.r = 0; ilmout.g = 0.5F; ilmout.b = 0;*/
				xm = xo + cx2 * ray->dir.x / ray->dir.y;
				ym = yo + cy2 * ray->dir.z / ray->dir.y;
			} else {
				/* facing right */
				double xo = 0.625 * sce->amb.map.width, yo = 0.5 * sce->amb.map.height;
				/*ilmout.r = 0; ilmout.g = 0.5F; ilmout.b = 0.5F;*/
				xm = xo + cx2 * ray->dir.x / ray->dir.y;
				ym = yo - cy2 * ray->dir.z / ray->dir.y;
			}
		} else if (fabs(ray->dir.z) > fabs(ray->dir.x) && fabs(ray->dir.z) > fabs(ray->dir.y)) {
			if (ray->dir.z > 0) {
				double xo = 0.375 * sce->amb.map.width, yo = 0.83333333 * sce->amb.map.height;
				/* facing top */
				/*ilmout.r = 0; ilmout.g = 0; ilmout.b = 0.5F;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.z;
				ym = yo - cy2 * ray->dir.x / ray->dir.z;
			} else {
				double xo = 0.375 * sce->amb.map.width, yo = 0.16666666 * sce->amb.map.height;
				/* facing bottom */
				/*ilmout.r = 0.5F; ilmout.g = 0; ilmout.b = 0.5F;*/
				xm = xo - cx2 * ray->dir.y / ray->dir.z;
				ym = yo + cy2 * ray->dir.x / ray->dir.z;
			}
		} 

		return map_texture(&sce->amb.map, xm, ym, NULL);

	} else {

		/* only ambient color */
		return sce->amb.clr;
	}
}

// radiance along ray
struct elmVEC radiance(struct elmRAY ray, struct elmSCENE *sce, int depth, pcg32_random_t* rng)
{
	double t; // distance to intersection
	int id = 0; // id of intersected object
	struct elmVEC hitnor = {0}, hitpoint = {0}; // object normal and hit point

	if (!intersect_scene(&ray, sce, &t, &id, &hitpoint, &hitnor))

		// if miss, return ambient
		return depth ? vec_times(ambient(sce, &ray), sce->amb.str) : ambient(sce, &ray); 

	else {

		struct elmOBJ *obj = &sce->obj[id]; // the hit object
		double n_dot_raydir = vec_dot(hitnor, ray.dir);
		struct elmVEC nl = n_dot_raydir < 0 ? hitnor : vec_times(hitnor, -1); // properly oriented surface normal
        struct elmVEC objclr = obj->mat.diffFunc ? obj->mat.diffFunc(&hitpoint, obj) : obj->mat.diffclr; /* object color */
		double p = vec_max(objclr); /* use maximum reflectivity amount for Russian roulette */
		struct elmVEC objemiclr = n_dot_raydir < 0 ? (obj->mat.emiFunc ? obj->mat.emiFunc(&hitpoint, obj) : obj->mat.emiclr) : elmvec(0, 0, 0); // emits only in the normal direction
		int objmatrefl = obj->mat.comp[0] == 1 ? elmDIFF : (obj->mat.comp[1] == 1 ? elmSPEC : (obj->mat.comp[2] == 1 ? elmREFR : int_rand(rng, obj->mat.comp, 3))); // decide with roulette
				
		// too deep recursion
		if (depth > 15)
			// return current emission
			return objemiclr; 

		// max refl with Russian roulette
		if (++depth > 5)
			if (double_rand(rng) < p)
				objclr = vec_times(objclr, 1 / p);
			else
				return objemiclr;

		/* depends on object material probability */
		if (objmatrefl == elmDIFF) {

			// Ideal DIFFUSE reflection: generate random point on the unit hemisphere

			// random azimuth angle
			double r1 = 2 * M_PI * double_rand(rng);

			// random distance from center
			double r2 = double_rand(rng), r2s = sqrt(r2);

			// the z axis is the normal
			struct elmVEC w = nl;

			// the u axis is perpendicular to w and is normalized
			struct elmVEC u = vec_norm(vec_cross(fabs(w.x) > .1 ? elmvec(0, 1, 0) : elmvec(1, 0, 0), w)); 

			// v is perpendicular to v and w and is normalized
			struct elmVEC v = vec_cross(w, u); 

			// move along u then along v then along W (hit the unit hemisphere)
			// and obtain a random point rdir on it
			struct elmVEC rdir = vec_norm(vec_plus(vec_times(u, cos(r1) * r2s), vec_plus(vec_times(v, sin(r1) * r2s), vec_times(w, sqrt(1 - r2))))); // d is a random reflection ray

			// go onwards by a small amount
			//struct elmVEC hitpoint2 = vec_plus(hitpoint, vec_times(rdir, 0.001));
			
			// find radiance on this random direction, weight by surface color, and add emissivity
			return vec_plus(objemiclr, vec_mult(objclr, radiance(elmray(hitpoint, rdir, id), sce, depth, rng)));

		} else if (objmatrefl == elmSPEC) {

			// Ideal SPECULAR reflection

			// reflected ray
			struct elmRAY refray = elmray(hitpoint, vec_minus(ray.dir, vec_times(hitnor, 2 * n_dot_raydir)), id);

			// return emissivity plus reflected radiance weighted by specular color
			return vec_plus(objemiclr, vec_mult(objclr, radiance(refray, sce, depth, rng)));
				
		} else  if (objmatrefl == elmREFR) {

			/* REFRACTION on dielectric surface - glass */
			struct elmRAY reflRay = elmray(hitpoint, vec_minus(ray.dir, vec_times(hitnor, 2 * n_dot_raydir)), id);     
			int into = vec_dot(hitnor, nl) > 0; /* Ray from outside going in? */
			double nc = 1, nt = 1.5, nnt = into ? nc / nt : nt / nc, ddn = vec_dot(ray.dir, nl), cos2t;
			
			/* if total internal reflection, reflect */
			if ((cos2t = 1 - nnt * nnt * (1 - ddn * ddn)) < 0) {

				/* Total internal reflection */
				return vec_plus(objemiclr, vec_mult(objclr, radiance(reflRay, sce, depth, rng)));

			} else {

				/* choose reflection or refraction */
				struct elmVEC tdir = vec_norm(vec_minus(vec_times(ray.dir, nnt), vec_times(hitnor, (into ? 1 : -1) * (ddn * nnt + sqrt(cos2t)))));
				double a = nt - nc, b = nt + nc, R0 = a * a / (b * b), c = 1 - (into ? -ddn : vec_dot(tdir, hitnor));
				double Re = R0 + (1 - R0) * c * c * c * c * c, Tr = 1 - Re, P = .25 + .5 * Re,
					RP = Re / P, TP = Tr / (1 - P);
				
				// Russian roulette
				return vec_plus(objemiclr, vec_mult(objclr, depth > 2 ? (double_rand(rng) < P ? 
					vec_times(radiance(reflRay, sce, depth, rng), RP) : vec_times(radiance(elmray(hitpoint, tdir, id), sce, depth, rng), TP)) :
					vec_plus(vec_times(radiance(reflRay, sce, depth, rng), Re), vec_times(radiance(elmray(hitpoint, tdir, id), sce, depth, rng), Tr))));
			}
		
		} else

			return elmvec(0,0,0);
	}
}

/* exposure correction function */
float apply_exposure(float l)
{
	float expo = -1.00f;
	return (1.0f - expf(l * expo));
}

/* encoding to srgb */
float srgb_encode(float c)
{
	if (c <= 0.0031308f) 
		return 12.92f * c;
	else
		return 1.055f * powf(c, 0.4166667f) - 0.055f; // Inverse gamma 2.4
}

/* decoding from srgb */
float srgb_decode(float c)
{
	if (c <= 0.04045f) 
		return c * 0.07739938F;
	else
		return powf((c + 0.055F) * 0.947867F, 2.4F); // Direct gamma 2.4
}

/* clip float to unsigned char */
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))

/* encode a pixel */
unsigned char encode_pixel(float pixval)
{
	// exposure correction
	float pixvalec = apply_exposure(pixval);

	// gamma correction
	float pixvalge = srgb_encode(pixvalec);

	// ready for file
	unsigned char pixvaluc = FLOAT2UCHAR(pixvalge);

	return pixvaluc;
}

struct threadenv {
	int nt, startcol, stepcol;
	int *die;
	struct elmSCENE *sce;
	int samps;
};

/* explore outer space with rays */
void threadproc(void *env)
{
	int y, x, s;
	struct threadenv *te = (struct threadenv *) env;
	pcg32_random_t rng = {time(NULL) ^ (intptr_t) &printf, (intptr_t) env};

	fprintf(stderr, "Thread %d cols %d,%d,%d,...\n", te->nt,
		te->startcol, te->startcol + te->stepcol, te->startcol + 2 * te->stepcol);

	// loop samples
	for (s = 0; s < te->samps && !*(te->die); s++) { 

		// Loop over image rows (top-bottom)
		for (y = 0; y < te->sce->cam.imgf.height; y++) { 

			// Loop cols (left-right)
			for (x = te->startcol; x < te->sce->cam.imgf.width; x += te->stepcol) {  

				int sx, sy, ns = 0;
				struct elmVEC r = {0,0,0}; // colors of smples

				// 2x2 subpixel rows
				for (sy = 0; sy < 2; sy++) { 

					// 2x2 subpixel cols
					for (sx = 0; sx < 2; sx++) { 

						/* random displacement around subpixel center */
						double r1 = 2 * double_rand(&rng), dx = r1 < 1 ? sqrt(r1) - 1 : 1 - sqrt(2 - r1);
						double r2 = 2 * double_rand(&rng), dy = r2 < 1 ? sqrt(r2) - 1 : 1 - sqrt(2 - r2);

						/* image plane pixel coord */
						struct elmVEC pixcoord = vec_plus(vec_plus(te->sce->cam.ipo, vec_times(te->sce->cam.cx, (sx + .5 + dx) / 2 + x)),
							vec_times(te->sce->cam.cy, (sy + .5 + dy) / 2 + y));

						/* primary ray direction */
						struct elmVEC pdir = vec_norm(vec_minus(pixcoord, te->sce->cam.view.orig));

						/* primary ray origin (out of camera) */
						struct elmVEC porig = vec_plus(te->sce->cam.view.orig, pdir);

						/* change if dof active */
						if (te->sce->cam.fdist) {

							/* eye-pixel distance */
							double pedist = vec_length(vec_minus(pixcoord, te->sce->cam.view.orig));
							
							/* focal point */
							struct elmVEC fp = vec_plus(te->sce->cam.view.orig, vec_times(pdir, pedist * te->sce->cam.fdist / te->sce->cam.fl));

							/* new origin */
							double ap = te->sce->cam.imgf.width;
							double rr1 = double_rand(&rng) * ap - ap / 2;
							double rr2 = double_rand(&rng) * ap - ap / 2;
							porig = vec_plus(vec_plus(pixcoord, vec_times(te->sce->cam.cx, rr1)), vec_times(te->sce->cam.cy, rr2));

							/* new direction */
							pdir = vec_norm(vec_minus(fp, porig));
						}

						/*if (y == 123 && x == 212)
							printf("Azz\n");*/

						/* accumulate radiance in that pixel */
						r = vec_plus(r, radiance(elmray(porig, pdir, -1), te->sce, 0, &rng)); 	

						/* count samples done */
						ns++;
					}
				}

				// accumulate radiance on this point
				te->sce->cam.imgf.buf[(y * te->sce->cam.imgf.width + x) * 3 + 0] += (float) r.x;
				te->sce->cam.imgf.buf[(y * te->sce->cam.imgf.width + x) * 3 + 1] += (float) r.y;
				te->sce->cam.imgf.buf[(y * te->sce->cam.imgf.width + x) * 3 + 2] += (float) r.z;
				te->sce->cam.imgf.samps[y * te->sce->cam.imgf.width + x] += ns;
			}
		}
	}

	fprintf(stderr, "Thread %d exiting\n", te->nt);
}

/* add plane */
int add_plane(struct elmSCENE *sce, struct elmVEC pos, struct elmVEC nor, struct elmVEC ori, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMPLANE;
	cobj->pla.pos = pos;
	cobj->pla.nor = vec_norm(nor);
	cobj->pla.ori = vec_norm(ori);
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->aabb.vmin = elmvec(vec_dot(cobj->pla.nor, elmvec(1,0,0)) == 1 ? 0 : -Inf,
		vec_dot(cobj->pla.nor, elmvec(0,1,0)) == 1 ? 0 : -Inf, vec_dot(cobj->pla.nor, elmvec(0,0,1)) == 1 ? 0 : -Inf);
	cobj->aabb.vmax = elmvec(vec_dot(cobj->pla.nor, elmvec(1,0,0)) == 1 ? 0 : +Inf,
		vec_dot(cobj->pla.nor, elmvec(0,1,0)) == 1 ? 0 : +Inf, vec_dot(cobj->pla.nor, elmvec(0,0,1)) == 1 ? 0 : +Inf);
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* add sphere */
int add_sphere(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMSPHERE;
	cobj->sph.rad = rad;
	cobj->sph.pos = pos;
	cobj->sph.nor = nor;
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->aabb.vmin = vec_minus(cobj->sph.pos, elmvec(cobj->sph.rad, cobj->sph.rad, cobj->sph.rad));
	cobj->aabb.vmax = vec_plus(cobj->sph.pos, elmvec(cobj->sph.rad, cobj->sph.rad, cobj->sph.rad));
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* add disc */
int add_disc(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMDISC;
	cobj->dsc.rad = rad;
	cobj->dsc.pos = pos;
	cobj->dsc.nor = vec_norm(nor);
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* add cylinder */
int add_cylinder(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, double height, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMCYLINDER;
	cobj->cyl.rad = rad;
	cobj->cyl.pos = pos;
	cobj->cyl.nor = vec_norm(nor);
	cobj->cyl.hght = height;
	cobj->cyl.pos2 = vec_plus(cobj->cyl.pos, vec_times(cobj->cyl.nor, cobj->cyl.hght));
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* add cone */
int add_cone(struct elmSCENE *sce, double rad, struct elmVEC pos, struct elmVEC nor, double height, double rad2, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMCONE;
	cobj->con.rad = rad;
	cobj->con.pos = pos;
	cobj->con.nor = vec_norm(nor);
	cobj->con.hght = height;
	cobj->con.rad2 = rad2;
	cobj->con.pos2 = vec_plus(cobj->con.pos, vec_times(cobj->con.nor, cobj->con.hght));
	cobj->con.apx = vec_plus(cobj->con.pos, vec_times(vec_minus(cobj->con.pos2, cobj->con.pos), cobj->con.rad / (cobj->con.rad - cobj->con.rad2)));
	cobj->con.tga = (cobj->con.rad - cobj->con.rad2) / cobj->con.hght;
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* add triangle */
int add_triangle(struct elmSCENE *sce, struct elmVEC pos, struct elmVEC pos2, struct elmVEC pos3, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMTRIANGLE;
	cobj->tri.pos = pos;
	cobj->tri.pos2 = pos2;
	cobj->tri.pos3 = pos3;
	cobj->tri.nor = vec_norm(vec_cross(vec_minus(cobj->tri.pos2, cobj->tri.pos), vec_minus(cobj->tri.pos3, cobj->tri.pos)));
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* add quadrangle */
int add_quadrangle(struct elmSCENE *sce, struct elmVEC pos, struct elmVEC pos2, struct elmVEC pos3, struct elmMAT mat, char name[32])
{
	struct elmOBJ *cobj = NULL;
	sce->obj = realloc(sce->obj, ++sce->nobjs * sizeof(struct elmOBJ));
	cobj = &sce->obj[sce->nobjs - 1];
	memset(cobj, 0, sizeof(struct elmOBJ));
	cobj->kind = ELMQUADRANGLE;
	cobj->quad.pos = pos;
	cobj->quad.pos2 = pos2;
	cobj->quad.pos3 = pos3;
	cobj->quad.pos4 = vec_minus(cobj->quad.pos3, vec_minus(cobj->quad.pos2, cobj->quad.pos));
	cobj->quad.nor = vec_norm(vec_cross(vec_minus(cobj->quad.pos2, cobj->quad.pos), vec_minus(cobj->quad.pos3, cobj->quad.pos)));
	cobj->mat = mat;
	strncpy(cobj->name, name, 32);
	cobj->id = sce->nobjs - 1;
	return sce->nobjs - 1;
}

/* prepare camera */
int prepare_camera(struct elmCAM *cam, struct elmVEC pos, struct elmVEC aim, double hfov, int width, int height)
{
	struct elmVEC ipc;
	double ipsw;

	/* prepare image */
	cam->imgf.width = width;
	cam->imgf.height = height;
	cam->imgf.buf = (float *) calloc(cam->imgf.height * cam->imgf.width * 3, sizeof(float));
	cam->imgf.samps = (int *) calloc(cam->imgf.height * cam->imgf.width, sizeof(int));

	// assign camera data
	cam->view = elmray(pos, vec_norm(vec_minus(aim, pos)), -1); // camera pos, d
	cam->hfov = hfov;
	cam->fdist = vec_length(vec_minus(aim, pos));

	// focal length
	cam->fl = cos(0.5 * hfov * M_PI / 180);

	// semi-width of image plane
	ipsw = tan(0.5 * hfov * M_PI / 180);

	// image plane center
	ipc = vec_plus(cam->view.orig, vec_times(cam->view.dir, cam->fl));

	// image plane horizontal director
	cam->cx = vec_times(vec_norm(vec_cross(elmvec(cam->view.dir.x, cam->view.dir.y, 0), elmvec(0, 0, 1))), 2 * ipsw / cam->imgf.width);

	// image plane vertical director
	cam->cy = vec_times(vec_norm(vec_cross(cam->cx, cam->view.dir)), -(2 * ipsw / cam->imgf.width));

	// image plane origin
	cam->ipo = vec_plus(ipc, vec_plus(vec_times(cam->cx, -cam->imgf.width / 2.0), vec_times(cam->cy, -cam->imgf.height / 2.0)));

	return 0;
}

/* populate scene */
struct elmIMGF logof = {0};
struct elmVEC floorClr(struct elmVEC *p, struct elmOBJ *obj)
{
	struct elmPLA *pla = &obj->pla;
	struct elmVEC clr0 = {0.9,0.9,0.9}, clr1 = {0.4,0.4,0.4};
	struct elmVEC x = vec_minus(*p, pla->pos);
	double newx = vec_dot(x, pla->ori);
	double newy = vec_dot(x, vec_cross(pla->nor, pla->ori));
	struct elmVEC flclr = (int) newx & 0x00000008 ^ (int) newy & 0x00000008 ? clr1 : clr0;
	double logow1 = -20, logow2 = +20, logoh1 = -20, logoh2 = +20; // logo position
	if (logof.width && newx >= logow1 && newx <= logow2 && newy >= logoh1 && newy <= logoh2) {
		double alpha; // transparency channel
		struct elmVEC logoclr = map_texture(&logof, (newx - logow1) * logof.width / (logow2 - logow1), (newy - logoh1) * logof.height / (logoh2 - logoh1), &alpha);
		return vec_plus(vec_times(flclr, 1 - alpha), vec_times(logoclr, alpha));
	} else
		return flclr;
}
int prepare_scene(struct elmSCENE *sce)
{
	// load logo
	if (1) {
		struct elmIMG logo = {0};
		int row, col;
		img_read("Unipg.png", &logo);
		logof.width = logo.width;
		logof.height = logo.height;
		logof.buf = calloc(logof.width * logof.height * 4, sizeof(float));
		for (row = 0; row < logo.height; row++) {
			for (col = 0; col < logo.width; col++) {
				logof.buf[(row * logof.width + col) * 4 + 0] = srgb_decode((float) logo.buffer[row * logo.stride + col * logo.depth * logo.bytes + 0] / 255.0F);
				logof.buf[(row * logof.width + col) * 4 + 1] = srgb_decode((float) logo.buffer[row * logo.stride + col * logo.depth * logo.bytes + 1] / 255.0F);
				logof.buf[(row * logof.width + col) * 4 + 2] = srgb_decode((float) logo.buffer[row * logo.stride + col * logo.depth * logo.bytes + 2] / 255.0F);
				logof.buf[(row * logof.width + col) * 4 + 3] = (float) logo.buffer[row * logo.stride + col * logo.depth * logo.bytes + 3] / 255.0F;
			}
		}
		free(logo.buffer);
	}

	// camera
	prepare_camera(&sce->cam, elmvec(-90,0,45), elmvec(5,0,0), 54, 640, 480);

	// plane hovered by light
    sce->obj[add_plane(sce, elmvec(-20,0,0), elmvec(0,0,1), elmvec(0.5,-1,0), elmmat(elmvec(0,0,0), elmvec(.67,.56,.56), elmvec(.67,.56,.56), 1, 0, 0), "Floor plane")].mat.diffFunc = &floorClr;
	add_sphere(sce, 12, elmvec(0,0,12), elmvec(0,0,1), elmmat(elmvec(0,0,0), elmvec(.32,.32,.32), elmvec(.32,.32,.32), 1, 0, 0), "Ball");
	//add_sphere(sce, 8, elmvec(-60,-30,62), elmvec(0,0,1), elmmat(elmvec(50,50,50), elmvec(.32,.32,.32), elmDIFF), "Light 1");
	//add_sphere(sce, 7, elmvec(-200,5,85), elmvec(0,0,1), elmmat(elmvec(60,60,50), elmvec(.32,.32,.32), elmDIFF), "Light 2");
	add_sphere(sce, 10, elmvec(-20,25,10), elmvec(0,0,1), elmmat(elmvec(0,0,0), elmvec(.90,.90,.90), elmvec(.90,.90,.90), 0, 1, 0), "Mirror");
	add_sphere(sce, 10, elmvec(-20,-25,10), elmvec(0,0,1), elmmat(elmvec(0,0,0), elmvec(.93,.93,.98), elmvec(.93,.93,.98), 0, 0, 1), "Glass");
	add_cylinder(sce, 5, elmvec(-40,4,0), elmvec(0,0,1), 8, elmmat(elmvec(0,0,0), elmvec(.98,.98,.90), elmvec(.98,.98,.90), 0, 0, 1), "Cylinder");
	add_cone(sce, 5, elmvec(-40,-10,0), elmvec(0,0,1), 16, 0, elmmat(elmvec(0,0,0), elmvec(.32,.32,.32), elmvec(.32,.32,.32), 1, 0, 0), "Cone");
	add_disc(sce, 2.5, elmvec(-38,15,2.5), elmvec(-1,0,0), elmmat(elmvec(1,5,1), elmvec(.4,.4,.5), elmvec(.4,.4,.5), 1, 0, 0), "Disc");
	add_triangle(sce, elmvec(-40,22,0), elmvec(-40,26,5), elmvec(-40,26,0), elmmat(elmvec(5,1,1), elmvec(.76,.15,.13), elmvec(.76,.15,.13), 1, 0, 0), "Triangle");
	add_quadrangle(sce, elmvec(-35,22,0), elmvec(-35,22,5), elmvec(-35,26,5), elmmat(elmvec(1,1,5), elmvec(.76,.15,.13), elmvec(.76,.15,.13), 1, 0, 0), "Rectangle");

	// ambient occlusion
	sce->amb.clr = elmvec(0.3,0.3,0.5);
	sce->amb.str = 0.99;

	// load the sky
	if (1) {

		struct elmIMG sky = {0};
		int row, col;
		char ambientfile[] = "skyboxsun25degtest.png";
		//char ambientfile[] = "skybox_texture.bmp";
		//char ambientfile[] = "grimmnight_large.bmp";

		fprintf(stderr, "Loading ambient \"%s\"...", ambientfile);
		img_read(ambientfile, &sky);
		sce->amb.map.width = sky.width;
		sce->amb.map.height = sky.height;
		sce->amb.map.buf = calloc(sky.width * sky.height * 3, sizeof(float));
		for (row = 0; row < sky.height; row++) {
			for (col = 0; col < sky.width; col++) {
				float P0 = srgb_decode((float) sky.buffer[row * sky.stride + col * sky.depth * sky.bytes + 0] / 255.0F);
				float P1 = srgb_decode((float) sky.buffer[row * sky.stride + col * sky.depth * sky.bytes + 1] / 255.0F);
				float P2 = srgb_decode((float) sky.buffer[row * sky.stride + col * sky.depth * sky.bytes + 2] / 255.0F);
				float thr = 1, gain = 60000.0F;
				if (P0 > thr && P1 > thr && P2 > thr) {
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 0] = (P0 - thr) * gain + thr;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 1] = (P1 - thr) * gain + thr;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 2] = (P2 - thr) * gain + thr;
				} else {
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 0] = P0;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 1] = P1;
					sce->amb.map.buf[(row * sce->amb.map.width + col) * 3 + 2] = P2;
				}
			}
		}
		free(sky.buffer);
		fprintf(stderr, " done\n");
	}

	return 1;
}

int prepare_scene_3diff(struct elmSCENE *sce)
{
	int r;

	// camera
	prepare_camera(&sce->cam, elmvec(-90,0,50), elmvec(0,0,15), 48, 640, 360);

	// plane hovered by light
	add_quadrangle(sce, elmvec(200,-200,120), elmvec(150,-200,120), elmvec(150,-170,220), elmmat(elmvec(200,200,200), elmvec(.76,.15,.13), elmvec(.76,.15,.13), 1, 0, 0), "Rectangle Light");
	add_plane(sce, elmvec(-20,0,0), elmvec(0,0,1), elmvec(0.5,-1,0), elmmat(elmvec(0,0,0), elmvec(.22,.22,.22), elmvec(.22,.22,.22), 1, 0, 0), "Floor plane"); 
	for (r = 0; r < 1; r++) {
		add_cone(sce, 12, elmvec(r * 28,28 - r * r,0), elmvec(0,0,1), 24, 0, elmmat(elmvec(0,0,0), elmvec(.95,.95,.90), elmvec(.95,.95,.90), 1, 0, 0), "Cone");
		add_sphere(sce, 12, elmvec(r * 28,0 - r * r,12), elmvec(0,0,1), elmmat(elmvec(0,0,0), elmvec(.95,.95,.90), elmvec(.95,.95,.90), 1, 0, 0), "Ball");
		add_cylinder(sce, 10, elmvec(r * 28,-28 - r * r,0), elmvec(0,0,1), 22, elmmat(elmvec(0,0,0), elmvec(.95,.95,.90), elmvec(.95,.95,.90), 1, 0, 0), "Cylinder");
	}

	// ambient occlusion
	sce->amb.clr = elmvec(0.8,0.8,0.8);
	sce->amb.str = 0.99;

	return 0;
}

/* dump to output file */
int image_dump(char *filename, struct elmIMGF *imgf)
{
	int x, y;
	struct elmIMGF imgf2 = *imgf;
	struct elmIMG img = {imgf->width, imgf->height, 3, 255};
	img.bytes = img.maxval > 255 ? 2 : 1;
	img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
	img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

	imgf2.buf = (float *) calloc(imgf2.height * imgf2.width * 3, sizeof(float));
	memcpy(imgf2.buf, imgf->buf, imgf2.height * imgf2.width * 3 * sizeof(float));
	//post_process(&imgf2, &imgf);

	for (y = 0; y < imgf->height; y++) {
		for (x = 0; x < imgf->width; x++) {
			img.buffer[y * img.stride + x * img.depth * img.bytes + 0] =
				encode_pixel(imgf2.buf[(y * imgf2.width + x) * 3 + 0] / imgf2.samps[y * imgf2.width + x]);
			img.buffer[y * img.stride + x * img.depth * img.bytes + 1] =
				encode_pixel(imgf2.buf[(y * imgf2.width + x) * 3 + 1] / imgf2.samps[y * imgf2.width + x]);
			img.buffer[y * img.stride + x * img.depth * img.bytes + 2] =
				encode_pixel(imgf2.buf[(y * imgf2.width + x) * 3 + 2] / imgf2.samps[y * imgf2.width + x]);
		}
	}

	free(imgf2.buf);

	/* dump image */
	img_write(filename, &img);

	/* free image */
	free(img.buffer);
	memset(&img, 0, sizeof(struct elmIMG));

	return 0;
}

/* main function */
int main(int argc, char **argv)
{
	int y, nt, die = 0;
	int numthreads = 6;
	struct elmSCENE scene = {0};
	int samps = 16384; // # samples
	struct threadenv *te = NULL;
	HANDLE *hands = NULL;
	double starttime = timeGetTime();
	char imagename[] = "image.png";

	/* prepare scene */
	prepare_scene_3diff(&scene);
	//prepare_scene(&scene);

	/* start threads */
	te = calloc(numthreads, sizeof(struct threadenv));
	hands = calloc(numthreads, sizeof(HANDLE));
	for (nt = 0; nt < numthreads; nt++) {
		te[nt].nt = nt;
		te[nt].startcol = nt;
		te[nt].stepcol = numthreads;
		te[nt].die = &die;
		te[nt].sce = &scene;
		te[nt].samps = samps;
		hands[nt] = (HANDLE) _beginthread(threadproc, 0, &te[nt]);
	}

	do {
		int sleepms = 2000;
		int KeyInfo = _kbhit() ? _getch() : 0;
		double currsamples = 0;
		static int dumpimg = 1, forcedump = 0;
		static double currtime, nextdumptime = 0;
		double totsamples = (double) scene.cam.imgf.width * scene.cam.imgf.height * 4 * samps;
		static double lastsamples = 0;

		currtime = timeGetTime();
		Sleep(sleepms);

		if (tolower(KeyInfo) == 'd')
			dumpimg = !dumpimg;
		else if (tolower(KeyInfo) == 'k')
			die = 1;
		else if (tolower(KeyInfo) == 'f')
			forcedump = 1;

		/* count samples done */
		for (y = 0; y < scene.cam.imgf.height * scene.cam.imgf.width; y++)
			currsamples += scene.cam.imgf.samps[y];

		/* write to file */
		if ((dumpimg && timeGetTime() >= nextdumptime) || forcedump) {
			double nowtime = timeGetTime();
			double remtime = totsamples / (currsamples / (nowtime - starttime)) - (nowtime - starttime);
			image_dump(imagename, &scene.cam.imgf);
			fprintf(stderr, "image '%s' dumped (E%02dh%02dm%02ds, R%02dh%02dm%02ds)\n", imagename,
				(int) floor((nowtime - starttime) / 3600000), (int) floor((nowtime - starttime) / 60000) % 60, (int) ((nowtime - starttime) / 1000) % 60,
				(int) floor(remtime / 3600000), (int) floor(remtime / 60000) % 60, (int) (remtime / 1000) % 60);
			nextdumptime = nowtime + 30000;
			forcedump = 0;
		}

		fprintf(stderr, "%.1f krays/s (%.1f spp)\n", (currsamples - lastsamples) / (timeGetTime() - currtime), currsamples / (scene.cam.imgf.height * scene.cam.imgf.width));
		lastsamples = currsamples;

	} while (!die);

	/* wait all threads */
	WaitForMultipleObjects(numthreads, hands, TRUE, INFINITE);

	/* save safely to disk */
	image_dump(imagename, &scene.cam.imgf);
	fprintf(stderr, "image dumped\n");

	/* free image, scene, and all */
	free(scene.cam.imgf.buf);
	free(scene.cam.imgf.samps);
	free(te);
	free(scene.obj);
	free(scene.amb.map.buf);
	free(hands);
	free(logof.buf);

	return EXIT_SUCCESS;
}
