/* a basic ray-tracer */
/* http://www.scratchapixel.com/lessons/3d-basic-rendering/introduction-to-ray-tracing/implementing-the-raytracing-algorithm */
/* https://en.wikipedia.org/wiki/Ray_tracing_(graphics) */

#define _USE_MATH_DEFINES
#include <math.h>
#include <time.h>
#include "elabmm.h"

extern const float NaNf = ((float)((1e308 * 10)*0.));
extern const float Inff = ((float)(1e308 * 10));

#define DEG2RAD(A) ((float) (A) * (float) M_PI / 180.0F)
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))

struct elmCOORD {
	float x, y, z;
};

struct elmCLR {
	float r, g, b;
};

struct elmMTR {
	struct elmCLR sc; /* basic color */
	float rr; /* reflective factor */
	float rf; /* refractive factor */
	float em; /* emissive facor */
};

struct elmROTM {
	float xx, xy, xz;
	float yx, yy, yz;
	float zx, zy, zz;
};

struct elmRAY {
	struct elmCOORD s;
	struct elmCOORD d;
};

struct elmCAM {
	struct elmCOORD o;
	float az, el;
	float fov;
	float ar;
	float fl;
	struct {
		float hs, vs, dx, dy;
	} wk;
};

struct elmPOLY {
	struct elmCOORD v1, v2, v3;
	struct elmMTR sm;
};

struct elmSPHE {
	struct elmCOORD c;
	float r;
	struct elmMTR sm;
};

struct elmOBJ {
	enum {ELMP, ELMS} t;
	union {
		struct elmPOLY p;
		struct elmSPHE s;
	} o;
	struct elmMTR *sm;
};

struct elmIMGF {
	int width, height;
	float *buf;
};

struct elmCLR skycolor = {0.25F,0.25F,0.3F};

struct elmCOORD versor(struct elmCOORD *v1, struct elmCOORD *v2)
{
	/* versor from v1 towards v2 */
	struct elmCOORD d = {v2->x - v1->x, v2->y - v1->y, v2->z - v1->z};
	float len = sqrtf(d.x * d.x + d.y * d.y + d.z * d.z);
	d.x /= len;
	d.y /= len;
	d.z /= len;
	return d;
}

int prepare_cam(struct elmCAM *cam, struct elmIMGF *imgf)
{
	/* camera size on the unit circle */
	cam->wk.hs = cam->fl * 2.0F * tanf(cam->fov / 2.0F);
	cam->wk.dx = cam->wk.hs / (float) imgf->width;
	cam->wk.dy = cam->wk.dx;
	cam->wk.vs = cam->wk.hs / cam->ar;

	return 0;
}

int primary_ray(struct elmRAY *pray, struct elmCAM *cam, int row, int col, int dorandom)
{
	struct elmCOORD orig = {0, 0, 0};
	struct elmCOORD pixelcoord = {
		cam->fl * cosf(cam->fov / 2.0F),
		cam->wk.hs / 2.0F - cam->wk.dx * (col + dorandom * ((float) rand() / (float) RAND_MAX)),
		cam->wk.vs / 2.0F - cam->wk.dy * (row + dorandom * ((float) rand() / (float) RAND_MAX))
	};

	/* origin of ray is the cam */
	memcpy(&pray->s, &cam->o, sizeof(struct elmCOORD));

	/* direction versor */
	pray->d = versor(&orig, &pixelcoord);

	/* rotate around z-axis according to azimut */
	pray->d.x = pray->d.x * cosf(cam->az) - pray->d.y * sinf(cam->az);
	pray->d.y = pray->d.x * sinf(cam->az) + pray->d.y * cosf(cam->az);

	/* rotate around new x-axis according to elevation */
	pray->d.x = -pray->d.z * sinf(cam->el) + pray->d.x * cosf(cam->el);
	pray->d.z =  pray->d.z * cosf(cam->el) + pray->d.x * sinf(cam->el);

	return 0;
}

float intersection(struct elmCOORD *hitp, struct elmCOORD *hitn, struct elmOBJ *obj, struct elmRAY *pray) 
{
	if (obj->t == ELMS) {

		/* sphere */
		struct elmSPHE *sphe = &(obj->o.s);
		struct elmCOORD v = {pray->s.x - sphe->c.x, pray->s.y - sphe->c.y, pray->s.z - sphe->c.z};
		float vd = (v.x * pray->d.x + v.y * pray->d.y + v.z * pray->d.z);
		float disc = vd * vd - (v.x * v.x + v.y * v.y + v.z * v.z) + sphe->r * sphe->r;
		if (disc < 0)
			return -1;
		else {
			float discs = sqrtf(disc);
			float t1 = - vd + discs, t2 = - vd - discs;
			if (t2 >= 0) {
				if (hitp) {
					hitp->x = pray->s.x + t2 * pray->d.x; 
					hitp->y = pray->s.y + t2 * pray->d.y; 
					hitp->z = pray->s.z + t2 * pray->d.z; 
					if (hitn)
						*hitn = versor(&sphe->c, hitp);
				}
				return t2;
			} else if (t1 >= 0) {
				if (hitp) {
					hitp->x = pray->s.x + t1 * pray->d.x; 
					hitp->y = pray->s.y + t1 * pray->d.y; 
					hitp->z = pray->s.z + t1 * pray->d.z; 
					if (hitn)
						*hitn = versor(&sphe->c, hitp);
				}
				return t1;
			} else
				return -1;
		}

	} else if (obj->t == ELMP) {

		/* polygon */
		return -1;
	}

	return -1;
}

struct elmCLR trace(struct elmRAY *ray, struct elmOBJ *obj, int nobj)
{
	struct elmCLR ilmout;
	int oo;
	float tol = 0.001F;

	/* trace ray */
	float mindist = +Inff;
	struct elmOBJ *cobj = NULL;
	struct elmCOORD hitp, hitn, chitp, chitn;

	/* search all objects */
	for (oo = 0; oo < nobj; oo++) {

		/* find intersection if any (not with lights) */
		float dist;
		if (obj[oo].sm->em == 0 && (dist = intersection(&hitp, &hitn, &obj[oo], ray)) >= 0) {

			/* save nearest object */
			if (dist < mindist) {
				mindist = dist;
				cobj = &obj[oo];
				chitp = hitp;
				chitn = hitn;
			}
		}
	}

	/* from far object */
	if (cobj) {

		struct elmCLR ill = {0,0,0};

		/* object color */
		ilmout.r = cobj->sm->sc.r;
		ilmout.g = cobj->sm->sc.g;
		ilmout.b = cobj->sm->sc.b;

		/* find the light arriving here */
		for (oo = 0; oo < nobj; oo++) {
			if (obj[oo].sm->em > 0) {

				/* ray of direct light */
				int ooo, blocked = 0;
				struct elmRAY dray;
				struct elmCOORD diffray = {
					chitp.x - obj[oo].o.s.c.x,
					chitp.y - obj[oo].o.s.c.y,
					chitp.z - obj[oo].o.s.c.z
				};
				float draylen, dist;
				dray.s = obj[oo].o.s.c;
				dray.d = versor(&dray.s, &chitp);
				draylen = sqrtf(diffray.x * diffray.x + diffray.y * diffray.y + diffray.z * diffray.z);

				/* is this ray blocked by something? */
				for (ooo = 0; ooo < nobj; ooo++) {
					if (ooo != oo && obj[ooo].sm->em == 0) {
						dist = intersection(NULL, NULL, &obj[ooo], &dray);
						if (dist >= 0 && draylen - dist >= dist * tol) {
							blocked = 1;
							break;
						}
					}
				}

				/* add this light */
				if (!blocked) {
					float atten = (obj[oo].o.s.r * obj[oo].o.s.r) / (draylen * draylen);
					float angleatten = 1 /*sqrtf(1 - (chitn.x * dray.d.x + chitn.y * dray.d.y + chitn.z * dray.d.z))*/;
					ill.r += obj[oo].sm->em * obj[oo].sm->sc.r * atten * angleatten;
					ill.g += obj[oo].sm->em * obj[oo].sm->sc.g * atten * angleatten;
					ill.b += obj[oo].sm->em * obj[oo].sm->sc.b * atten * angleatten;
				}
			}
		}

		/* modify by illumination */
		ilmout.r *= ill.r;
		ilmout.g *= ill.g;
		ilmout.b *= ill.b;

	} else {

		/* hit the sky */
		ilmout.r = skycolor.r;
		ilmout.g = skycolor.g;
		ilmout.b = skycolor.b;
	}

	return ilmout;
}

float apply_exposure(float l)
{
	float expo = -1.00f;
	return (1.0f - expf(l * expo));
}

float srgb_encode(float c)
{
	if (c <= 0.0031308f) 
		return 12.92f * c;
	else
		return 1.055f * powf(c, 0.4166667f) - 0.055f; // Inverse gamma 2.4
}

unsigned char encode_pixel(float pixval)
{
	// exposure correction
	float pixvalec = apply_exposure(pixval);

	// gamma correction
	float pixvalge = srgb_encode(pixvalec);

	// ready for file
	unsigned char pixvaluc = FLOAT2UCHAR(pixvalge);

	return pixvaluc;
}

int main(void)
{
	int row = 0, col = 0, samp = 0;
	struct elmIMGF imgf = {640, 480};
	struct elmCAM cam = {
		{-1.5F, 0, 0.6F}, /* o */
		DEG2RAD(0), /* az */
		DEG2RAD(0), /* el */
		DEG2RAD(60), /* fov */
		(float) imgf.width / (float) imgf.height, /* ar */
		0.022F /* fl */
	};
	struct elmPOLY polys[] = {
		{{4,0,0}, {5,0,0}, {4,1,0}}
	};
	int npoly = sizeof(polys) / sizeof(struct elmPOLY);
	struct elmSPHE sphes[] = {
		{{4,0,0.8F}, 0.8F, {{0.2F,1,0.5F}, 1, 0, 0}}, // large back greenish sphere
		{{3,1.1F,0.8F}, 0.5F, {{1,0,0}, 1, 0, 0}}, // small lateral red sphere
		{{2,0.3F,0.2F}, 0.2F, {{0,0.2F,0.95F}, 1, 0, 0}}, // tiny front blue sphere
		{{0,0,-50000}, 50000, {{0.95F,0.95F,0.8F}, 0, 0, 0}}, // infinite floor sphere
		{{3.8F,0,4.5F}, 0.1F, {{1,1,1}, 0, 0, 1000}} // white light
	};
	int nsphe = sizeof(sphes) / sizeof(struct elmSPHE);
	struct elmOBJ obj[(sizeof(polys)/sizeof(struct elmPOLY)) + (sizeof(sphes)/sizeof(struct elmSPHE))] = {0};
	int nobj = 0, oo;
	int nrays = 60;

	/* prepare list of objects */
	/*for (oo = 0; oo < npoly; oo++, nobj++) {
		obj[nobj].t = ELMP;
		obj[nobj].o.p = polys[oo];
		obj[nobj].sm = &obj[nobj].o.p.sm;
	}*/
	for (oo = 0; oo < nsphe; oo++, nobj++) {
		obj[nobj].t = ELMS;
		obj[nobj].o.s = sphes[oo];
		obj[nobj].sm = &obj[nobj].o.s.sm;
	}

	/* initialize random seed: */
	srand((unsigned int) time(NULL));

	/* prepare image */
	imgf.buf = (float *) calloc(imgf.height * imgf.width * 3, sizeof(float));

	/* prepare camera */
	prepare_cam(&cam, &imgf);

	/* scan image pixels */
	for (row = 0; row < imgf.height; row++) {
		for (col = 0; col < imgf.width; col++) {

			if (row == 395) {
				__asm nop;
			}

			/* start from black */
			imgf.buf[(row * imgf.width + col) * 3 + 0] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 1] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 2] = 0;

			/* multiple samples */
			for (samp = 0; samp < max(1, nrays); samp++) {

				/* primary ray for that pixel */
				struct elmCLR ilm = {0,0,0};
				struct elmRAY pray = {0};
				primary_ray(&pray, &cam, row, col, nrays > 0);

				/* trace ray */
				ilm = trace(&pray, obj, nobj);

				/* assign illumination */
				imgf.buf[(row * imgf.width + col) * 3 + 0] += ilm.r;
				imgf.buf[(row * imgf.width + col) * 3 + 1] += ilm.g;
				imgf.buf[(row * imgf.width + col) * 3 + 2] += ilm.b;
			}

			/* renormalize */
			if (nrays > 1) {
				imgf.buf[(row * imgf.width + col) * 3 + 0] /= (float) nrays;
				imgf.buf[(row * imgf.width + col) * 3 + 1] /= (float) nrays;
				imgf.buf[(row * imgf.width + col) * 3 + 2] /= (float) nrays;
			}
		}

		fprintf(stderr, "%d/%d (%.1f%%)\r", row + 1, imgf.height, 100 * (float) (row + 1) / (float) imgf.height);
	}
	fprintf(stderr, "\n");

	/* write to file */
	{
		struct elmIMG img = {imgf.width, imgf.height, 3, 255};
		img.bytes = img.maxval > 255 ? 2 : 1;
		img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
		img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

		for (row = 0; row < imgf.height; row++) {
			for (col = 0; col < imgf.width; col++) {
				img.buffer[row * img.stride + col * img.depth * img.bytes + 0] = encode_pixel(imgf.buf[(row * imgf.width + col) * 3 + 0]);
				img.buffer[row * img.stride + col * img.depth * img.bytes + 1] = encode_pixel(imgf.buf[(row * imgf.width + col) * 3 + 1]);
				img.buffer[row * img.stride + col * img.depth * img.bytes + 2] = encode_pixel(imgf.buf[(row * imgf.width + col) * 3 + 2]);
			}
		}

		/* dump image */
		bmp_write("out.bmp", &img);

		/* free image */
		free(img.buffer);
		memset(&img, 0, sizeof(struct elmIMG));
	}

	/* free float image */
	free(imgf.buf);
	memset(&imgf, 0, sizeof(struct elmIMGF));

	return 0;
}
