/* a basic ray-tracer */
/* http://www.scratchapixel.com/lessons/3d-basic-rendering/introduction-to-ray-tracing/implementing-the-raytracing-algorithm */
/* https://en.wikipedia.org/wiki/Ray_tracing_(graphics) */

#define _USE_MATH_DEFINES
#include <math.h>
#include <time.h>
#include "elabmm.h"

extern const float NaNf = ((float)((1e308 * 10)*0.));
extern const float Inff = ((float)(1e308 * 10));

#define DEG2RAD(A) ((float) (A) * (float) M_PI / 180.0F)
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))

struct elmCOORD {
	float x, y, z;
};

struct elmCLR {
	float r, g, b;
};

struct elmMTR {
	struct elmCLR sc; /* basic color */
	float rr; /* reflective factor */
	float rf; /* refractive factor */
	float em; /* emissive facor */
};

struct elmROTM {
	float xx, xy, xz;
	float yx, yy, yz;
	float zx, zy, zz;
};

struct elmRAY {
	struct elmCOORD s;
	struct elmCOORD d;
};

struct elmCAM {
	struct elmCOORD o;
	float az, el;
	float fov;
	float ar;
	float fl;
	struct {
		float hs, vs, dx, dy;
	} wk;
};

struct elmPOLY {
	struct elmCOORD v1, v2, v3;
	struct elmMTR sm;
};

struct elmSPHE {
	struct elmCOORD c;
	float r;
	struct elmMTR sm;
};

struct elmOBJ {
	enum {ELMP, ELMS} t;
	union {
		struct elmPOLY p;
		struct elmSPHE s;
	} o;
	struct elmMTR *sm;
};

struct elmIMGF {
	int width, height;
	float *buf;
};

struct elmCLR skycolor = {0.15F,0.15F,0.25F};

struct elmCOORD versor(struct elmCOORD *v1, struct elmCOORD *v2)
{
	/* versor from v1 towards v2 */
	struct elmCOORD d = {v2->x - v1->x, v2->y - v1->y, v2->z - v1->z};
	float len = sqrtf(d.x * d.x + d.y * d.y + d.z * d.z);
	d.x /= len;
	d.y /= len;
	d.z /= len;
	return d;
}

int prepare_cam(struct elmCAM *cam, struct elmIMGF *imgf)
{
	/* camera size on the unit circle */
	cam->wk.hs = cam->fl * 2.0F * tanf(cam->fov / 2.0F);
	cam->wk.dx = cam->wk.hs / (float) imgf->width;
	cam->wk.dy = cam->wk.dx;
	cam->wk.vs = cam->wk.hs / cam->ar;

	return 0;
}

int primary_ray(struct elmRAY *pray, struct elmCAM *cam, int row, int col, int nrays, int r)
{
	struct elmCOORD orig = {0, 0, 0};
	struct elmCOORD pixelcoord = {
		cam->fl * cosf(cam->fov / 2.0F),
		cam->wk.hs / 2.0F - cam->wk.dx * ((float) col + 0.5F),
		cam->wk.vs / 2.0F - cam->wk.dy * ((float) row + 0.5F)
	};

	/* decide pixel position exactly */
	if (nrays < 0) {
		pixelcoord.y -= cam->wk.dx * (-0.5F + (float) rand() / (float) RAND_MAX);
		pixelcoord.z -= cam->wk.dy * (-0.5F + (float) rand() / (float) RAND_MAX);
	} else if (nrays > 0) {
		int l = (int) sqrtf(nrays);
		pixelcoord.y -= cam->wk.dx * (-0.5F + (float) (r % l) / (float) l);
		pixelcoord.z -= cam->wk.dy * (-0.5F + (float) (r / l) / (float) l);
	}

	/* origin of ray is the cam */
	memcpy(&pray->s, &cam->o, sizeof(struct elmCOORD));

	/* direction versor */
	pray->d = versor(&orig, &pixelcoord);

	/* rotate around z-axis according to azimut */
	pray->d.x = pray->d.x * cosf(cam->az) - pray->d.y * sinf(cam->az);
	pray->d.y = pray->d.x * sinf(cam->az) + pray->d.y * cosf(cam->az);

	/* rotate around new x-axis according to elevation */
	pray->d.x = -pray->d.z * sinf(cam->el) + pray->d.x * cosf(cam->el);
	pray->d.z =  pray->d.z * cosf(cam->el) + pray->d.x * sinf(cam->el);

	return 0;
}

float intersection(struct elmCOORD *hitp, struct elmCOORD *hitn, struct elmOBJ *obj, struct elmRAY *pray) 
{
	float t = +Inff;

	if (obj->t == ELMS) {

		/* sphere */
		struct elmSPHE *sphe = &(obj->o.s);
		struct elmCOORD v = {pray->s.x - sphe->c.x, pray->s.y - sphe->c.y, pray->s.z - sphe->c.z};
		float vd = (v.x * pray->d.x + v.y * pray->d.y + v.z * pray->d.z);
		float disc = vd * vd - (v.x * v.x + v.y * v.y + v.z * v.z) + sphe->r * sphe->r;
		if (disc < 0)
			return -1; /* not hit */
		else {
			/* hit */
			float discs = sqrtf(disc);
			float t1 = - vd + discs, t2 = - vd - discs;
			if (t2 >= 0.01F && t2 < t) {
				t = t2;
			}
			if (t1 >= 0.01F && t1 < t) {
				t = t1;
			}
			if (t < +Inff) {
				if (hitp) {
					hitp->x = pray->s.x + t * pray->d.x; 
					hitp->y = pray->s.y + t * pray->d.y; 
					hitp->z = pray->s.z + t * pray->d.z; 
					if (hitn)
						*hitn = versor(&sphe->c, hitp);
				}
				return t;
			} else
				return -1;
		}

	} else if (obj->t == ELMP) {

		/* polygon */
		return -1;
	}

	return -1;
}

struct elmCLR trace(struct elmRAY *ray, struct elmOBJ *obj, int nobj)
{
	struct elmCLR ilmout;
	int oo;
	float tol = 0.000F;

	/* trace ray */
	float mindist = +Inff;
	struct elmOBJ *cobj = NULL;
	struct elmCOORD hitp, hitn, chitp, chitn;

	/* search all objects */
	for (oo = 0; oo < nobj; oo++) {

		/* find intersection if any (not with lights) */
		float dist;
		if (obj[oo].sm->em == 0 && (dist = intersection(&hitp, &hitn, &obj[oo], ray)) >= 0) {

			/* save nearest object */
			if (dist < mindist) {
				mindist = dist;
				cobj = &obj[oo];
				chitp = hitp;
				chitn = hitn;
			}
		}
	}

	/* from far object */
	if (cobj) {

		struct elmCLR ill = {0,0,0};

		/* object color */
		ilmout.r = cobj->sm->sc.r;
		ilmout.g = cobj->sm->sc.g;
		ilmout.b = cobj->sm->sc.b;

		/* find the light arriving here */
		for (oo = 0; oo < nobj; oo++) {
			if (obj[oo].sm->em > 0) {

				struct elmSPHE *light = &obj[oo].o.s;

				/* ray of direct light */
				int ooo, blocked = 0;
				struct elmRAY dray;
				struct elmCOORD diffray = {
					chitp.x - light->c.x,
					chitp.y - light->c.y,
					chitp.z - light->c.z
				};
				float draylen, dist;
				dray.s = light->c;
				dray.d = versor(&dray.s, &chitp);
				draylen = sqrtf(diffray.x * diffray.x + diffray.y * diffray.y + diffray.z * diffray.z);

				/* is this ray blocked by something? */
				for (ooo = 0; ooo < nobj; ooo++) {
					if (ooo != oo && obj[ooo].sm->em == 0) {
						dist = intersection(NULL, NULL, &obj[ooo], &dray);
						if (dist >= 0 && draylen - dist >= dist * tol) {
							/* is it the same sphere */
							if (cobj->t == ELMS && cobj == &obj[ooo]) {
								float dn = chitn.x * dray.d.x + chitn.y * dray.d.y + chitn.z * dray.d.z;
								if (dn > 0) {
									blocked = 1;
									break;
								}
							} else {
								blocked = 1;
								break;
							}
						}
					}
				}

				/* add this light */
				if (!blocked) {
					float pathgain = (light->r * light->r) / (draylen * draylen);
					float dh = chitn.x * dray.d.x + chitn.y * dray.d.y + chitn.z * dray.d.z;
					float anglegain = dh < 0 ? -dh : dh;
					ill.r += light->sm.em * light->sm.sc.r * pathgain * anglegain;
					ill.g += light->sm.em * light->sm.sc.g * pathgain * anglegain;
					ill.b += light->sm.em * light->sm.sc.b * pathgain * anglegain;
				}
			}
		}

		/* modify by illumination */
		ilmout.r *= ill.r;
		ilmout.g *= ill.g;
		ilmout.b *= ill.b;

	} else {

		/* hit the sky */
		ilmout.r = skycolor.r;
		ilmout.g = skycolor.g;
		ilmout.b = skycolor.b;
	}

	return ilmout;
}

float apply_exposure(float l)
{
	float expo = -1.00f;
	return (1.0f - expf(l * expo));
}

float srgb_encode(float c)
{
	if (c <= 0.0031308f) 
		return 12.92f * c;
	else
		return 1.055f * powf(c, 0.4166667f) - 0.055f; // Inverse gamma 2.4
}

unsigned char encode_pixel(float pixval)
{
	// exposure correction
	float pixvalec = apply_exposure(pixval);

	// gamma correction
	float pixvalge = srgb_encode(pixvalec);

	// ready for file
	unsigned char pixvaluc = FLOAT2UCHAR(pixvalge);

	return pixvaluc;
}

int main(void)
{
	int row = 0, col = 0, samp = 0;
	struct elmIMGF imgf = {640, 480};
	struct elmCAM cam = {
		{-1.5F, 0, 0.6F}, /* o */
		DEG2RAD(0), /* az */
		DEG2RAD(0), /* el */
		DEG2RAD(60), /* fov */
		(float) imgf.width / (float) imgf.height, /* ar */
		0.022F /* fl */
	};
	struct elmPOLY polys[] = {
		{{4,0,0}, {5,0,0}, {4,1,0}}
	};
	int npoly = sizeof(polys) / sizeof(struct elmPOLY);
	struct elmSPHE sphes[] = {
		{{4,0,0.8F}, 0.8F, {{0.2F,1,0.5F}, 1, 0, 0}}, // large back greenish sphere
		{{3,1.1F,0.8F}, 0.5F, {{1,0,0}, 1, 0, 0}}, // small lateral red sphere
		{{2,0.3F,0.2F}, 0.2F, {{0,0.2F,0.95F}, 1, 0, 0}}, // tiny front blue sphere
		{{0,0,-50000}, 50000, {{0.95F,0.95F,0.8F}, 0, 0, 0}}, // infinite floor sphere
		{{3.7F,0,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.7F,0.1,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.7F,0.2,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.8F,0,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.8F,0.1,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.8F,0.2,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.9F,0,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.9F,0.1,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}}, // white light
		{{3.9F,0.2,3.4F}, 0.1F, {{1,1,1}, 0, 0, 25}} // white light
	};
	int nsphe = sizeof(sphes) / sizeof(struct elmSPHE);
	struct elmOBJ obj[(sizeof(polys)/sizeof(struct elmPOLY)) + (sizeof(sphes)/sizeof(struct elmSPHE))] = {0};
	int nobj = 0, oo;
	int nrays = -64;
	int npasses = max(1, abs(nrays));

	/* prepare list of objects */
	/*for (oo = 0; oo < npoly; oo++, nobj++) {
		obj[nobj].t = ELMP;
		obj[nobj].o.p = polys[oo];
		obj[nobj].sm = &obj[nobj].o.p.sm;
	}*/
	for (oo = 0; oo < nsphe; oo++, nobj++) {
		obj[nobj].t = ELMS;
		obj[nobj].o.s = sphes[oo];
		obj[nobj].sm = &obj[nobj].o.s.sm;
	} 

	/* initialize random seed: */
	srand((unsigned int) time(NULL));

	/* prepare image */
	imgf.buf = (float *) calloc(imgf.height * imgf.width * 3, sizeof(float));

	/* prepare camera */
	prepare_cam(&cam, &imgf);

	/* start from black image */
	for (row = 0; row < imgf.height; row++) {
		for (col = 0; col < imgf.width; col++) {
			imgf.buf[(row * imgf.width + col) * 3 + 0] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 1] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 2] = 0;
		}
	}

	/* scan image pixels */
	for (samp = 0; samp < npasses; samp++) {
		for (row = 0; row < imgf.height; row++) {
			for (col = 0; col < imgf.width; col++) {

				/*if (row == 395) __asm nop;*/

				/* primary ray for that pixel */
				struct elmCLR ilm = {0,0,0};
				struct elmRAY pray = {0};
				primary_ray(&pray, &cam, row, col, nrays, samp);

				/* trace ray */
				ilm = trace(&pray, obj, nobj);

				/* increment illumination */
				imgf.buf[(row * imgf.width + col) * 3 + 0] += ilm.r;
				imgf.buf[(row * imgf.width + col) * 3 + 1] += ilm.g;
				imgf.buf[(row * imgf.width + col) * 3 + 2] += ilm.b;
			}
			fprintf(stderr, "%d/%d (%.1f%%)\r", samp + 1, npasses, 100 * (float) (row + 1 + samp * imgf.height) / (float) (imgf.height * npasses));
		}


		/* write to file */
		if (samp == npasses - 1 || samp % 4 == 3) {
			struct elmIMG img = {imgf.width, imgf.height, 3, 255};
			img.bytes = img.maxval > 255 ? 2 : 1;
			img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
			img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

			for (row = 0; row < imgf.height; row++) {
				for (col = 0; col < imgf.width; col++) {
					img.buffer[row * img.stride + col * img.depth * img.bytes + 0] =
						encode_pixel(imgf.buf[(row * imgf.width + col) * 3 + 0] / fabsf(samp + 1));
					img.buffer[row * img.stride + col * img.depth * img.bytes + 1] =
						encode_pixel(imgf.buf[(row * imgf.width + col) * 3 + 1] / fabsf(samp + 1));
					img.buffer[row * img.stride + col * img.depth * img.bytes + 2] =
						encode_pixel(imgf.buf[(row * imgf.width + col) * 3 + 2] / fabsf(samp + 1));
				}
			}

			/* dump image */
			bmp_write("out.bmp", &img);

			/* free image */
			free(img.buffer);
			memset(&img, 0, sizeof(struct elmIMG));
		}

		/* renormalize */
		/*if (abs(nrays) > 1) {
			imgf.buf[(row * imgf.width + col) * 3 + 0] /= fabsf(nrays);
			imgf.buf[(row * imgf.width + col) * 3 + 1] /= fabsf(nrays);
			imgf.buf[(row * imgf.width + col) * 3 + 2] /= fabsf(nrays);
		}*/
	}
	fprintf(stderr, "\n");


	/* free float image */
	free(imgf.buf);
	memset(&imgf, 0, sizeof(struct elmIMGF));

	return 0;
}
