/* a basic ray-tracer */
/* http://www.scratchapixel.com/lessons/3d-basic-rendering/introduction-to-ray-tracing/implementing-the-raytracing-algorithm */
/* https://en.wikipedia.org/wiki/Ray_tracing_(graphics) */

#define _USE_MATH_DEFINES
#include <math.h>
#include <time.h>
#include "elabmm.h"

extern const float NaNf = ((float)((1e308 * 10)*0.));
extern const float Inff = ((float)(1e308 * 10));

#define DEG2RAD(A) ((float) (A) * (float) M_PI / 180.0F)
#define FLOAT2UCHAR(P) ((P) > 1.0F ? 255 : ((P) < 0 ? 0 : (unsigned char) ((P) * 255.0F + 0.5F)))
#define NAMELEN 32

struct elmCOORD {
	float x, y, z;
};

struct elmCLR {
	float r, g, b;
};

struct elmMTR {
	struct elmCLR sc; /* basic color */
	float rr; /* reflective factor (one minus diffusive) */
	float rf; /* refractive factor */
	float em; /* emissive factor */
};

struct elmROTM {
	float xx, xy, xz;
	float yx, yy, yz;
	float zx, zy, zz;
};

struct elmRAY {
	struct elmCOORD s;
	struct elmCOORD d;
};

struct elmCAM {
	struct elmCOORD o;
	float az, el;
	float fov;
	float ar;
	float fl;
	struct {
		float hs, vs, dx, dy;
	} wk;
};

struct elmPOLY {
	struct elmCOORD v1, v2, v3;
	struct elmMTR sm;
	char name[NAMELEN];
};

struct elmSPHE {
	struct elmCOORD c;
	float r;
	struct elmMTR sm;
	char name[NAMELEN];
};

struct elmOBJ {
	enum {ELMP, ELMS} t;
	union {
		struct elmPOLY p;
		struct elmSPHE s;
	} o;
	struct elmMTR *sm;
	char *name;
};

struct elmIMGF {
	int width, height;
	int bord;
	float *buf;
};

int normalize(struct elmCOORD *d)
{
	float len = sqrtf(d->x * d->x + d->y * d->y + d->z * d->z);
	d->x /= len;
	d->y /= len;
	d->z /= len;
	return 0;
}

struct elmCOORD versor(struct elmCOORD *v1, struct elmCOORD *v2)
{
	/* versor from v1 towards v2 */
	struct elmCOORD d = {v2->x - v1->x, v2->y - v1->y, v2->z - v1->z};
	normalize(&d);
	return d;
}

int prepare_cam(struct elmCAM *cam, struct elmIMGF *imgf)
{
	/* camera size on the unit circle */
	cam->wk.hs = cam->fl * 2.0F * tanf(cam->fov / 2.0F);
	cam->wk.dx = cam->wk.hs / (float) imgf->width;
	cam->wk.dy = cam->wk.dx;
	cam->wk.vs = cam->wk.hs / cam->ar;

	return 0;
}

int primary_ray(struct elmRAY *pray, struct elmCAM *cam, int row, int col, int nrays, int r)
{
	struct elmCOORD orig = {0, 0, 0};
	struct elmCOORD pixelcoord = {
		cam->fl * cosf(cam->fov / 2.0F),
		cam->wk.hs / 2.0F - cam->wk.dx * ((float) col + 0.5F),
		cam->wk.vs / 2.0F - cam->wk.dy * ((float) row + 0.5F)
	};

	/* decide pixel position exactly */
	if (nrays < 0) {
		pixelcoord.y -= cam->wk.dx * (-0.5F + (float) rand() / (float) RAND_MAX);
		pixelcoord.z -= cam->wk.dy * (-0.5F + (float) rand() / (float) RAND_MAX);
	} else if (nrays > 0) {
		int l = (int) sqrtf(nrays);
		pixelcoord.y -= cam->wk.dx * (-0.5F + (float) (r % l) / (float) l);
		pixelcoord.z -= cam->wk.dy * (-0.5F + (float) (r / l) / (float) l);
	}

	/* origin of ray is the cam */
	memcpy(&pray->s, &cam->o, sizeof(struct elmCOORD));

	/* direction versor */
	pray->d = versor(&orig, &pixelcoord);

	/* rotate around z-axis according to azimut */
	pray->d.x = pray->d.x * cosf(cam->az) - pray->d.y * sinf(cam->az);
	pray->d.y = pray->d.x * sinf(cam->az) + pray->d.y * cosf(cam->az);

	/* rotate around new x-axis according to elevation */
	pray->d.x = -pray->d.z * sinf(cam->el) + pray->d.x * cosf(cam->el);
	pray->d.z =  pray->d.z * cosf(cam->el) + pray->d.x * sinf(cam->el);

	return 0;
}

/* http://geomalgorithms.com/a06-_intersect-2.html */
// intersect3D_RayTriangle(): find the 3D intersection of a ray with a triangle
//    Input:  a ray R, and a triangle T
//    Output: *I = intersection point (when it exists)
//    Return: -1 = triangle is degenerate (a segment or point)
//             0 =  disjoint (no intersect)
//             1 =  intersect in unique point I1
//             2 =  are in the same plane
//int intersection_triangle(Ray R, Triangle T, Point* I)
float intersection_triangle(struct elmCOORD *hitp, struct elmCOORD *hitn, struct elmPOLY *tri, struct elmRAY *pray)
{
    struct elmCOORD u, v, n; // triangle vectors
    struct elmCOORD w0, w; // ray vectors
    float r, a, b; // params to calc ray-plane intersect
	float small_num = 0.00000001F; // anything that avoids division overflow
    float uu, uv, vv, wu, wv, D;
    float s, t;
	struct elmCOORD P;

    // get triangle edge vectors and plane normal
	u.x = tri->v2.x - tri->v1.x;
	u.y = tri->v2.y - tri->v1.y;
	u.z = tri->v2.z - tri->v1.z;
	v.x = tri->v3.x - tri->v1.x;
	v.y = tri->v3.y - tri->v1.y;
	v.z = tri->v3.z - tri->v1.z;

	// cross product
	n.x = u.y * v.z - u.z * v.y;
	n.y = u.z * v.x - u.x * v.z;
	n.z = u.x * v.y - u.y * v.x;

    // triangle is degenerate
    // do not deal with this case
	if (n.x == 0 && n.y == 0 && n.z == 0)
		return -1;

	w0.x = pray->s.x - tri->v1.x;
	w0.y = pray->s.y - tri->v1.y;
	w0.z = pray->s.z - tri->v1.z;

	a = -(n.x * w0.x + n.y * w0.y + n.z * w0.z);
	b = n.x * pray->d.x + n.y * pray->d.y + n.z * pray->d.z; 
    if (fabs(b) < small_num) {     // ray is  parallel to triangle plane
        if (a == 0)                 // ray lies in triangle plane
            return -1;
        else
			return -1;              // ray disjoint from plane
    }

    // get intersect point of ray with triangle plane
    r = a / b;
    if (r < 0)                    // ray goes away from triangle
        return -1;                   // => no intersect
    // for a segment, also test if (r > 1.0) => no intersect

    // intersect point of ray and plane
	P.x = pray->s.x + r * pray->d.x;
	P.y = pray->s.y + r * pray->d.y;
	P.z = pray->s.z + r * pray->d.z;

    // is I inside T?
    uu = u.x * u.x + u.y * u.y + u.z * u.z;
    uv = u.x * v.x + u.y * v.y + u.z * v.z;
    vv = v.x * v.x + v.y * v.y + v.z * v.z;
	w.x = P.x - tri->v1.x;
	w.y = P.y - tri->v1.y;
	w.z = P.z - tri->v1.z;
	wu = w.x * u.x + w.y * u.y + w.z * u.z;
	wv = w.x * v.x + w.y * v.y + w.z * v.z;
    D = uv * uv - uu * vv;

    // get and test parametric coords
    s = (uv * wv - vv * wu) / D;
    if (s < 0 || s > 1) // I is outside T
        return -1;
    t = (uv * wu - uu * wv) / D;
    if (t < 0 || (s + t) > 1) // I is outside T
        return -1;

	if (hitn) {
		*hitn = n;
		normalize(hitn);
		if (pray->d.x * hitn->x + pray->d.y * hitn->y + pray->d.z * hitn->z > 0) {
			hitn->x = -hitn->x;
			hitn->y = -hitn->y;
			hitn->z = -hitn->z;
		}
	}
	if (hitp)
		*hitp = P;

    return r; // I is in T
}

float intersection(struct elmCOORD *hitp, struct elmCOORD *hitn, struct elmOBJ *obj, struct elmRAY *pray) 
{
	float t = +Inff;

	if (obj->t == ELMS) {

		/* sphere */
		struct elmSPHE *sphe = &(obj->o.s);
		struct elmCOORD v = {pray->s.x - sphe->c.x, pray->s.y - sphe->c.y, pray->s.z - sphe->c.z};
		float vd = (v.x * pray->d.x + v.y * pray->d.y + v.z * pray->d.z);
		float disc = vd * vd - (v.x * v.x + v.y * v.y + v.z * v.z) + sphe->r * sphe->r;
		if (disc < 0)
			return -1; /* not hit */
		else {
			/* hit */
			float discs = sqrtf(disc);
			float t1 = - vd + discs, t2 = - vd - discs;
			if (t2 >= 0.01F && t2 < t) {
				t = t2;
			}
			if (t1 >= 0.01F && t1 < t) {
				t = t1;
			}
			if (t < +Inff) {
				if (hitp) {
					hitp->x = pray->s.x + t * pray->d.x; 
					hitp->y = pray->s.y + t * pray->d.y; 
					hitp->z = pray->s.z + t * pray->d.z; 
					if (hitn)
						*hitn = versor(&sphe->c, hitp);
				}
				return t;
			} else
				return -1;
		}

	} else if (obj->t == ELMP) {

		/* polygon */
		return intersection_triangle(hitp, hitn, &obj->o.p, pray);
	}

	return -1;
}

struct elmIMG sky;

struct elmCLR map_texture(struct elmIMG *img, float x, float y)
{
	struct elmCLR ilmout = {0,0,0};
	float u, v;
	int col, row;
	struct elmCLR P0, P1, P2, P3;
	if (x < 0)
		x = 0;
	else if (x > img->width - 2)
		x = (float) img->width - 2;
	if (y < 0)
		y = 0;
	else if (y > img->height - 2)
		y = (float) img->height - 2;
	u = x - floorf(x);
	v = 1 - (y - floorf(y));
	col = (int) x;
	row = img->height - 1 - (int) y;
	P0.r = (float) img->buffer[row * img->stride + col * img->depth * img->bytes + 0] / 255.0F;
	P0.g = (float) img->buffer[row * img->stride + col * img->depth * img->bytes + 1] / 255.0F;
	P0.b = (float) img->buffer[row * img->stride + col * img->depth * img->bytes + 2] / 255.0F;
	P1.r = (float) img->buffer[row * img->stride + (col + 1) * img->depth * img->bytes + 0] / 255.0F;
	P1.g = (float) img->buffer[row * img->stride + (col + 1) * img->depth * img->bytes + 1] / 255.0F;
	P1.b = (float) img->buffer[row * img->stride + (col + 1) * img->depth * img->bytes + 2] / 255.0F;
	P2.r = (float) img->buffer[(row + 1) * img->stride + col * img->depth * img->bytes + 0] / 255.0F;
	P2.g = (float) img->buffer[(row + 1) * img->stride + col * img->depth * img->bytes + 1] / 255.0F;
	P2.b = (float) img->buffer[(row + 1) * img->stride + col * img->depth * img->bytes + 2] / 255.0F;
	P3.r = (float) img->buffer[(row + 1) * img->stride + (col + 1) * img->depth * img->bytes + 0] / 255.0F;
	P3.g = (float) img->buffer[(row + 1) * img->stride + (col + 1) * img->depth * img->bytes + 1] / 255.0F;
	P3.b = (float) img->buffer[(row + 1) * img->stride + (col + 1) * img->depth * img->bytes + 2] / 255.0F;
	ilmout.r = ((1 - u) * P0.r + u * P1.r) * (1 - v) + ((1 - u) * P2.r + u * P3.r) * v;
	ilmout.g = ((1 - u) * P0.g + u * P1.g) * (1 - v) + ((1 - u) * P2.g + u * P3.g) * v;
	ilmout.b = ((1 - u) * P0.b + u * P1.b) * (1 - v) + ((1 - u) * P2.b + u * P3.b) * v;
	return ilmout;
}

struct elmCLR map_skycube(struct elmRAY *ray)
{
	struct elmCLR skycolor = {0.15F,0.15F,0.25F};
	//struct elmCLR skycolor = {1,1,1};
	//struct elmCLR skycolor = {1,0,0};
	struct elmCLR ilmout = skycolor;

	/* map on the cube */
	if (sky.buffer) {

		float cx2 = (float) sky.width / 8.0F, cy2 = (float) sky.height / 6.0F;
		float xm, ym;

		if (fabsf(ray->d.x) > fabsf(ray->d.y) && fabsf(ray->d.x) > fabsf(ray->d.z)) {
			if (ray->d.x > 0) {
				/* facing front */
				float xo = 0.375F * (float) sky.width, yo = 0.5F * (float) sky.height;
				/*ilmout.r = 0.5F;
				ilmout.g = 0;
				ilmout.b = 0;*/
				xm = xo - cx2 * ray->d.y / ray->d.x;
				ym = yo + cy2 * ray->d.z / ray->d.x;
			} else {
				float xo = 0.875F * (float) sky.width, yo = 0.5F * (float) sky.height;
				/* facing back */
				/*ilmout.r = 0.5F;
				ilmout.g = 0.5F;
				ilmout.b = 0;*/
				xm = xo - cx2 * ray->d.y / ray->d.x;
				ym = yo - cy2 * ray->d.z / ray->d.x;
			}
		} else if (fabsf(ray->d.y) > fabsf(ray->d.x) && fabsf(ray->d.y) > fabsf(ray->d.z)) {
			if (ray->d.y > 0) {
				float xo = 0.125F * (float) sky.width, yo = 0.5F * (float) sky.height;
				/* facing left */
				/*ilmout.r = 0;
				ilmout.g = 0.5F;
				ilmout.b = 0;*/
				xm = xo + cx2 * ray->d.x / ray->d.y;
				ym = yo + cy2 * ray->d.z / ray->d.y;
			} else {
				/* facing right */
				float xo = 0.625F * (float) sky.width, yo = 0.5F * (float) sky.height;
				/*ilmout.r = 0;
				ilmout.g = 0.5F;
				ilmout.b = 0.5F;*/
				xm = xo + cx2 * ray->d.x / ray->d.y;
				ym = yo - cy2 * ray->d.z / ray->d.y;
			}
		} else if (fabsf(ray->d.z) > fabsf(ray->d.x) && fabsf(ray->d.z) > fabsf(ray->d.y)) {
			if (ray->d.z > 0) {
				float xo = 0.375F * (float) sky.width, yo = 0.83333333F * (float) sky.height;
				/* facing top */
				/*ilmout.r = 0;
				ilmout.g = 0;
				ilmout.b = 0.5F;*/
				xm = xo - cx2 * ray->d.y / ray->d.z;
				ym = yo - cy2 * ray->d.x / ray->d.z;
			} else {
				float xo = 0.375F * (float) sky.width, yo = 0.16666666F * (float) sky.height;
				/* facing bottom */
				/*ilmout.r = 0.5F;
				ilmout.g = 0;
				ilmout.b = 0.5F;*/
				xm = xo - cx2 * ray->d.y / ray->d.z;
				ym = yo + cy2 * ray->d.x / ray->d.z;
			}
		} else {
			return ilmout;
		}
		ilmout = map_texture(&sky, xm, ym);
	}

	return ilmout;
}

struct elmCLR trace(struct elmRAY *ray, struct elmOBJ *obj, int nobj, int bounces_left)
{
	struct elmCLR ilmout = {0,0,0}, ilmdiff = {0,0,0}, ilmrefl = {0,0,0}, ilmem = {0,0,0};
	int oo;
	float tol = 0.000F, small_num = 0.000001F;

	/* trace ray */
	float mindist = +Inff;
	struct elmOBJ *cobj = NULL;
	struct elmCOORD hitp, hitn, chitp, chitn;

	/* ran out of bounces */
	if (bounces_left <= 0)
		return ilmout;

	/* search all objects */
	for (oo = 0; oo < nobj; oo++) {

		/* find intersection if any (not with lights) */
		float dist;
		if (/*obj[oo].sm->em == 0 &&*/ (dist = intersection(&hitp, &hitn, &obj[oo], ray)) >= 0) {

			/* save nearest object */
			if (dist < mindist) {
				mindist = dist;
				cobj = &obj[oo];
				chitp = hitp;
				chitn = hitn;
			}
		}
	}

	/* from far object */
	if (cobj) {

		struct elmCLR ill = {0,0,0};

		/* object color */
		ilmdiff.r = 0;
		ilmdiff.g = 0;
		ilmdiff.b = 0;

		/* if diffusive, find the light arriving here */
		if (cobj->sm->rr < 1) {

			/* object color */
			ilmdiff.r = (1 - cobj->sm->rr) * cobj->sm->sc.r;
			ilmdiff.g = (1 - cobj->sm->rr) * cobj->sm->sc.g;
			ilmdiff.b = (1 - cobj->sm->rr) * cobj->sm->sc.b;

			/* browse all objects */
			for (oo = 0; oo < nobj; oo++) {

				/* look from light sources and reflectors */
				if (obj[oo].sm->em > 0) {

					struct elmSPHE *light = &obj[oo].o.s;

					/* ray of direct light */
					int ooo, blocked = 0;
					struct elmRAY dray;
					struct elmCOORD diffray = {
						chitp.x - light->c.x,
						chitp.y - light->c.y,
						chitp.z - light->c.z
					};
					float draylen, dist;
					dray.s = light->c;
					dray.d = versor(&dray.s, &chitp);
					draylen = sqrtf(diffray.x * diffray.x + diffray.y * diffray.y + diffray.z * diffray.z);

					/* is this ray blocked by something? */
					for (ooo = 0; ooo < nobj; ooo++) {
						if (ooo != oo && obj[ooo].sm->em == 0) {
							dist = intersection(NULL, NULL, &obj[ooo], &dray);
							if (dist >= 0 && draylen - dist >= dist * tol) {
								/* is it the same sphere */
								if (cobj->t == ELMS && cobj == &obj[ooo]) {
									float dn = chitn.x * dray.d.x + chitn.y * dray.d.y + chitn.z * dray.d.z;
									if (dn > 0) {
										blocked = 1;
										break;
									}
								} else if (cobj->t == ELMP && cobj == &obj[ooo]) {
									/* it is the same triangle */
									float dn = chitn.x * dray.d.x + chitn.y * dray.d.y + chitn.z * dray.d.z;
									if (dn > 0) {
										blocked = 1;
										break;
									}
								} else {
									blocked = 1;
									break;
								}
							}
						}
					}

					/* add this light */
					if (!blocked) {
						float pathgain = (light->r * light->r) / (draylen * draylen);
						float dh = chitn.x * dray.d.x + chitn.y * dray.d.y + chitn.z * dray.d.z;
						float anglegain = dh < 0 ? -dh : dh;
						ill.r += light->sm.em * light->sm.sc.r * pathgain * anglegain;
						ill.g += light->sm.em * light->sm.sc.g * pathgain * anglegain;
						ill.b += light->sm.em * light->sm.sc.b * pathgain * anglegain;
					}
				
					/* modify by illumination */
					ilmdiff.r *= ill.r;
					ilmdiff.g *= ill.g;
					ilmdiff.b *= ill.b;

				}
			}
		}

		/* this object is also reflective */
		if (cobj->sm->rr > 0) {

			/* reflected ray */
			struct elmRAY rray;
			float dn = ray->d.x * chitn.x + ray->d.y * chitn.y + ray->d.z * chitn.z; 
			float fresnel = 0.9F * powf(1 - dn, 3);
			rray.d.x = ray->d.x - 2.0F * dn * chitn.x;
			rray.d.y = ray->d.y - 2.0F * dn * chitn.y;
			rray.d.z = ray->d.z - 2.0F * dn * chitn.z;
			normalize(&rray.d);
			rray.s.x = chitp.x + small_num * rray.d.x;
			rray.s.y = chitp.y + small_num * rray.d.y;
			rray.s.z = chitp.z + small_num * rray.d.z;

			/* trace this back  */
			ilmrefl = trace(&rray, obj, nobj, bounces_left - 1);

			/* fresnel part */
			ilmrefl.r *= fresnel * cobj->sm->sc.r * cobj->sm->rr;
			ilmrefl.g *= fresnel * cobj->sm->sc.g * cobj->sm->rr;
			ilmrefl.b *= fresnel * cobj->sm->sc.b * cobj->sm->rr;
		}

		/* it is also a light */
		if (cobj->sm->em > 0) {
			ilmem.r = cobj->sm->em * cobj->sm->sc.r;
			ilmem.g = cobj->sm->em * cobj->sm->sc.g;
			ilmem.b = cobj->sm->em * cobj->sm->sc.b;
		}

		/* compose everything */
		ilmout.r = ilmdiff.r + ilmrefl.r + ilmem.r;
		ilmout.g = ilmdiff.g + ilmrefl.g + ilmem.g;
		ilmout.b = ilmdiff.b + ilmrefl.b + ilmem.b;

	} else {

		/* hit the sky */
		ilmout = map_skycube(ray);
	}

	return ilmout;
}

float apply_exposure(float l)
{
	float expo = -1.00f;
	return (1.0f - expf(l * expo));
}

float srgb_encode(float c)
{
	if (c <= 0.0031308f) 
		return 12.92f * c;
	else
		return 1.055f * powf(c, 0.4166667f) - 0.055f; // Inverse gamma 2.4
}

unsigned char encode_pixel(float pixval)
{
	// exposure correction
	float pixvalec = apply_exposure(pixval);

	// gamma correction
	float pixvalge = srgb_encode(pixvalec);

	// ready for file
	unsigned char pixvaluc = FLOAT2UCHAR(pixvalge);

	return pixvaluc;
}

#pragma warning (disable:4305)

int post_process(struct elmIMGF *imgf2, struct elmIMGF *imgf)
{
	int row, col, Lx = 9, Ly = 9, x, y, N = 2 * Lx + 1;
	float kernn[] = {
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0001, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0001, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0002, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0003, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0005, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0005, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.001, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.01, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	                0.0001, 0.0001, 0.0002, 0.0003, 0.0005, 0.0005, 0.001, 0.005, 0.01, 1, 0.01, 0.005, 0.001, 0.0005, 0.0005, 0.0003, 0.0002, 0.0001, 0.0001, 
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.01, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.005, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.001, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0005, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0005, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0003, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0002, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0001, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0.0001, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	}, *kern = kernn + (N + 1) * Ly, k = 1.0 / 1.032;

	/* filter image */
	for (row = 0; row < imgf->height; row++) {
		for (col = 0; col < imgf->width; col++) {
			for (y = -Ly; y <= Ly; y++) {
				int roww = row - y;
				if (roww >= 0 && roww < imgf->height) {
					for (x = -Lx; x <= Lx; x++) {
						int coll = col - x;
						if (coll >= 0 && coll < imgf->width) {
							imgf2->buf[(row * imgf2->width + col) * 3 + 0] +=
								imgf->buf[(roww * imgf->width + coll) * 3 + 0] * kern[y * N + x] * k;
							imgf2->buf[(row * imgf2->width + col) * 3 + 1] += 
								imgf->buf[(roww * imgf->width + coll) * 3 + 1] * kern[y * N + x] * k;
							imgf2->buf[(row * imgf2->width + col) * 3 + 2] +=
								imgf->buf[(roww * imgf->width + coll) * 3 + 2] * kern[y * N + x] * k;
						}
					}
				}
			}
		}
	}

	return 0;
}

int main(void)
{
	int row = 0, col = 0, samp = 0;
	struct elmIMGF imgf = {640, 480, 8};
	struct elmCAM cam = {
		{-1.5F, 0, 0.6F}, /* o */
		DEG2RAD(0), /* az */
		DEG2RAD(0), /* el */
		DEG2RAD(60), /* fov */
		(float) imgf.width / (float) imgf.height, /* ar */
		0.022F /* fl */
	};
	struct elmPOLY polys[] = {
		// V1 V2 V3 color refl refr emiss name
		{{2.6,0.7,0.3F}, {3.4,0.7,0.3F}, {2.6,1.4,0.3F}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // top
		{{2.6,1.4,0.3F}, {3.4,0.7,0.3F}, {3.4,1.4,0.3F}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // top
		{{2.6,0.7,0.3F}, {3.4,0.7,0.3F}, {3.4,0.7,0}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // right
		{{2.6,0.7,0}, {3.4,0.7,0}, {2.6,0.7,0.3}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // right
		{{2.6,1.4,0.3}, {2.6,0.7,0.3}, {2.6,0.7,0}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // front
		{{2.6,1.4,0}, {2.6,0.7,0}, {2.6,1.4,0.3}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // front
		{{2.6,1.4,0.3F}, {3.4,1.4,0.3F}, {3.4,1.4,0}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // left
		{{2.6,1.4,0}, {3.4,1.4,0}, {2.6,1.4,0.3}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // left
		{{3.4,1.4,0.3}, {3.4,0.7,0.3}, {3.4,0.7,0}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // back
		{{3.4,1.4,0}, {3.4,0.7,0}, {3.4,1.4,0.3}, {{1,0,1}, 0, 0, 0}, "cube 1"}, // back
		/*{{4,3,0}, {6,3,0}, {4,3.03,1}, {{0.2,0.2,0.2}, 1, 0, 0}, "mirror"}, // mirror bottom
		{{4,3.03,1}, {6,3.03,1}, {6,3,0}, {{0.2,0.2,0.2}, 1, 0, 0}, "mirror"}, // mirror bottom
		{{4.03,3.03,1}, {4.03,5,1}, {4,3,0}, {{0.2,0.2,0.2}, 1, 0, 0}, "mirror"}, // mirror bottom
		{{4,3,0}, {4,5,0}, {4.03,5,1}, {{0.2,0.2,0.2}, 1, 0, 0}, "mirror"}, // mirror bottom*/
		/*{{3,-1,0}, {2.5,-1.3,0}, {2.5,-1.3,2}, {{0.2,0.2,0.2}, 1, 0, 0}, "mirror 1"}, // mirror
		{{3,-1,0}, {3,-1,2}, {2.5,-1.3,2}, {{0.2,0.2,0.2}, 1, 0, 0}, "mirror 1"}, // mirror*/
	};
	int npoly = sizeof(polys) / sizeof(struct elmPOLY);
	struct elmSPHE sphes[] = {
		// coord radius color refl refr emiss name
		/*{{5,4,1.6F}, 1.6F, {{0.2F,0.2F,0.2F}, 1, 0, 0}, "huge reflecting ball"}, // huge back reflecting sphere 
		{{7.3,2.5,1.6F}, 1.6F, {{0.2F,0.2F,0.2F}, 1, 0, 0}, "huge reflecting ball 2"}, // huge back reflecting sphere 
		{{5,-2,1.0F}, 1.0F, {{0,1,0.8F}, 1, 0, 0}, "cyan reflecting ball"}, // huge back reflecting sphere 
		{{4,0,0.8F}, 0.8F, {{0.2F,1,0.5F}, 0, 0, 0}, "large greenish ball"}, // large main greenish sphere*/
		{{3,1.1F,0.8F}, 0.5F, {{1,0,0}, 0, 0, 0}, "small reddish ball"}, // small lateral red sphere
		/*{{2,0.3F,0.2F}, 0.2F, {{0,0.2F,0.95F}, 0, 0, 0}, "tiny blue ball"}, // tiny front blue sphere*/
		{{0,0,-50000}, 50000, {{0.95F,0.95F,0.8F}, 0, 0, 0}, "floor"}, // infinite floor sphere
		{{3.8F,0,3.55F}, 0.2F, {{1,1,1}, 0, 0, 330}, "light"}, // white light
	};
	int nsphe = sizeof(sphes) / sizeof(struct elmSPHE);
	struct elmOBJ obj[(sizeof(polys)/sizeof(struct elmPOLY)) + (sizeof(sphes)/sizeof(struct elmSPHE))] = {0};
	int nobj = 0, oo;
	int nrays = 0;
	int npasses = max(1, abs(nrays));

	/* prepare list of objects */
	for (oo = 0; oo < npoly; oo++, nobj++) {
		obj[nobj].t = ELMP;
		obj[nobj].o.p = polys[oo];
		obj[nobj].sm = &obj[nobj].o.p.sm;
		obj[nobj].name = obj[nobj].o.p.name;
	}
	for (oo = 0; oo < nsphe; oo++, nobj++) {
		obj[nobj].t = ELMS;
		obj[nobj].o.s = sphes[oo];
		obj[nobj].sm = &obj[nobj].o.s.sm;
		obj[nobj].name = obj[nobj].o.s.name;
	} 

	/* initialize random seed: */
	srand((unsigned int) time(NULL));

	/* prepare image */
	imgf.buf = (float *) calloc(imgf.height * imgf.width * 3, sizeof(float));

	/* prepare camera */
	prepare_cam(&cam, &imgf);

	/* load the sky */
	//bmp_read("skybox_texture.bmp", &sky);
	//bmp_read("grimmnight_large.bmp", &sky);
	//bmp_read("skyboxsun25degtest.bmp", &sky);

	/* start from black image */
	for (row = 0; row < imgf.height; row++) {
		for (col = 0; col < imgf.width; col++) {
			imgf.buf[(row * imgf.width + col) * 3 + 0] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 1] = 0;
			imgf.buf[(row * imgf.width + col) * 3 + 2] = 0;
		}
	}

	/* scan image pixels */
	for (samp = 0; samp < npasses; samp++) {
		for (row = 0; row < imgf.height; row++) {
			for (col = 0; col < imgf.width; col++) {


				/* primary ray for that pixel */
				struct elmCLR ilm = {0,0,0};
				struct elmRAY pray = {0};
				primary_ray(&pray, &cam, row, col, nrays, samp);

				if (row == 311) {
					__asm nop
				};

				/* trace ray */
				ilm = trace(&pray, obj, nobj, 5);

				/* increment illumination */
				imgf.buf[(row * imgf.width + col) * 3 + 0] += ilm.r;
				imgf.buf[(row * imgf.width + col) * 3 + 1] += ilm.g;
				imgf.buf[(row * imgf.width + col) * 3 + 2] += ilm.b;
			}
			fprintf(stderr, "%d/%d (%.1f%%)\r", samp + 1, npasses, 100 * (float) (row + 1 + samp * imgf.height) / (float) (imgf.height * npasses));
		}

		/* write to file */
		if (samp == npasses - 1 || samp % 4 == 3) {
			struct elmIMGF imgf2 = imgf;
			struct elmIMG img = {imgf.width, imgf.height, 3, 255};
			img.bytes = img.maxval > 255 ? 2 : 1;
			img.stride = 16 * ((img.width * img.depth * img.bytes + 15) / 16);
			img.buffer = (unsigned char *) calloc(img.height * img.stride, 1);

			imgf2.buf = (float *) calloc(imgf2.height * imgf2.width * 3, sizeof(float));
			memcpy(imgf2.buf, imgf.buf, imgf2.height * imgf2.width * 3 * sizeof(float));
			//post_process(&imgf2, &imgf);

			for (row = 0; row < imgf.height; row++) {
				for (col = 0; col < imgf.width; col++) {
					img.buffer[row * img.stride + col * img.depth * img.bytes + 0] =
						encode_pixel(imgf2.buf[(row * imgf2.width + col) * 3 + 0] / fabsf(samp + 1));
					img.buffer[row * img.stride + col * img.depth * img.bytes + 1] =
						encode_pixel(imgf2.buf[(row * imgf2.width + col) * 3 + 1] / fabsf(samp + 1));
					img.buffer[row * img.stride + col * img.depth * img.bytes + 2] =
						encode_pixel(imgf2.buf[(row * imgf2.width + col) * 3 + 2] / fabsf(samp + 1));
				}
			}

			free(imgf2.buf);
		
			/* dump image */
			bmp_write("out.bmp", &img);

			/* free image */
			free(img.buffer);
			memset(&img, 0, sizeof(struct elmIMG));
		}
	}
	fprintf(stderr, "\n");

	/* free float image */
	free(imgf.buf);
	memset(&imgf, 0, sizeof(struct elmIMGF));

	/* free the sky */
	free(sky.buffer);
	memset(&sky, 0, sizeof(struct elmIMG));

	return 0;
}
