/* restart again from scratch */
/* conversion of smallpt using prof. Cline slides */
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* coordinate/vector/color triplets */
struct elmVEC {
	double x, y, z; /* the three components */
};

/* component-wise addition */
struct elmVEC vec_plus(struct elmVEC *a, struct elmVEC *b)
{
	struct elmVEC c = {a->x + b->x, a->y + b->y, a->z + b->z};
	return c;
}

/* component-wise subtraction */
struct elmVEC vec_minus(struct elmVEC *a, struct elmVEC *b)
{
	struct elmVEC c = {a->x - b->x, a->y - b->y, a->z - b->z};
	return c;
}

/* component-wise scalar-vector multiplication */
struct elmVEC vec_times(struct elmVEC *a, double b)
{
	struct elmVEC c = {b * a->x, b * a->y, b * a->z};
	return c;
}

/* component-wise vector-vector multiplication */
struct elmVEC vec_mult(struct elmVEC *a, struct elmVEC *b)
{
	struct elmVEC c = {a->x * b->x, a->y * b->y, a->z * b->z};
	return c;
}

/* normalization to 1 */
struct elmVEC vec_norm(struct elmVEC *a)
{
	*a = vec_times(a, 1 / sqrt(a->x * a->x + a->y * a->y + a->z * a->z));
	return *a;
}

/* scalar (dot) product of vectors */
double vec_dot(struct elmVEC *a, struct elmVEC *b)
{
	return (a->x * b->x + a->y * b->y + a->z * b->z);
}

/* vector (cross) product of vectors */
struct elmVEC vec_cross(struct elmVEC *a, struct elmVEC *b)
{
	struct elmVEC c = {a->y * b->z - a->z * b->y, a->z * b->x - a->x * b->z, a->x * b->y - a->y * b->x};
	return c;
}

/* a ray */ 
struct elmRAY {
	struct elmVEC orig; /* origin */
	struct elmVEC dir; /* direction */
};

/* a sphere */
struct elmSPHERE {
	double rad; /* radius */
	struct elmVEC pos; /* position */
	struct elmVEC em; /* emission */
	struct elmVEC clr; /* color */
	enum {elmDIFF, elmSPEC, elmREFR} refl; /* reflection type: diffuse, specular, refractive */
};

/* intersection between sphere and ray: returns distance or 0 if miss */
double sphere_intersect(struct elmSPHERE *sph, struct elmRAY *ray)
{
	/* solve the 2nd degree equation */
	struct elmVEC op = vec_minus(&sph->pos, &ray->orig);
	double t, eps = 1E-4; /* fudge factor */
	double b = vec_dot(&op, &ray->dir); /* in quadratic solver */
	double det = b * b - vec_dot(&op, &op) + sph->rad * sph->rad;
	if (det < 0)
		return 0; /* ray misses sphere */
	else
		det = sqrt(det);
	t = b - det;
	if (t > eps)
		return t;
	else {
		t = b + det;
		if (t > eps)
			return t;
		else
			return 0;
	} /* return smaller positive t */
}

/* Scene: radius, position, emission, color, material */
struct elmSPHERE spheres[] = {
	{1e5, {1e5+1,40.8,81.6}, {0}, {.75,.25,.25}, elmDIFF}, //Left
	{1e5, {-1e5+99,40.8,81.6}, {0}, {.25,.25,.75}, elmDIFF}, //Rght
	{1e5, {50,40.8, 1e5}, {0}, {.75,.75,.75}, elmDIFF},//Back
	{1e5, {50,40.8,-1e5+170}, {0}, {0}, elmDIFF},//Frnt
	{1e5, {50, 1e5, 81.6}, {0}, {.75,.75,.75}, elmDIFF},//Botm
	{1e5, {50,-1e5+81.6,81.6}, {0}, {.75,.75,.75}, elmDIFF},//Top
	{16.5, {27,16.5,47}, {0}, {.999,.999,.999}, elmSPEC},//Mirr
	{16.5, {73,16.5,78}, {0}, {.999,.999,.999}, elmREFR},//Glas
	{600, {50,681.6-.27,81.6}, {12,12,12}, {0}, elmDIFF} //Lite
};

int numSpheres = sizeof(spheres) / sizeof(struct elmSPHERE);

/* clamping function */
_inline double clamp(double x)
{
	return x<0 ? 0 : x>1 ? 1 : x;
}

/* double to integer conversion for saving in file */
_inline int toInt(double x)
{
	return (int) (pow(clamp(x), 1/2.2) * 255 + .5);
}

/* intersects ray with scene */
_inline int intersect(struct elmRAY *ray, double *t, int *id)
{
  double n = sizeof(spheres) / sizeof(struct elmSPHERE), d, inf = *t = 1e20;
  int i;
  for (i = (int) n; i--;) {
	  if ((d = sphere_intersect(&spheres[i], ray)) && d < *t) {
		  *t = d;
		  *id = i;
	  }
  }

  return *t < inf;
}

/* random number between 0 and 1 */
double erand48(unsigned short xseed[3])
{
	return (double) rand() / (double) RAND_MAX;
}

// radiance along ray
struct elmVEC radiance(struct elmRAY *ray, int depth, unsigned short *Xi, int E)
{
	double t; // distance to intersection
	int id = 0; // id of intersected object
	struct elmVEC vec = {0}, x, x1, n, nl, f, vec1;
	struct elmSPHERE *obj;
	double p;

	if (!intersect(ray, &t, &id))
		return vec; // if miss, return black

	obj = &spheres[id]; // the hit object

	if (depth > 10)
		return vec;

	/* ray intersection point */
	x1 = vec_times(&ray->dir, t);
	x = vec_plus(&ray->orig, &x1);

	// sphere normal
	n = vec_minus(&x, &obj->pos);
	vec_norm(&n);

	// properly oriented surface normal
	nl = vec_dot(&n, &ray->dir) < 0 ? n : vec_times(&n, -1); 

	/* object color */
	f = obj->clr;

	/* use maximum reflectivity amount for Russian roulette */
	p = f.x > f.y && f.x > f.z ? f.x : f.y > f.z ? f.y : f.z; // max refl
	if (++depth > 5 || !p)
		if (erand48(Xi) < p)
			f = vec_times(&f, 1 / p);
		else
			return vec_times(&obj->em, E); //R.R.

	/* depends on object material */
	if (obj->refl == elmDIFF) {

		// Ideal DIFFUSE reflection
		double r1 = 2 * M_PI * erand48(Xi); // angle around
		double r2 = erand48(Xi), r2s = sqrt(r2); // distance from center
		struct elmVEC w = nl; // w =  normal
		struct elmVEC u1 = {0,1,0}, u2 = {1,0,0}, u3 = fabs(w.x) > .1 ? u1 : u2, u = vec_cross(&u3, &w),
			v, d, d1, d2, d3, d4;
		struct elmRAY ray2;
		vec_norm(&u); // u is perpendicular to w
		v = vec_cross(&w, &u); // v is perpendicular to v and w
		d1 = vec_times(&u, cos(r1) * r2s);
		d2 = vec_times(&v, sin(r1) * r2s);
		d3 = vec_times(&w, sqrt(1 - r2));
		d4 = vec_plus(&d1, &d2);
		d = vec_plus(&d4, &d3); // d is a random reflection ray
		vec_norm(&d);
		ray2.orig = x;
		ray2.dir = d;
		vec1 = radiance(&ray2, depth, Xi, 0); // recursive call with random ray and no emissive term
		vec1 = vec_mult(&f, &vec1);

		return vec_plus(&obj->em, &vec1);

	} else if (obj->refl == elmSPEC) {

		// Ideal SPECULAR reflection
		struct elmRAY ray2;
		ray2.orig = x;
		ray2.dir = vec_times(&n, 2 * vec_dot(&n, &ray->dir));
		vec1 = radiance(&ray2, depth, Xi, 1);
		vec1 = vec_mult(&f, &vec1);

		return vec_plus(&obj->em, &vec1);
	
	} else {

		/* REFRACTION on dielectric surface - glass */
		struct elmRAY reflRay;     
		int into = vec_dot(&n, &nl) > 0; /* Ray from outside going in? */
		double nc = 1, nt = 1.5, nnt = into ? nc / nt : nt / nc, ddn = vec_dot(&ray->dir, &nl), cos2t;
		
		/* Ideal dielectric REFlection */
		reflRay.orig = x;
		reflRay.dir = vec_times(&n, 2 * vec_dot(&n, &ray->dir));

		/* if total internal reflection, reflect */
		if ((cos2t = 1 - nnt * nnt * (1 - ddn * ddn)) < 0) {

			/* Total internal reflection */
			vec1 = radiance(&reflRay, depth, Xi, 1);
			vec1 = vec_mult(&f, &vec1);
			return vec_plus(&obj->em, &vec1);

		} else {

			/* choose reflection or refraction */
			struct elmVEC tdir1 = vec_times(&ray->dir, nnt),
				tdir2 = vec_times(&n, (into ? 1 : -1) * (ddn * nnt + sqrt(cos2t))),
				tdir3 = vec_minus(&tdir1, &tdir2), tdir = vec_norm(&tdir3), vec2, vec3;
			double a = nt - nc, b = nt + nc, R0 = a * a / (b * b), c = 1 - (into ? -ddn : vec_dot(&tdir, &n));
			double Re = R0 + (1 - R0) * c * c * c * c * c, Tr = 1 - Re, P = .25 + .5 * Re, RP = Re / P, TP = Tr / (1 - P);
			if (depth > 2) {
				// Russian roulette
				if (erand48(Xi) < P) {
					vec1 = radiance(&reflRay, depth, Xi, 1);
					vec2 = vec_times(&vec1, RP);
				} else {
					struct elmRAY ray2;
					ray2.orig = x;
					ray2.dir = tdir;
					vec1 = radiance(&ray2, depth, Xi, 1);
					vec2 = vec_times(&vec1, TP);
				}
			} else {
				struct elmRAY ray2;
				ray2.orig = x;
				ray2.dir = tdir;
				vec = radiance(&reflRay, depth, Xi, 1);
				vec = vec_times(&vec, Re);
				vec1 = radiance(&ray2, depth, Xi, 1);
				vec1 = vec_times(&vec1, Tr);
				vec2 = vec_plus(&vec, &vec1);
			}

			vec3 = vec_mult(&f, &vec2);
			return vec_plus(&obj->em, &vec3);
		}
	}
}

/* main function */
int main(int argc, char **argv)
{
	int w = 1024, h = 768; // image size
	int samps = argc == 2 ? atoi(argv[1]) / 4 : 2; // # samples, default 1
	struct elmRAY cam = {{50,52,295.6}, {0,-0.042612,-1}}; // camera pos, dir
	struct elmVEC cx = {w * .5135 / h, 0, 0}; // x dir increment
	struct elmVEC cy; // y direction increment with cross product
	struct elmVEC r; // colors of smples
	struct elmVEC *c = (struct elmVEC *) calloc(w * h, sizeof(struct elmVEC));
	int y, i;
	FILE *f;

	vec_norm(&cam.dir);
	cy = vec_cross(&cx, &cam.dir);
	vec_norm(&cy);
	cy = vec_times(&cy, .5135);

	for (y = 0; y < h; y++) { // Loop over image rows
		unsigned short x = 0, Xi[3] = {0, 0, y * y * y};
		int sx, sy, s;
		fprintf(stderr, "\rRendering (%d spp) %5.2f%%", samps * 4, 100. * y / (h - 1));
		for (x = 0; x < w; x++)   // Loop cols
			for (sy = 0, i = (h - y - 1) * w + x; sy < 2; sy++)     // 2x2 subpixel rows
				for (sx = 0; sx < 2; sx++, r.x = r.y = r.z = 0) {        // 2x2 subpixel cols
					struct elmVEC r1, r2;
					for (s = 0; s < samps; s++) {
						/* tent filter? */
						double r1 = 2 * erand48(Xi), dx = r1 < 1 ? sqrt(r1) - 1 : 1 - sqrt(2 - r1);
						double r2 = 2 * erand48(Xi), dy = r2 < 1 ? sqrt(r2) - 1 : 1 - sqrt(2 - r2);
						struct elmVEC d1 = vec_times(&cx, ((sx + .5 + dx) / 2 + x) / w - .5);
						struct elmVEC d2 = vec_times(&cy, ((sy + .5 + dy) / 2 + y) / h - .5);
						struct elmVEC d3 = vec_plus(&d1, &d2);
						struct elmVEC d = vec_plus(&d3, &cam.dir);
						struct elmVEC d4 = vec_times(&d, 140);
						struct elmRAY ray;
						struct elmVEC vec;
						ray.orig = vec_plus(&cam.orig, &d4);
						ray.dir = vec_norm(&d);
						vec = radiance(&ray, 0, Xi, 1);
						vec = vec_times(&vec, 1.0 / samps);
						r = vec_plus(&r, &vec);
					} // Camera rays are pushed ^^^^^ forward to start in interior
					r1.x = clamp(r.x), r1.y = clamp(r.y), r1.z = clamp(r.z);
					r2 = vec_times(&r1, .25);
					c[i] = vec_plus(&c[i], &r2);
				}
	}
	f = fopen("image.ppm", "w");         // Write image to PPM file.
	fprintf(f, "P3\n%d %d\n%d\n", w, h, 255);
	for (i = 0; i < w * h; i++)
		fprintf(f,"%d %d %d ", toInt(c[i].x), toInt(c[i].y), toInt(c[i].z));

  
  return EXIT_SUCCESS;
}
